// SciTE - Scintilla based Text Editor
// KeyWords.cxx - colourise for particular languages
// Copyright 1998-2000 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>

#include "Platform.h"

#ifdef PLAT_GTK
#include "WinDefs.h"
#endif

#include "PropSet.h"
#include "KeyWords.h"
#include "Scintilla.h"

#ifdef PLAT_GTK
LRESULT SendMessage(WindowID w, UINT msg, WPARAM wParam = 0, LPARAM lParam = 0);
#endif

inline bool IsLeadByte(int codePage, char ch) {
#ifdef PLAT_GTK
	// TODO: support DBCS under GTK+
	return false;
#else
	return codePage && IsDBCSLeadByteEx(codePage, ch);
#endif
}

bool wordInList(char *word, char **list) {
	if (0 == list)
		return false;
	for (int i = 0; list[i][0]; i++) {
		// Initial test is to mostly avoid slow function call
		if ((list[i][0] == word[0]) && (0 == strcmp(list[i], word)))
			return true;
	}
	return false;
}

bool wordInListInsensitive(char *word, char **list) {
	if (0 == list)
		return false;
	for (int i = 0; list[i][0]; i++) {
		// Initial test is to mostly avoid slow function call
		if ((toupper(list[i][0]) == toupper(word[0])) && 
			(EqualCaseInsensitive(list[i], word)))
			return true;
	}
	return false;
}

void colourSegHwnd(WindowID hwnd, unsigned int start, unsigned int end, char chAttr) {
	// Only perform styling if non empty range
	if (end != start - 1) {
		if (end < start) {
			Platform::DebugPrintf("Bad colour positions %d - %d\n", start, end);
		}
		//Platform::DebugPrintf("Colour %d %0d-%0d\n", chAttr, start, end);
		SendMessage(hwnd, SCI_SETSTYLING, end - start + 1, chAttr);
	}
}

void classifyWord(char *cdoc, unsigned int start, unsigned int end, const char *, char **keywords, WindowID hwnd) {
	char s[100];
	bool wordIsNumber = isdigit(cdoc[start]) || (cdoc[start] == '.');
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = 0;
	if (wordIsNumber)
		chAttr = 4;
	else {
		if (wordInList(s, keywords))
			chAttr = 5;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
}

inline bool iswordchar(char ch) {
	return isalnum(ch) || ch == '.' || ch == '_';
}

inline bool iswordstart(char ch) {
	return isalnum(ch) || ch == '_';
}

inline bool isoperator(char ch) {
	if (isalnum(ch))
		return false;
	// '.' left out as it is used to make up numbers
	if (ch == '%' || ch == '^' || ch == '&' || ch == '*' ||
	        ch == '(' || ch == ')' || ch == '-' || ch == '+' ||
	        ch == '=' || ch == '|' || ch == '{' || ch == '}' ||
	        ch == '[' || ch == ']' || ch == ':' || ch == ';' ||
	        ch == '<' || ch == '>' || ch == ',' || ch == '/' ||
	        ch == '?' || ch == '!' || ch == '.' || ch == '~')
		return true;
	return false;
}

enum eState {
    eDefault = 0,
    eComment = 1,
    eLineComment = 2,
    eDocComment = 3,
    eNumber = 4,
    eWord = 5,
    eString = 6,
    eChar = 7,
    ePunct = 8,
    ePreProc = 9,
    eOperator = 10,
    eIdentifier = 11
};

void classifyWordCpp(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd) {
	char s[100];
	bool wordIsNumber = isdigit(cdoc[start]) || (cdoc[start] == '.');
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = eIdentifier;
	if (wordIsNumber)
		chAttr = eNumber;
	else {
		if (wordInList(s, keywords))
			chAttr = eWord;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
}

static void ColouriseCppDoc(int codePage, char *cdoc, int lengthDoc, int initStyle, char **keywords, WindowID hwnd) {
	eState state = static_cast<eState>(initStyle);
	char chPrev = ' ';
	char chNext = cdoc[0];
	int startSeg = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		char ch = chNext;
		chNext = ' ';
		if (i + 1 < lengthDoc)
			chNext = cdoc[i + 1];
		if (IsLeadByte(codePage, ch)) {	// dbcs
			if (i + 2 < lengthDoc) {
				chNext = cdoc[i + 2];
			}
			chPrev = ' ';
			i += 1;
			continue;
		}

		if (state == eDefault) {
			if (iswordstart(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eWord;
				startSeg = i;
			} else if (ch == '/' && chNext == '*') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eComment;
				startSeg = i;
			} else if (ch == '/' && chNext == '/') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eLineComment;
				startSeg = i;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eString;
				startSeg = i;
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eChar;
				startSeg = i;
			} else if (ch == '#') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = ePreProc;
				startSeg = i;
			} else if (isoperator(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				colourSegHwnd(hwnd, i, i, eOperator);
				startSeg = i + 1;
			}
		} else if (state == eWord) {
			if (!iswordchar(ch)) {
				classifyWordCpp(cdoc, startSeg, i - 1, keywords, hwnd);
				state = eDefault;
				startSeg = i;
				if (ch == '/' && chNext == '*') {
					state = eComment;
				} else if (ch == '/' && chNext == '/') {
					state = eLineComment;
				} else if (ch == '\"') {
					state = eString;
				} else if (ch == '\'') {
					state = eChar;
				} else if (ch == '#') {
					state = ePreProc;
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, eOperator);
					state = eDefault;
					startSeg = i + 1;
				}
			}
		} else {
			if (state == ePreProc) {
				if ((ch == '\r' || ch == '\n') && (chPrev != '\\')) {
					state = eDefault;
					colourSegHwnd(hwnd, startSeg, i - 1, ePreProc);
					startSeg = i;
				}
			} else if (state == eComment) {
				if (ch == '/' && chPrev == '*' && ((i > startSeg + 2) || ((initStyle == eComment) && (startSeg == 0)))) {
					state = eDefault;
					colourSegHwnd(hwnd, startSeg, i, eComment);
					startSeg = i + 1;
				}
			} else if (state == eLineComment) {
				if (ch == '\r' || ch == '\n') {
					colourSegHwnd(hwnd, startSeg, i - 1, eLineComment);
					state = eDefault;
					startSeg = i;
				}
			} else if (state == eString) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\"') {
					colourSegHwnd(hwnd, startSeg, i, eString);
					state = eDefault;
					i++;
					ch = chNext;
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
					startSeg = i;
				}
			} else if (state == eChar) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\'') {
					colourSegHwnd(hwnd, startSeg, i, eChar);
					state = eDefault;
					i++;
					ch = chNext;
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
					startSeg = i;
				}
			}
			if (state == eDefault) {    // One of the above succeeded
				if (ch == '/' && chNext == '*') {
					state = eComment;
				} else if (ch == '/' && chNext == '/') {
					state = eLineComment;
				} else if (ch == '\"') {
					state = eString;
				} else if (ch == '\'') {
					state = eChar;
				} else if (ch == '#') {
					state = ePreProc;
				} else if (iswordstart(ch)) {
					state = eWord;
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, eOperator);
					startSeg = i + 1;
				}
			}
		}
		chPrev = ch;
	}
	if (startSeg < lengthDoc)
		colourSegHwnd(hwnd, startSeg, lengthDoc - 1, state);
}

static void ColouriseJavaDoc(int codePage, char *cdoc, int lengthDoc, int initStyle, const char *language, char **keywords, WindowID hwnd) {
	eState state = static_cast<eState>(initStyle);
	char chPrev = ' ';
	char chNext = cdoc[0];
	char chNext2 = cdoc[0];
	int startSeg = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		char ch = chNext;
		chNext = ' ';
		if (i + 1 < lengthDoc)
			chNext = cdoc[i + 1];
		if (i + 2 < lengthDoc)
			chNext2 = cdoc[i + 2];

		if (IsLeadByte(codePage, ch)) {	// dbcs
			if (i + 2 < lengthDoc) {
				chNext = cdoc[i + 2];
			}
			if (i + 3 < lengthDoc) {
				chNext2 = cdoc[i + 3];
			}
			chPrev = ' ';
			i += 1;
			continue;
		}

		if (state == eDefault) {
			if (iswordstart(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eWord;
				startSeg = i;
			} else if (ch == '/' && chNext == '*') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				if (chNext2 == '*')
					state = eDocComment;
				else
					state = eComment;
				startSeg = i;
			} else if (ch == '/' && chNext == '/') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eLineComment;
				startSeg = i;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eString;
				startSeg = i;
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eChar;
				startSeg = i;
			} else if (isoperator(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				colourSegHwnd(hwnd, i, i, eOperator);
				startSeg = i + 1;
			}
		} else if (state == eWord) {
			if (!iswordchar(ch)) {
				classifyWord(cdoc, startSeg, i - 1, language, keywords, hwnd);
				state = eDefault;
				startSeg = i;
				if (ch == '/' && chNext == '*') {
					if (chNext2 == '*')
						state = eDocComment;
					else
						state = eComment;
				} else if (ch == '/' && chNext == '/') {
					state = eLineComment;
				} else if (ch == '\"') {
					state = eString;
				} else if (ch == '\'') {
					state = eChar;
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, eOperator);
					state = eDefault;
					startSeg = i + 1;
				}
			}
		} else {
			if (state == eComment) {
				if (ch == '/' && chPrev == '*' && ((i > startSeg + 2) || ((initStyle == eComment) && (startSeg == 0)))) {
					state = eDefault;
					colourSegHwnd(hwnd, startSeg, i, eComment);
					startSeg = i + 1;
				}
			} else if (state == eDocComment) {
				if (ch == '/' && chPrev == '*' && ((i > startSeg + 2) || ((initStyle == eComment) && (startSeg == 0)))) {
					state = eDefault;
					colourSegHwnd(hwnd, startSeg, i, eDocComment);
					startSeg = i + 1;
				}
			} else if (state == eLineComment) {
				if (ch == '\r' || ch == '\n') {
					colourSegHwnd(hwnd, startSeg, i - 1, eLineComment);
					state = eDefault;
					startSeg = i;
				}
			} else if (state == eString) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\"') {
					colourSegHwnd(hwnd, startSeg, i, 7);
					state = eDefault;
					i++;
					ch = chNext;
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
					startSeg = i;
				}
			} else if (state == eChar) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\'') {
					colourSegHwnd(hwnd, startSeg, i, 7);
					state = eDefault;
					i++;
					ch = chNext;
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
					startSeg = i;
				}
			}
			if (state == eDefault) {    // One of the above succeeded
				if (ch == '/' && chNext == '*') {
					if (chNext2 == '*')
						state = eDocComment;
					else
						state = eComment;
				} else if (ch == '/' && chNext == '/') {
					state = eLineComment;
				} else if (ch == '\"') {
					state = eString;
				} else if (ch == '\'') {
					state = eChar;
				} else if (iswordstart(ch)) {
					state = eWord;
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, eOperator);
					startSeg = i + 1;
				}
			}
		}
		chPrev = ch;
	}
	if (startSeg < lengthDoc)
		colourSegHwnd(hwnd, startSeg, lengthDoc, state);
}

void classifyWordVB(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd) {
	char s[100];
	bool wordIsNumber = isdigit(cdoc[start]) || (cdoc[start] == '.');
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = 0;
	if (wordIsNumber)
		chAttr = 4;
	else {
		if (wordInListInsensitive(s, keywords))
			chAttr = 5;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
}

static void ColouriseVBDoc(int codePage, char *cdoc, int lengthDoc, int initStyle, const char *, char **keywords, WindowID hwnd) {
	eState state = static_cast<eState>(initStyle);
	char chPrev = ' ';
	char chNext = cdoc[0];
	char chNext2 = cdoc[0];
	int startSeg = 0;
	for (int i = 0; i < lengthDoc; i++) {
		char ch = chNext;
		chNext = ' ';
		if (i + 1 < lengthDoc)
			chNext = cdoc[i + 1];
		if (i + 2 < lengthDoc)
			chNext2 = cdoc[i + 2];

		if (IsLeadByte(codePage, ch)) {	// dbcs
			if (i + 2 < lengthDoc) {
				chNext = cdoc[i + 2];
			}
			if (i + 3 < lengthDoc) {
				chNext2 = cdoc[i + 3];
			}
			chPrev = ' ';
			i += 1;
			continue;
		}

		if (state == eDefault) {
			if (iswordstart(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eWord;
				startSeg = i;
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eLineComment;
				startSeg = i;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, eDefault);
				state = eString;
				startSeg = i;
			}
		} else if (state == eWord) {
			if (!iswordchar(ch)) {
				if ((i - startSeg == 3) &&
				        toupper(cdoc[startSeg]) == 'R' &&
				        toupper(cdoc[startSeg + 1]) == 'E' &&
				        toupper(cdoc[startSeg + 2]) == 'M' ) {
					colourSegHwnd(hwnd, startSeg, i - 1, eLineComment);
					state = eLineComment;
					startSeg = i;
				} else {
					classifyWordVB(cdoc, startSeg, i - 1, keywords, hwnd);
					state = eDefault;
					startSeg = i;
					if (ch == '\'') {
						state = eLineComment;
					} else if (ch == '\"') {
						state = eString;
					}
				}
			}
		} else {
			if (state == eLineComment) {
				if (ch == '\r' || ch == '\n') {
					colourSegHwnd(hwnd, startSeg, i - 1, eLineComment);
					state = eDefault;
					startSeg = i;
				}
			} else if (state == eString) {
				// VB doubles quotes to preserve them
				if (ch == '\"') {
					colourSegHwnd(hwnd, startSeg, i, 7);
					state = eDefault;
					i++;
					ch = chNext;
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
					startSeg = i;
				}
			}
			if (state == eDefault) {    // One of the above succeeded
				if (ch == '\'') {
					state = eLineComment;
				} else if (ch == '\"') {
					state = eString;
				} else if (iswordstart(ch)) {
					state = eWord;
				}
			}
		}
		chPrev = ch;
	}
	if (startSeg < lengthDoc)
		colourSegHwnd(hwnd, startSeg, lengthDoc, state);
}

enum ePyState {
    ePyDefault = 0,
    ePyComment = 1,
    ePyNumber = 2,
    ePyString = 3,
    ePyChar = 4,
    ePyWord = 5,
    eTriple = 6,
    eTripleDouble = 7,
    eClassName = 8,
    eDefName = 9,
    ePyOperator = 10,
    ePyIdentifier = 11,
    ePyCommentEOL = 12	// Only used in pyro to get block comments
};

void classifyWordPy(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd, char *prevWord) {
	char s[100];
	bool wordIsNumber = isdigit(cdoc[start]);
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = ePyIdentifier;
	if (0 == strcmp(prevWord, "class"))
		chAttr = eClassName;
	else if (0 == strcmp(prevWord, "def"))
		chAttr = eDefName;
	else if (wordIsNumber)
		chAttr = ePyNumber;
	else if (wordInList(s, keywords))
		chAttr = ePyWord;
	colourSegHwnd(hwnd, start, end, chAttr);
	strcpy(prevWord, s);
}

static void ColourisePyDoc(int codePage, char *cdoc, int lengthDoc, int initStyle, char **keywords, WindowID hwnd) {
	//Platform::DebugPrintf("Python coloured\n");
	char prevWord[200];
	prevWord[0] = '\0';
	if (lengthDoc == 0)
		return;
	ePyState state = static_cast<ePyState>(initStyle);
	char chPrev = ' ';
	char chPrev2 = ' ';
	char chNext = cdoc[0];
	char chNext2 = cdoc[0];
	int startSeg = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		char ch = chNext;
		chNext = ' ';
		if (i + 1 < lengthDoc)
			chNext = cdoc[i + 1];
		chNext2 = ' ';
		if (i + 2 < lengthDoc)
			chNext2 = cdoc[i + 2];

		if (IsLeadByte(codePage, ch)) {	// dbcs
			if (i + 2 < lengthDoc) {
				chNext = cdoc[i + 2];
			}
			if (i + 3 < lengthDoc) {
				chNext2 = cdoc[i + 3];
			}
			chPrev = ' ';
			chPrev2 = ' ';
			i += 1;
			continue;
		}

		if (state == ePyDefault) {
			if (iswordstart(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, ePyDefault);
				state = ePyWord;
				startSeg = i;
			} else if (ch == '#') {
				colourSegHwnd(hwnd, startSeg, i - 1, ePyDefault);
				state = ePyComment;
				startSeg = i;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, ePyDefault);
				startSeg = i;
				if (chNext == '\"' && chNext2 == '\"') {
					i += 2;
					state = eTripleDouble;
					ch = ' ';
					chPrev = ' ';
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
				} else {
					state = ePyString;
				}
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, ePyDefault);
				startSeg = i;
				if (chNext == '\'' && chNext2 == '\'') {
					i += 2;
					state = eTriple;
					ch = ' ';
					chPrev = ' ';
					chNext = ' ';
					if (i + 1 < lengthDoc)
						chNext = cdoc[i + 1];
				} else {
					state = ePyChar;
				}
			} else if (isoperator(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, ePyDefault);
				colourSegHwnd(hwnd, i, i, ePyOperator);
				startSeg = i + 1;
			}
		} else if (state == ePyWord) {
			if (!iswordchar(ch)) {
				classifyWordPy(cdoc, startSeg, i - 1, keywords, hwnd, prevWord);
				state = ePyDefault;
				startSeg = i;
				if (ch == '#') {
					state = ePyComment;
				} else if (ch == '\"') {
					if (chNext == '\"' && chNext2 == '\"') {
						i += 2;
						state = eTripleDouble;
						ch = ' ';
						chPrev = ' ';
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					} else {
						state = ePyString;
					}
				} else if (ch == '\'') {
					if (chNext == '\'' && chNext2 == '\'') {
						i += 2;
						state = eTriple;
						ch = ' ';
						chPrev = ' ';
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					} else {
						state = ePyChar;
					}
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, ePyOperator);
					state = ePyDefault;
					startSeg = i + 1;
				}
			}
		} else {
			if (state == ePyComment) {
				if (ch == '\r' || ch == '\n') {
					colourSegHwnd(hwnd, startSeg, i - 1, ePyComment);
					state = ePyDefault;
					startSeg = i;
				}
			} else if (state == ePyString) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\"') {
					colourSegHwnd(hwnd, startSeg, i, ePyString);
					state = ePyDefault;
					startSeg = i + 1;
				}
			} else if (state == ePyChar) {
				if (ch == '\\') {
					if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
						i++;
						ch = chNext;
						chNext = ' ';
						if (i + 1 < lengthDoc)
							chNext = cdoc[i + 1];
					}
				} else if (ch == '\'') {
					colourSegHwnd(hwnd, startSeg, i, ePyChar);
					state = ePyDefault;
					startSeg = i + 1;
				}
			} else if (state == eTriple) {
				if (ch == '\'' && chPrev == '\'' && chPrev2 == '\'') {
					colourSegHwnd(hwnd, startSeg, i, eTriple);
					state = ePyDefault;
					startSeg = i + 1;
				}
			} else if (state == eTripleDouble) {
				if (ch == '\"' && chPrev == '\"' && chPrev2 == '\"') {
					colourSegHwnd(hwnd, startSeg, i, eTripleDouble);
					state = ePyDefault;
					startSeg = i + 1;
				}
			}
		}
		chPrev2 = chPrev;
		chPrev = ch;
	}
	if (startSeg <= lengthDoc) {
		if (state == ePyDefault) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, ePyDefault);
		} else if (state == ePyWord) {
			classifyWordPy(cdoc, startSeg, lengthDoc, keywords, hwnd, prevWord);
		} else if (state == ePyComment) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, ePyComment);
		} else if (state == ePyString) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, ePyString);
		} else if (state == ePyChar) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, ePyChar);
		} else if (state == eTriple) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, eTriple);
		} else if (state == eTripleDouble) {
			colourSegHwnd(hwnd, startSeg, lengthDoc, eTripleDouble);
		}
	}
}

static void ColouriseBatchLine(char *lineBuffer, int lengthLine, WindowID hwnd) {
	if (0 == strncmp(lineBuffer, "REM", 3)) {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 1);
	} else if (0 == strncmp(lineBuffer, "rem", 3)) {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 1);
	} else if (0 == strncmp(lineBuffer, "SET", 3)) {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 2);
	} else if (0 == strncmp(lineBuffer, "set", 3)) {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 2);
	} else if (lineBuffer[0] == ':') {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 3);
	} else {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 0);
	}
}

static void ColouriseBatchDoc(char *cdoc, int lengthDoc, int, WindowID hwnd) {
	char lineBuffer[1024];
	unsigned int linePos = 0;
	for (int i = 0; i < lengthDoc; i++) {
		lineBuffer[linePos++] = cdoc[i];
		if (cdoc[i] == '\r' || cdoc[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
			ColouriseBatchLine(lineBuffer, linePos, hwnd);
			linePos = 0;
		}
	}
	if (linePos > 0)
		ColouriseBatchLine(lineBuffer, linePos, hwnd);
}

enum eHTState {
    eHTDefault = 0,
    eHTTag = 1,
    eHTUnknownTag = 2,
    eHTAttrib = 3,
    eHTUnknownAttrib = 4,
    eHTNumber = 5,
    eHTDString = 6,
    eHTSString = 7,
    eHTOther = 8,
    eHTComment = 9,
    eHTEntity = 10,
    eHTJSStartScript = 11,
    eHTJSDefault = 12,
    eHTJSComment = 13,
    eHTJSLineComment = 14,
    eHTJSDocComment = 15,
    eHTJSNumber = 16,
    eHTJSWord = 17,
    eHTJSKeyword = 18,
    eHTJSDQString = 19,
    eHTJSSQString = 20,
    eHTJSSymbols = 21
};

void classifyAttribHTML(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd) {
	char s[100];
	s[0] = '\0';
	bool wordIsNumber = isdigit(cdoc[start]) || (cdoc[start] == '.') ||
	                    (cdoc[start] == '-') || (cdoc[start] == '#');
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = eHTUnknownAttrib;
	if (wordIsNumber)
		chAttr = eHTNumber;
	else {
		if (wordInListInsensitive(s, keywords))
			chAttr = eHTAttrib;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
}

// Yeah, I know
static bool lastWasScript = false;

eHTState classifyTagHTML(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd) {
	lastWasScript = false;
	char s[100];
	// Copy after the '<'
	s[0] = '\0';
	for (unsigned int i = 0; i < end - start && i < 30; i++) {
		s[i] = cdoc[start + i + 1];
		s[i + 1] = '\0';
	}
	eHTState chAttr = eHTUnknownTag;
	if (s[0] == '!' && s[1] == '-' && s[2] == '-') {	//Comment
		chAttr = eHTComment;
	} else if (s[0] == '/') {	// Closing tag
		if (wordInListInsensitive(s + 1, keywords))
			chAttr = eHTTag;
	} else {
		if (wordInListInsensitive(s, keywords))
			chAttr = eHTTag;
		if (EqualCaseInsensitive(s, "SCRIPT"))
			lastWasScript = true;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
	return chAttr;
}

void classifyWordHTJS(char *cdoc, unsigned int start, unsigned int end, char **keywords, WindowID hwnd) {
	char s[100];
	bool wordIsNumber = isdigit(cdoc[start]) || (cdoc[start] == '.');
	for (unsigned int i = 0; i < end - start + 1 && i < 30; i++) {
		s[i] = cdoc[start + i];
		s[i + 1] = '\0';
	}
	char chAttr = eHTJSWord;
	if (wordIsNumber)
		chAttr = eHTJSNumber;
	else {
		if (wordInList(s, keywords))
			chAttr = eHTJSKeyword;
	}
	colourSegHwnd(hwnd, start, end, chAttr);
}

inline bool ishtmlwordchar(char ch) {
	return isalnum(ch) || ch == '.' || ch == '-' || ch == '_' || ch == ':' || ch == '!' || ch == '#';
}

static void ColouriseHyperTextDoc(int codePage, char *cdoc, int lengthDoc, int initStyle,
	                                  char **keywords, char **keywords2, WindowID hwnd) {
	eHTState state = static_cast<eHTState>(initStyle);
	char chPrev = ' ';
	int startSeg = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		char ch = cdoc[i];
		char chNext = ' ';
		if (i + 1 < lengthDoc)
			chNext = cdoc[i + 1];
		char chNext2 = ' ';
		if (i + 2 < lengthDoc)
			chNext2 = cdoc[i + 2];
		if (IsLeadByte(codePage, ch)) {	// dbcs
			chPrev = ' ';
			i += 1;
			continue;
		}

		if (state == eHTDefault) {
			if (ch == '<') {
				colourSegHwnd(hwnd, startSeg, i - 1, eHTDefault);
				state = eHTUnknownTag;
				startSeg = i;
			} else if (ch == '&') {
				colourSegHwnd(hwnd, startSeg, i - 1, eHTDefault);
				state = eHTEntity;
				startSeg = i;
			}
		} else if (state == eHTComment) {
			if ((ch == '>') && (chPrev == '-')) {
				colourSegHwnd(hwnd, startSeg, i, state);
				startSeg = i + 1;
				state = eHTDefault;
			}
		} else if (state == eHTEntity) {
			if (ch == ';') {
				colourSegHwnd(hwnd, startSeg, i, state);
				startSeg = i + 1;
				state = eHTDefault;
			}
		} else if (state == eHTUnknownTag) {
			if (!ishtmlwordchar(ch) && ch != '/' && ch != '-') {
				eHTState eClass = classifyTagHTML(cdoc, startSeg, i - 1, keywords, hwnd);
				if (ch == '>') {
					colourSegHwnd(hwnd, i, i, eHTTag);
					startSeg = i + 1;
					if (lastWasScript)
						state = eHTJSStartScript;
					else
						state = eHTDefault;
				} else {
					if (eClass == eHTComment) {
						startSeg = i;
						state = eHTComment;
					} else {
						startSeg = i;
						state = eHTOther;
					}
				}
			}
		} else if (state == eHTAttrib) {
			if (!ishtmlwordchar(ch) && ch != '/' && ch != '-') {
				classifyAttribHTML(cdoc, startSeg, i - 1, keywords, hwnd);
				if (ch == '>') {
					colourSegHwnd(hwnd, i, i, eHTTag);
					startSeg = i + 1;
					if (lastWasScript)
						state = eHTJSStartScript;
					else
						state = eHTDefault;
				} else {
					state = eHTOther;
					startSeg = i;
				}
			}
		} else if (state == eHTOther) {
			if (ch == '>') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				colourSegHwnd(hwnd, i, i, eHTTag);
				startSeg = i + 1;
				if (lastWasScript)
					state = eHTJSStartScript;
				else
					state = eHTDefault;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTDString;
				startSeg = i;
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTSString;
				startSeg = i;
			} else if (ishtmlwordchar(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTAttrib;
				startSeg = i;
			}
		} else if (state == eHTDString) {
			if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i, eHTDString);
				state = eHTOther;
				startSeg = i + 1;
			}
		} else if (state == eHTSString) {
			if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i, eHTSString);
				state = eHTOther;
				startSeg = i + 1;
			}
		} else if (state == eHTJSDefault || state == eHTJSStartScript) {
			if (iswordstart(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTJSWord;
				startSeg = i;
			} else if (ch == '/' && chNext == '*') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				if (chNext2 == '*')
					state = eHTJSDocComment;
				else
					state = eHTJSComment;
				startSeg = i;
			} else if (ch == '/' && chNext == '/') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTJSLineComment;
				startSeg = i;
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTJSDQString;
				startSeg = i;
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTJSSQString;
				startSeg = i;
			} else if ((ch == '<') && (chNext == '/')) {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				state = eHTUnknownTag;
				startSeg = i;
			} else if (isoperator(ch)) {
				colourSegHwnd(hwnd, startSeg, i - 1, state);
				colourSegHwnd(hwnd, i, i, eHTJSSymbols);
				state = eHTJSDefault;
				startSeg = i + 1;
			} else if ((ch == ' ') || (ch == '\t')) {
				if (state == eHTJSStartScript) {
					colourSegHwnd(hwnd, startSeg, i - 1, state);
					state = eHTJSDefault;
					startSeg = i;
				}
			}
		} else if (state == eHTJSWord) {
			if (!iswordchar(ch)) {
				classifyWordHTJS(cdoc, startSeg, i - 1, keywords2, hwnd);
				//colourSegHwnd(hwnd, startSeg, i - 1, eHTJSKeyword);
				state = eHTJSDefault;
				startSeg = i;
				if (ch == '/' && chNext == '*') {
					if (chNext2 == '*')
						state = eHTJSDocComment;
					else
						state = eHTJSComment;
				} else if (ch == '/' && chNext == '/') {
					state = eHTJSLineComment;
				} else if (ch == '\"') {
					state = eHTJSDQString;
				} else if (ch == '\'') {
					state = eHTJSSQString;
				} else if (isoperator(ch)) {
					colourSegHwnd(hwnd, startSeg, i, eHTJSSymbols);
					state = eHTJSDefault;
					startSeg = i + 1;
				}
			}
		} else if (state == eHTJSComment) {
			if (ch == '/' && chPrev == '*') {
				state = eHTJSDefault;
				colourSegHwnd(hwnd, startSeg, i, eHTJSComment);
				startSeg = i + 1;
			}
		} else if (state == eHTJSDocComment) {
			if (ch == '/' && chPrev == '*') {
				state = eHTJSDefault;
				colourSegHwnd(hwnd, startSeg, i, eHTJSDocComment);
				startSeg = i + 1;
			}
		} else if (state == eHTJSLineComment) {
			if (ch == '\r' || ch == '\n') {
				colourSegHwnd(hwnd, startSeg, i - 1, eHTJSLineComment);
				state = eHTJSDefault;
				startSeg = i;
			} else if ((ch == '<') && (chNext == '/')) {
				// Common to hide end script tag in comment
				colourSegHwnd(hwnd, startSeg, i - 1, eHTJSLineComment);
				state = eHTUnknownTag;
				startSeg = i;
			}
		} else if (state == eHTJSDQString) {
			if (ch == '\\') {
				if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
					i++;
				}
			} else if (ch == '\"') {
				colourSegHwnd(hwnd, startSeg, i, eHTJSDQString);
				state = eHTJSDefault;
				i++;
				ch = chNext;
				chNext = ' ';
				if (i + 1 < lengthDoc)
					chNext = cdoc[i + 1];
				startSeg = i;
			}
		} else if (state == eHTJSSQString) {
			if (ch == '\\') {
				if (chNext == '\"' || chNext == '\'' || chNext == '\\') {
					i++;
				}
			} else if (ch == '\'') {
				colourSegHwnd(hwnd, startSeg, i, eHTJSSQString);
				state = eHTJSDefault;
				i++;
				ch = chNext;
				chNext = ' ';
				if (i + 1 < lengthDoc)
					chNext = cdoc[i + 1];
				startSeg = i;
			}
		}
		if (state == eHTJSDefault) {    // One of the above succeeded
			if (ch == '/' && chNext == '*') {
				if (chNext2 == '*')
					state = eHTJSDocComment;
				else
					state = eHTJSComment;
			} else if (ch == '/' && chNext == '/') {
				state = eHTJSLineComment;
			} else if (ch == '\"') {
				state = eHTJSDQString;
			} else if (ch == '\'') {
				state = eHTJSSQString;
			} else if (iswordstart(ch)) {
				state = eHTJSKeyword;
			} else if (isoperator(ch)) {
				colourSegHwnd(hwnd, startSeg, i, eHTJSSymbols);
				startSeg = i + 1;
			}
		}
		chPrev = ch;
	}
	if (startSeg < lengthDoc)
		colourSegHwnd(hwnd, startSeg, lengthDoc - 1, state);
}

static void ColourisePropsLine(char *lineBuffer, int lengthLine, WindowID hwnd) {
	int i = 0;
	while (isspace(lineBuffer[i]) && (i < lengthLine))	// Skip initial spaces
		i++;
	if (lineBuffer[i] == '#' || lineBuffer[i] == '!' || lineBuffer[i] == ';') {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 1);
	}
	else if (lineBuffer[i] == '[') {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 2);
	}
	else if (lineBuffer[i] == '@') {
		colourSegHwnd(hwnd, 0, i, 4);
		if (lineBuffer[++i] == '=')
			colourSegHwnd(hwnd, i, i, 3);
		if (++i < lengthLine)
			colourSegHwnd(hwnd, i, lengthLine - 1, 0);
	}
	else {
		while (lineBuffer[i] != '=' && (i < lengthLine))	// Search the '=' character
			i++;
		if (lineBuffer[i] == '=') {
			colourSegHwnd(hwnd, 0, i - 1, 0);
			colourSegHwnd(hwnd, i, i, 3);
			if (++i < lengthLine)
				colourSegHwnd(hwnd, i, lengthLine - 1, 0);
		}
		else
			colourSegHwnd(hwnd, 0, lengthLine - 1, 0);
	}
}

static void ColourisePropsDoc(char *cdoc, int lengthDoc, int, WindowID hwnd) {
	char lineBuffer[1024];
	unsigned int linePos = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		lineBuffer[linePos++] = cdoc[i];
		if (cdoc[i] == '\r' || cdoc[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
			lineBuffer[linePos] = '\0';
			ColourisePropsLine(lineBuffer, linePos, hwnd);
			linePos = 0;
		}
	}
	if (linePos > 0)
		ColourisePropsLine(lineBuffer, linePos, hwnd);
}

static void ColouriseMakeLine(char *lineBuffer, int lengthLine, WindowID hwnd) {
	int i = 0;
	while (isspace(lineBuffer[i]) && (i < lengthLine))
		i++;
	if (lineBuffer[i] == '#' || lineBuffer[i] == '!') {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 1);
	} else {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 0);
	}
}

static void ColouriseMakeDoc(char *cdoc, int lengthDoc, int, WindowID hwnd) {
	char lineBuffer[1024];
	unsigned int linePos = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		lineBuffer[linePos++] = cdoc[i];
		if (cdoc[i] == '\r' || cdoc[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
			ColouriseMakeLine(lineBuffer, linePos, hwnd);
			linePos = 0;
		}
	}
	if (linePos > 0)
		ColouriseMakeLine(lineBuffer, linePos, hwnd);
}

static void ColouriseErrorListLine(char *lineBuffer, int lengthLine, WindowID hwnd) {
	if (lineBuffer[0] == '>') {
		// Command or return status
		colourSegHwnd(hwnd, 0, lengthLine - 1, 4);
	} else if (strstr(lineBuffer, "File \"") && strstr(lineBuffer, ", line ")) {
		colourSegHwnd(hwnd, 0, lengthLine - 1, 1);
	} else {
		// Look for <filename>:<line>:message
		int state = 0;
		for (int i = 0; i < lengthLine; i++) {
			if (state == 0 && lineBuffer[i] == ':' && isdigit(lineBuffer[i + 1])) {
				state = 1;
			} else if (state == 0 && lineBuffer[i] == '(') {
				state = 10;
			} else if (state == 1 && isdigit(lineBuffer[i])) {
				state = 2;
			} else if (state == 2 && lineBuffer[i] == ':') {
				state = 3;
			} else if (state == 2 && !isdigit(lineBuffer[i])) {
				state = 99;
			} else if (state == 10 && isdigit(lineBuffer[i])) {
				state = 11;
			} else if (state == 11 && lineBuffer[i] == ')') {
				state = 12;
			} else if (state == 12 && lineBuffer[i] == ':') {
				state = 13;
			} else if (state == 11 && !isdigit(lineBuffer[i])) {
				state = 99;
			}
		}
		if (state == 3) {
			colourSegHwnd(hwnd, 0, lengthLine - 1, 2);
		} else if (state == 13) {
			colourSegHwnd(hwnd, 0, lengthLine - 1, 3);
		} else {
			colourSegHwnd(hwnd, 0, lengthLine - 1, 0);
		}
	}
}

static void ColouriseErrorListDoc(char *cdoc, int lengthDoc, int, WindowID hwnd) {
	char lineBuffer[1024];
	unsigned int linePos = 0;
	for (int i = 0; i <= lengthDoc; i++) {
		lineBuffer[linePos++] = cdoc[i];
		if (cdoc[i] == '\r' || cdoc[i] == '\n' || (linePos >= sizeof(lineBuffer) - 1)) {
			ColouriseErrorListLine(lineBuffer, linePos, hwnd);
			linePos = 0;
		}
	}
	if (linePos > 0)
		ColouriseErrorListLine(lineBuffer, linePos, hwnd);
}

void ColouriseDoc(int codePage, char *cdoc, int startPos, int lengthDoc, int initStyle,
                  const char *language, char **keywords, char **keywords2, WindowID hwnd) {
	//Platform::DebugPrintf("ColouriseDoc <%s>\n", language);
	SendMessage(hwnd, SCI_STARTSTYLING, startPos, 31);
	if (0 == strcmp(language, "python")) {
		ColourisePyDoc(codePage, cdoc, lengthDoc, initStyle, keywords, hwnd);
	} else if (0 == strcmp(language, "batch")) {
		ColouriseBatchDoc(cdoc, lengthDoc, initStyle, hwnd);
	} else if (0 == strcmp(language, "java")) {
		ColouriseJavaDoc(codePage, cdoc, lengthDoc, initStyle, language, keywords, hwnd);
	} else if (0 == strcmp(language, "javascript")) {
		ColouriseJavaDoc(codePage, cdoc, lengthDoc, initStyle, language, keywords, hwnd);
	} else if (0 == strcmp(language, "vb")) {
		ColouriseVBDoc(codePage, cdoc, lengthDoc, initStyle, language, keywords, hwnd);
	} else if (0 == strcmp(language, "cpp")) {
		ColouriseCppDoc(codePage, cdoc, lengthDoc, initStyle, keywords, hwnd);
	} else if (0 == strcmp(language, "props")) {
		ColourisePropsDoc(cdoc, lengthDoc, initStyle, hwnd);
	} else if (0 == strcmp(language, "makefile")) {
		ColouriseMakeDoc(cdoc, lengthDoc, initStyle, hwnd);
	} else if (0 == strcmp(language, "hypertext")) {
		ColouriseHyperTextDoc(codePage, cdoc, lengthDoc, initStyle, keywords, keywords2, hwnd);
	} else if (0 == strcmp(language, "errorlist")) {
		ColouriseErrorListDoc(cdoc, lengthDoc, initStyle, hwnd);
	} else {
		// Null language means all style bytes are 0 so just mark the end - no need to fill in.
		SendMessage(hwnd, SCI_STARTSTYLING, startPos + lengthDoc - 1, 31);
		colourSegHwnd(hwnd, 0, 0, 0);
		//colourSegHwnd(hwnd, 0, lengthDoc, 0);
	}

}
