// SciTE - Scintilla based Text Editor
// SciTE.cxx - main code for the editor
// Copyright 1998-2000 by Neil Hodgson <neilh@scintilla.org>
// The License.txt file describes the conditions under which this software may be distributed.

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/stat.h>

#include "Platform.h"

#ifdef PLAT_GTK

#include <unistd.h>
#include <glib.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include "WinDefs.h"

#else

#include <io.h>
#include <process.h>
#include <mmsystem.h>
#ifdef _MSC_VER
#include <direct.h>
#endif
#ifdef __BORLANDC__
#include <dir.h>
#endif

#endif

#include "SciTE.h"
#include "PropSet.h"
#include "KeyWords.h"
#include "Scintilla.h"

#ifdef STATIC_BUILD
const char appName[] = "Sc1";
#else
const char appName[] = "SciTE";
#endif

#ifdef unix
const char pathSepString[] = "/";
const char pathSepChar = '/';
#else
const char pathSepString[] = "\\";
const char pathSepChar = '\\';
#endif

#ifdef PLAT_GTK

#define MAX_PATH 260

static void destroyDialog(GtkWidget *dialog) {
}

LRESULT SendMessage(GtkWidget *w, UINT msg, WPARAM wParam = 0, LPARAM lParam = 0) {
	return scintilla_send_message(SCINTILLA(w), msg, wParam, lParam);
}

void SetFocus(GtkWidget *hwnd) {
	SendMessage(hwnd, SCI_GRABFOCUS, 0, 0);
}

static void _fullpath(char *fullPath, const char *basePath, int size) {
	strncpy(fullPath, basePath, size);
	fullPath[size - 1] = '\0';
}

static GtkWidget *messageBoxDialog = 0;
static int messageBoxResult = 0;

static gint messageBoxKey(GtkWidget *w, GdkEventKey *event, gpointer p) {
	if (event->keyval == GDK_Escape) {
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		gtk_widget_destroy(GTK_WIDGET(w));
		messageBoxDialog = 0;
		messageBoxResult = reinterpret_cast<int>(p);
	}
	return 0;
}

static void messageBoxOK(GtkWidget *, gpointer p) {
	gtk_widget_destroy(GTK_WIDGET(messageBoxDialog));
	messageBoxDialog = 0;
	messageBoxResult = reinterpret_cast<int>(p);
}

static GtkWidget *AddMBButton(GtkWidget *dialog, const char *label,
	                              int val, bool isDefault = false) {
	GtkWidget * button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
	                   GtkSignalFunc(messageBoxOK), reinterpret_cast<gpointer>(val));
	if (isDefault) {
		GTK_WIDGET_SET_FLAGS(GTK_WIDGET(button), GTK_CAN_DEFAULT);
	}
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->action_area),
	                   button, TRUE, TRUE, 0);
	if (isDefault) {
		gtk_widget_grab_default(GTK_WIDGET(button));
	}
	gtk_widget_show(button);
	return button;
}

void SetAboutMessage(HWND wsci);

#define MB_ABOUTBOX	0x100000L

int MessageBox(GtkWidget *wParent, const char *m, const char *t = appName, int style = MB_OK) {
	if (!messageBoxDialog) {
		messageBoxResult = -1;
		messageBoxDialog = gtk_dialog_new();
		gtk_window_set_title(GTK_WINDOW(messageBoxDialog), t);
		gtk_container_border_width(GTK_CONTAINER(messageBoxDialog), 0);

		gtk_signal_connect(GTK_OBJECT(messageBoxDialog),
		                   "destroy", GtkSignalFunc(destroyDialog), 0);

		int escapeResult = IDOK;
		if ((style & 0xf) == MB_OK) {
			AddMBButton(messageBoxDialog, "OK", IDOK, true);
		} else {
			AddMBButton(messageBoxDialog, "Yes", IDYES, true);
			AddMBButton(messageBoxDialog, "No", IDNO);
			escapeResult = IDNO;
			if (style == MB_YESNOCANCEL) {
				AddMBButton(messageBoxDialog, "Cancel", IDCANCEL);
				escapeResult = IDCANCEL;
			}
		}
		gtk_signal_connect(GTK_OBJECT(messageBoxDialog),
		                   "key_press_event", GtkSignalFunc(messageBoxKey),
		                   reinterpret_cast<gpointer>(escapeResult));

		if (style & MB_ABOUTBOX) {
			GtkWidget *explanation = scintilla_new();
			scintilla_set_id(SCINTILLA(explanation), 0);
			gtk_box_pack_start(GTK_BOX(GTK_DIALOG(messageBoxDialog)->vbox),
			                   explanation, TRUE, TRUE, 0);
			gtk_widget_set_usize(GTK_WIDGET(explanation), 320, 180);
			gtk_widget_show_all(explanation);
			SetAboutMessage(explanation);
		} else {
			GtkWidget *label = gtk_label_new(m);
			gtk_misc_set_padding(GTK_MISC(label), 10, 10);
			gtk_box_pack_start(GTK_BOX(GTK_DIALOG(messageBoxDialog)->vbox),
			                   label, TRUE, TRUE, 0);
			gtk_widget_show(label);
		}

		// Mark it as a modal transient dialog
		gtk_window_set_modal(GTK_WINDOW(messageBoxDialog), TRUE);
		gtk_window_set_transient_for (GTK_WINDOW(messageBoxDialog),
		                              GTK_WINDOW(wParent));

		gtk_widget_show(messageBoxDialog);

		while (messageBoxResult < 0) {
			gtk_main_iteration();
		}
	}
	return messageBoxResult;
}

DWORD timeGetTime() {
	return 0;
}

#endif

// AddStyledText only called from About so static size buffer is OK
void AddStyledText(HWND hwnd, const char *s, int attr) {
	char buf[1000];
	int len = strlen(s);
	for (int i = 0; i < len; i++) {
		buf[i*2] = s[i];
		buf[i*2 + 1] = attr;
	}
	SendMessage(hwnd, SCI_ADDSTYLEDTEXT, len*2,
	            reinterpret_cast<LPARAM>(static_cast<char *>(buf)));
}

void SetAboutStyle(HWND wsci, int style, Colour fore, Colour back, int size) {
	SendMessage(wsci, SCI_STYLESETFORE, style, fore.AsLong());
	SendMessage(wsci, SCI_STYLESETBACK, style, back.AsLong());
	SendMessage(wsci, SCI_STYLESETSIZE, style, size);
}

void SetAboutMessage(HWND wsci) {
	if (wsci) {
		SendMessage(wsci, SCI_STYLESETBACK, STYLE_DEFAULT, Colour(0, 0, 0).AsLong());
		SendMessage(wsci, EM_SETREADONLY, 0, 0);
		SetAboutStyle(wsci, 0, Colour(0xff, 0xff, 0xff), Colour(0, 0, 0x80), 24);
#ifdef STATIC_BUILD
		AddStyledText(wsci, "Sc1  \n", 0);
#else
		AddStyledText(wsci, "SciTE\n", 0);
#endif
		SetAboutStyle(wsci, 1, Colour(0xff, 0xff, 0xff), Colour(0, 0, 0x80), 12);
		AddStyledText(wsci, "Version 1.2 \n", 1);
		SetAboutStyle(wsci, 2, Colour(0xff, 0xff, 0xff), Colour(0, 0, 0x80), 12);
		SendMessage(wsci, SCI_STYLESETITALIC, 2, 1);
		AddStyledText(wsci, "by Neil Hodgson.\n", 2);
		SetAboutStyle(wsci, 3, Colour(0xff, 0xff, 0xff), Colour(0, 0, 0x80), 12);
		AddStyledText(wsci, "December 1998-January 2000.\n", 3);
		SetAboutStyle(wsci, 4, Colour(0, 0xff, 0xff), Colour(0, 0, 0x80), 10);
		AddStyledText(wsci, "http://www.scintilla.org", 4);
	}
}

static int ControlIDOfCommand(WPARAM wParam) {
	return wParam & 0xffff;
}

#ifdef PLAT_WIN

HWND currentDlg = 0;

// DefaultDlg, DoDialog is a bit like something in PC Magazine May 28, 1991, page 357
// DefaultDlg is only used for about box
int PASCAL DefaultDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM) {
	switch (message) {

	case WM_INITDIALOG: {
			HWND wsci = GetDlgItem(hDlg, IDABOUTSCINTILLA);
			SetAboutMessage(wsci);
		}
		return TRUE;

	case WM_CLOSE:
		SendMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
		break;

	case WM_COMMAND:
		if (ControlIDOfCommand(wParam) == IDOK) {
			EndDialog(hDlg, IDOK);
			return TRUE;
		} else if (ControlIDOfCommand(wParam) == IDCANCEL) {
			EndDialog(hDlg, IDCANCEL);
			return FALSE;
		}
	}

	return FALSE;
}

int DoDialog(HINSTANCE hInst, const char *resName, HWND hWnd, DLGPROC lpProc,
             DWORD dwInitParam) {
	int result = -1;

	if (lpProc == NULL)
		lpProc = reinterpret_cast<DLGPROC>(DefaultDlg);

	if (!dwInitParam)
		result = DialogBox(hInst, resName, hWnd, lpProc);
	else
		result = DialogBoxParam(hInst, resName, hWnd, lpProc, dwInitParam);

	if (result == -1) {
		DWORD dwError = GetLastError();
		SString errormsg = "Failed to create Dialog box ";
		errormsg += SString(dwError).c_str();
		errormsg += ".";
		MessageBox(hWnd, errormsg.c_str(), appName, MB_OK);
	}

	return result;
}
#endif

const char propGlobalFileName[] = "SciTEGlobal.properties";
const char propFileName[] = "SciTE.properties";

class RecentFile {
public:
	char fileName[MAX_PATH];
	int lineNumber;
	RecentFile() {
		fileName[0] = '\0';
		lineNumber = -1;
	}
};

enum JobSubsystem { jobCLI = 0, jobGUI = 1};
class Job {
public:
	SString command;
	SString directory;
	JobSubsystem jobType;

	Job();
	void Clear();
};

Job::Job() {
	Clear();
}

void Job::Clear() {
	command = "";
	directory = "";
	jobType = jobCLI;
}


typedef EntryMemory < 10 > ComboMemory;

class SciTEWindow {
#ifdef PLAT_WIN
	static HINSTANCE hInstance;
	static char *className;
#endif
	char windowName[MAX_PATH + 20];
	char fullPath[MAX_PATH];
	char fileName[MAX_PATH];
	char fileExt[MAX_PATH];
	char dirName[MAX_PATH];
	char dirNameAtExecute[MAX_PATH];
	time_t fileModTime;

	enum { fileStackMax = 10 };
	RecentFile recentFileStack[fileStackMax];
	enum { fileStackCmdID = IDM_MRUFILE };

	char findWhat[200];
	char replaceWhat[200];
#ifdef PLAT_WIN
	FINDREPLACE fr;
	char openWhat[200];
	int filterDefault;
#endif
	Window wFindReplace;
	bool replacing;
	bool havefound;
	bool matchCase;
	bool wholeWord;
	bool reverseFind;
	ComboMemory memFinds;
	ComboMemory memReplaces;
	ComboMemory memFiles;

	int codePage;
	SString language;
	char **keywords;
	char *keywordlist;
	char **keywords2;
	char *keywordlist2;
	char **apis;
	char *apilist;
	SString functionDefinition;

	Window wSciTE;
	Window wEditor;
	Window wOutput;
#ifdef PLAT_GTK
	Window wDivider;
	Point ptOld;
	GdkGC *xor_gc;
//#define STATUSBAR
#ifdef STATUSBAR
	Window wStatusBar;
	guint sbContextID;
	bool sbVisible;
	SString sbValue;
#endif
#endif
	int dialogsOnScreen;
	enum { toolMax = 10 };

	int heightOutput;
	int heightOutputStartDrag;
	Point ptStartDrag;
	bool capturedMouse;
	bool firstPropertiesRead;
	bool splitVertical;
	bool bufferedDraw;
	bool bracesCheck;
	bool bracesSloppy;
	int bracesStyle;
	bool isSlaved;

	bool margin;
	int marginWidth;
	enum { marginWidthDefault = 20};

	bool lineNumbers;
	int lineNumbersWidth;
	enum { lineNumbersWidthDefault = 40};

	// On GTK+, heightBar should be 12 to be consistent with other apps but this is too big
	// for my tastes.
	enum { heightBar = 7};

	bool usePalette;
	bool clearBeforeExecute;
	bool allowMenuActions;
	bool isDirty;
	bool isBuilt;
	bool executing;
	enum { commandMax = 2 };
	int commandCurrent;
	Job jobQueue[commandMax];
	long cancelFlag;

#ifdef PLAT_GTK
	// Control of sub process
	int icmd;
	int originalEnd;
	int fdFIFO;
	int pidShell;
	char resultsFile[200];
	int inputHandle;
#endif

	PropSet propsEmbed;
	PropSet propsBase;
	PropSet props;

#ifdef PLAT_GTK
	Window content;
	bool savingHTML;
	Window fileSelector;
	Window findInFilesDialog;
	GtkWidget *comboFiles;
	Window gotoDialog;
	GtkWidget *gotoEntry;
	GtkWidget *toggleWord;
	GtkWidget *toggleCase;
	GtkWidget *toggleReverse;
	GtkWidget *comboFind;
	GtkWidget *comboReplace;
	GtkItemFactory *itemFactory;
	GtkAccelGroup *accelGroup;
#endif

	void ReadGlobalPropFile();
	void ReadLocalPropFile();

	LRESULT SendEditor(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0);
	LRESULT SendOutput(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0);
	void SendChildren(UINT msg, WPARAM wParam = 0, LPARAM lParam = 0);
	int LengthDocument();
	void GetRange(Window &win, int start, int end, char *text);
	void Colourise(int start = 0, int end = -1, bool editor = true);
	void FindMatchingBracePosition(bool editor, int &braceAtCaret, int &braceOpposite);
	void BraceMatch(bool editor);

	void SetWindowName();
	void SetFileName(const char *openName);
	void New();
	bool Exists(const char *dir, const char *path, char *testPath);
	void Open(const char *file = 0, bool initialCmdLine = false);
	int SaveIfUnsure(bool forceQuestion = false);
	void Save();
	void SaveAs(char *file = 0);
	void SaveToHTML(const char *saveName);
	void SaveAsHTML();
	void OpenProperties(bool local);
	void Print();
	void PrintSetup();
	void SetSelection(int anchor, int currentPos);
	void HandleFindReplace();
	void SelectionIntoProperties();
	void SelectionIntoFind();
	void Find();
	void FindNext();
	void FindInFiles();
	void Replace();
	void ReplaceOnce();
	void ReplaceAll();
	void GoLineDialog();
	void GoMatchingBrace();
#ifdef PLAT_GTK
	void FindReplaceGrabFields();
#endif
	void FindReplace(bool replace);
	void OutputAppendString(const char *s, int len = -1);
#ifdef PLAT_GTK
	void ContinueExecute();
#endif
	void Execute();
	void GoMessage(int dir);
	void StartCallTip();
	void ContinueCallTip();
	void StartAutoComplete();
	void CharAdded(char ch);
	int GetCurrentLineNumber();
	void AddCommand(const SString &cmd, const SString &dir, JobSubsystem jobType);
	void AboutDialog();
	void QuitProgram();
	void Command(WPARAM wParam, LPARAM lParam);
	void Notify(SCNotification *notification);

	void CheckAMenuItem(WORD wIDCheckItem, BOOL val);
	void EnableAMenuItem(WORD wIDCheckItem, BOOL val);
	void CheckMenus();
	void SizeSubWindows();

	void DestroyMenuItem(int menuNumber, int itemID);
	void SetMenuItem(int menuNumber, int position, int itemID, const char *text);
	void DeleteFileStackMenu();
	void SetFileStackMenu();
	void DropFileStackTop();
	void AddFileToStack(const char *file, int line = -1);
	void RememberLineNumberStack(const char *file, int line = -1);
	void StackMenu(int pos);

	void RemoveToolsMenu();
	void SetToolsMenu();
	JobSubsystem SubsystemType(const char *cmd, int item = -1);
	void ToolsMenu(int item);

	void ReadProperties();
	void SetOneStyle(Window &win, int style, const char *s);
	void SetStyleFor(Window &win, const char *language);

	void CheckReload();
	void Activate(bool activeApp);
	PRectangle GetClientRectangle();
	void Redraw();
	int NormaliseSplit(int splitPos);
	void MoveSplit(Point ptNewDrag);

#ifdef PLAT_GTK
	// GTK+ Signal Handlers

	static void OpenCancelSignal(GtkWidget *w, SciTEWindow *scitew);
	static void OpenKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew);
	static void OpenOKSignal(GtkWidget *w, SciTEWindow *scitew);
	static void SaveAsSignal(GtkWidget *w, SciTEWindow *scitew);

	static void FindInFilesSignal(GtkWidget *w, SciTEWindow *scitew);
	static void FindInFilesCancelSignal(GtkWidget *w, SciTEWindow *scitew);
	static void FindInFilesKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew);

	static void GotoCancelSignal(GtkWidget *w, SciTEWindow *scitew);
	static void GotoKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew);
	static void GotoSignal(GtkWidget *w, SciTEWindow *scitew);

	static void FRCancelSignal(GtkWidget *w, SciTEWindow *scitew);
	static void FRKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew);
	static void FRFindSignal(GtkWidget *w, SciTEWindow *scitew);
	static void FRReplaceSignal(GtkWidget *w, SciTEWindow *scitew);
	static void FRReplaceAllSignal(GtkWidget *w, SciTEWindow *scitew);

	static void IOSignal(SciTEWindow *scitew);

	static gint MoveResize(GtkWidget *widget, GtkAllocation *allocation, SciTEWindow *scitew);
	static void QuitSignal(GtkWidget *w, SciTEWindow *scitew);
	static void MenuSignal(SciTEWindow *scitew, guint action, GtkWidget *w);
	static void CommandSignal(GtkWidget *w, gint wParam, gpointer lParam, SciTEWindow *scitew);
	static void NotifySignal(GtkWidget *w, gint wParam, gpointer lParam, SciTEWindow *scitew);

	void DividerXOR(Point pt);
	static gint DividerExpose(GtkWidget *widget, GdkEventExpose *ose, SciTEWindow *scitew);
	static gint DividerMotion(GtkWidget *widget, GdkEventMotion *event, SciTEWindow *scitew);
	static gint DividerPress(GtkWidget *widget, GdkEventButton *event, SciTEWindow *scitew);
	static gint DividerRelease(GtkWidget *widget, GdkEventButton *event, SciTEWindow *scitew);
#endif

public:

	SciTEWindow();
	~SciTEWindow();

	void Run(const char *cmdLine);
#ifdef PLAT_WIN
	void Paint(Surface *surfaceWindow, PRectangle rcPaint);
	LRESULT WndProc(UINT iMessage, WPARAM wParam, LPARAM lParam);
	static BOOL CALLBACK GrepDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	static BOOL CALLBACK GoLineDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam);
	static void Register(HINSTANCE hInstance_);
	static LRESULT PASCAL TWndProc(
		    HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam);
#endif
	void ProcessExecute();
	WindowID GetID() { return wSciTE.GetID(); }
};

#ifdef PLAT_WIN
char *SciTEWindow::className = NULL;
HINSTANCE SciTEWindow::hInstance = 0;
#endif

SciTEWindow::SciTEWindow() {
	codePage = 0;
	language = "java";
	keywords = 0;
	keywordlist = 0;
	keywords2 = 0;
	keywordlist2 = 0;
	apis = 0;
	apilist = 0;
	functionDefinition = 0;

	dialogsOnScreen = 0;

	heightOutput = 0;

	allowMenuActions = true;
	isDirty = false;
	isBuilt = false;
	executing = false;
	commandCurrent = 0;
	InterlockedExchange(&cancelFlag, 0L);

#ifdef PLAT_GTK
	// Control of sub process
	icmd = 0;
	originalEnd = 0;
	fdFIFO = 0;
	pidShell = 0;
	sprintf(resultsFile, "/tmp/SciTE%x.results",
	        static_cast<int>(getpid()));
	inputHandle = 0;
#endif

	ptStartDrag.x = 0;
	ptStartDrag.y = 0;
	capturedMouse = false;
	firstPropertiesRead = true;
	splitVertical = false;
	bufferedDraw = true;
	bracesCheck = true;
	bracesSloppy = false;
	bracesStyle = 0;
	isSlaved = false;

	margin = false;
	marginWidth = marginWidthDefault;
	lineNumbers = false;
	lineNumbersWidth = lineNumbersWidthDefault;
	usePalette = false;

	clearBeforeExecute = false;
	findWhat[0] = '\0';
	replaceWhat[0] = '\0';
#ifdef PLAT_WIN
	memset(&fr, 0, sizeof(fr));
	strcpy(openWhat, "Custom Filter");
	openWhat[strlen(openWhat) + 1] = '\0';
	filterDefault = 1;
#endif
	replacing = false;
	havefound = false;
	matchCase = false;
	wholeWord = false;
	reverseFind = false;

	windowName[0] = '\0';
	fullPath[0] = '\0';
	fileName[0] = '\0';
	fileExt[0] = '\0';
	dirName[0] = '\0';
	dirNameAtExecute[0] = '\0';
	fileModTime = 0;

	propsBase.superPS = &propsEmbed;
	props.superPS = &propsBase;

#ifdef PLAT_WIN
	// Read properties resource into propsEmbed
	//The embedded properties file is to allow distributions to be sure
	// that they have a sensible default configuration even if the properties
	// files are missing. Also allows a one file distribution of Sc1.EXE.
	HRSRC handProps = ::FindResource(hInstance, "Embedded", "Properties");
	if (handProps) {
		DWORD size = ::SizeofResource(hInstance, handProps);
		HGLOBAL hmem = ::LoadResource(hInstance, handProps);
		if (hmem) {
			const void *pv = ::LockResource(hmem);
			if (pv) {
				propsEmbed.ReadFromMemory(
				    reinterpret_cast<const char *>(pv), size);
				UnlockResource(hmem);
			}
		}
		::FreeResource(handProps);
	}
#endif

	ReadGlobalPropFile();

#ifdef PLAT_GTK
	ptOld = Point(0, 0);
	xor_gc = 0;
	savingHTML = false;
	comboFiles = 0;
	gotoEntry = 0;
	toggleWord = 0;
	toggleCase = 0;
	toggleReverse = 0;
	comboFind = 0;
	comboReplace = 0;
	itemFactory = 0;
#else
	// Pass 'this' pointer in lpParam of CreateWindow().
	int left = props.GetInt("position.left", CW_USEDEFAULT);
	int top = props.GetInt("position.top", CW_USEDEFAULT);
	int width = props.GetInt("position.width", CW_USEDEFAULT);
	int height = props.GetInt("position.height", CW_USEDEFAULT);
	wSciTE = ::CreateWindowEx(
	             WS_EX_CLIENTEDGE,
	             className,
	             windowName,
	             WS_CAPTION | WS_SYSMENU | WS_THICKFRAME |
	             WS_MINIMIZEBOX | WS_MAXIMIZEBOX |
	             WS_MAXIMIZE | WS_CLIPCHILDREN,
	             left, top, width, height,
	             NULL,
	             NULL,
	             hInstance,
	             reinterpret_cast<LPSTR>(this));
	if (!wSciTE.Created())
		exit(FALSE);
#endif
}

#ifdef PLAT_WIN
void SciTEWindow::Register(HINSTANCE hInstance_) {
	const char resourceName[] = "SciTE";

	WNDCLASS wndclass;     // Structure used to register Windows class.

	className = "SciTEWindow";
	hInstance = hInstance_;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = SciTEWindow::TWndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = sizeof(SciTEWindow*);
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(hInstance, resourceName);
	wndclass.hCursor = NULL;
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = resourceName;
	wndclass.lpszClassName = className;

	if (!::RegisterClass(&wndclass))
		::exit(FALSE);
}

#endif

bool GetDefaultPropertiesFileName(char *pathDefaultProps, unsigned int lenPath) {
#ifdef PLAT_GTK
	strncpy(pathDefaultProps, getenv("HOME"), lenPath);
	strncat(pathDefaultProps, pathSepString, lenPath);
	strncat(pathDefaultProps, propGlobalFileName, lenPath);
	return true;
#else
	GetModuleFileName(0, pathDefaultProps, lenPath);
	char *lastSlash = strrchr(pathDefaultProps, pathSepChar);
	if (lastSlash && ((lastSlash + 1 - pathDefaultProps + strlen(propGlobalFileName)) < lenPath)) {
		strcpy(lastSlash + 1, propGlobalFileName);
		return true;
	} else {
		return false;
	}
#endif
}

void SciTEWindow::ReadGlobalPropFile() {
	char propfile[MAX_PATH + 20];
	if (GetDefaultPropertiesFileName(propfile, sizeof(propfile))) {
		propsBase.Read(propfile);
	}
}

void SciTEWindow::ReadLocalPropFile() {
	char propfile[MAX_PATH + 20];
	if (dirName[0])
		strcpy(propfile, dirName);
	else
		getcwd(propfile, sizeof(propfile));
	strcat(propfile, pathSepString);
	strcat(propfile, propFileName);
	props.Read(propfile);
	//Platform::DebugPrintf("Reading local properties from %s\n", propfile);

	// TODO: Grab these from Platform and update when environment says to
	props.Set("Chrome", "#C0C0C0");
	props.Set("ChromeHighlight", "#FFFFFF");
}

LRESULT SciTEWindow::SendEditor(UINT msg, WPARAM wParam, LPARAM lParam) {
	return SendMessage(wEditor.GetID(), msg, wParam, lParam);
}

LRESULT SciTEWindow::SendOutput(UINT msg, WPARAM wParam, LPARAM lParam) {
	return SendMessage(wOutput.GetID(), msg, wParam, lParam);
}

void SciTEWindow::SendChildren(UINT msg, WPARAM wParam, LPARAM lParam) {
	SendEditor(msg, wParam, lParam);
	SendOutput(msg, wParam, lParam);
}

void SciTEWindow::DestroyMenuItem(int menuNumber, int itemID) {
#ifdef PLAT_GTK
	// On GTK+ menu items are just hidden rather than destroyed as they can not be recreated in the middle of a menu
	// The menuNumber is ignored as all menu items in GTK+ can be found from the root of the menu tree
	GtkWidget *item = gtk_item_factory_get_widget_by_action(itemFactory, itemID);
	if (item) {
		//Platform::DebugPrintf("Destroying[%0d]\n", itemID);
		gtk_widget_hide(item);
	}
#else
	// On Windows menu items are destroyed as they can not be hidden and they can be recreated in any position
	HMENU hmenuBar = ::GetMenu(wSciTE.GetID());
	HMENU hmenu = ::GetSubMenu(hmenuBar, menuNumber);
	DeleteMenu(hmenu, itemID, MF_BYCOMMAND);
#endif
}

void SciTEWindow::SetMenuItem(int menuNumber, int position, int itemID, const char *text) {
#ifdef PLAT_GTK
	// On GTK+ the menuNumber and position are ignored as the menu item already exists and is in the right
	// place so only needs to be shown and have its text set.
	GtkWidget *item = gtk_item_factory_get_widget_by_action(itemFactory, itemID);
	if (item) {
		GList *al = gtk_container_children(GTK_CONTAINER(item));
		for (unsigned int ii = 0; ii < g_list_length(al); ii++) {
			gpointer d = g_list_nth(al, ii);
			GtkWidget **w = (GtkWidget **)d;
			gtk_label_set_text(GTK_LABEL(*w), text);
		}
		g_list_free(al);
		gtk_widget_show(item);
	}
#else
	// On Windows the menu items are modified if they already exist or created
	HMENU hmenuBar = ::GetMenu(wSciTE.GetID());
	HMENU hmenu = ::GetSubMenu(hmenuBar, menuNumber);
	if (::GetMenuState(hmenu, itemID, MF_BYCOMMAND) == 0xffffffff) {
		if (text[0])
			::InsertMenu(hmenu, position, MF_BYPOSITION, itemID, text);
		else
			::InsertMenu(hmenu, position, MF_BYPOSITION | MF_SEPARATOR, itemID, text);
	} else {
		::ModifyMenu(hmenu, position, MF_BYCOMMAND, itemID, text);
	}
#endif
}

void SciTEWindow::DeleteFileStackMenu() {
	for (int stackPos = 1; stackPos < fileStackMax; stackPos++) {
		DestroyMenuItem(0, fileStackCmdID + stackPos);
	}
	DestroyMenuItem(0, IDM_MRU_SEP);
}

void SciTEWindow::SetFileStackMenu() {
	int menuStart = 6;
	if (recentFileStack[1].fileName[0]) {
		SetMenuItem(0, menuStart, IDM_MRU_SEP, "");
		for (int stackPos = 1; stackPos < fileStackMax; stackPos++) {
			int itemID = fileStackCmdID + stackPos;
			if (recentFileStack[stackPos].fileName[0]) {
				SetMenuItem(0, menuStart + stackPos, itemID, recentFileStack[stackPos].fileName);
			}
		}
	}
}

void SciTEWindow::DropFileStackTop() {
	DeleteFileStackMenu();
	for (int stackPos = 0; stackPos < fileStackMax - 1; stackPos++)
		recentFileStack[stackPos] = recentFileStack[stackPos + 1];
	strcpy(recentFileStack[fileStackMax - 1].fileName, "");
	recentFileStack[fileStackMax - 1].lineNumber = -1;
	SetFileStackMenu();
}

void SciTEWindow::AddFileToStack(const char *file, int line) {
	DeleteFileStackMenu();
	if (file[strlen(file) - 1] == '\\')
		return;
	int stackPos;
	int eqPos = fileStackMax - 1;
	for (stackPos = 0; stackPos < fileStackMax; stackPos++)
		if (strcmp(recentFileStack[stackPos].fileName, file) == 0)
			eqPos = stackPos;
	for (stackPos = eqPos; stackPos > 0; stackPos--)
		recentFileStack[stackPos] = recentFileStack[stackPos - 1];
	strcpy(recentFileStack[0].fileName, file);
	recentFileStack[0].lineNumber = line;
	SetFileStackMenu();
}

void SciTEWindow::RememberLineNumberStack(const char *file, int line) {
	for (int stackPos = 0; stackPos < fileStackMax; stackPos++) {
		if (strcmp(recentFileStack[stackPos].fileName, file) == 0) {
			recentFileStack[stackPos].lineNumber = line;
		}
	}
}

void SciTEWindow::StackMenu(int pos) {
	//Platform::DebugPrintf("Stack menu %d\n", pos);
	if (pos == -1) {
		int lastPos = -1;
		for (int stackPos = 1; stackPos < fileStackMax; stackPos++) {
			if (recentFileStack[stackPos].fileName[0])
				lastPos = stackPos;
		}
		if (lastPos > 0) {
			//Platform::DebugPrintf("Opening lpos %d %s\n",recentFileStack[lastPos].lineNumber,recentFileStack[lastPos].fileName);
			int line = recentFileStack[lastPos].lineNumber;
			SString fileNameCopy = recentFileStack[lastPos].fileName;
			Open(fileNameCopy.c_str());
			if (line != -1)
				SendEditor(SCI_GOTOLINE, line);
		}
	} else {
		if ((pos == 0) && (recentFileStack[pos].fileName[0] == '\0')) {	// Empty
			New();
			ReadProperties();
		} else if (recentFileStack[pos].fileName[0] != '\0') {
			int line = recentFileStack[pos].lineNumber;
			//Platform::DebugPrintf("Opening pos %d %s\n",recentFileStack[pos].lineNumber,recentFileStack[pos].fileName);
			SString fileNameCopy = recentFileStack[pos].fileName;
			Open(fileNameCopy.c_str());
			if (line != -1)
				SendEditor(SCI_GOTOLINE, line);
		}
	}
}

void SciTEWindow::RemoveToolsMenu() {
	for (int pos = 0; pos < toolMax; pos++) {
		DestroyMenuItem(2, IDM_TOOLS + pos);
	}
}

void SciTEWindow::SetToolsMenu() {
	//command.name.0.*.py=Edit in PythonWin
	//command.0.*.py="c:\program files\python\pythonwin\pythonwin" /edit c:\coloreditor.py
	RemoveToolsMenu();
	const int menuPos = 3;
	for (int item = 0; item < toolMax; item++) {
		int itemID = IDM_TOOLS + item;
		SString prefix = "command.name.";
		prefix += SString(item).c_str();
		prefix += ".";
		SString commandName = props.GetNewExpand(prefix.c_str(), fileName);
		if (commandName.length()) {
			SString sMenuItem = commandName;
#ifdef PLAT_WIN
			sMenuItem += "\tCtrl+";
			sMenuItem += SString(item).c_str();
#endif
			SetMenuItem(2, menuPos + item, itemID, sMenuItem.c_str());
		}
	}
}

JobSubsystem SciTEWindow::SubsystemType(const char *cmd, int item) {
	JobSubsystem jobType = jobCLI;
	SString subsysprefix = cmd;
	if (item >= 0) {
		subsysprefix += SString(item).c_str();
		subsysprefix += ".";
	}
	SString subsystem = props.GetNewExpand(subsysprefix.c_str(), fileName);
	if (subsystem[0] == '1')
		jobType = jobGUI;
	return jobType;
}

void SciTEWindow::ToolsMenu(int item) {
	SelectionIntoProperties();

	SString prefix = "command.";
	prefix += SString(item).c_str();
	prefix += ".";
	SString command = props.GetNewExpand(prefix.c_str(), fileName);
	if (command.length()) {
		if (SaveIfUnsure() != IDCANCEL) {
			SString isfilter = "command.is.filter.";
			isfilter += SString(item).c_str();
			isfilter += ".";
			SString filter = props.GetNewExpand(isfilter.c_str(), fileName);
			if (filter[0] == '1')
				fileModTime -= 1;

			JobSubsystem jobType = SubsystemType("command.subsystem.", item);

			AddCommand(command, "", jobType);
			if (commandCurrent > 0)
				Execute();
		}
	}
}

static int IntFromHexDigit(const char ch) {
	if (isdigit(ch))
		return ch - '0';
	else if (ch >= 'A' && ch <= 'F')
		return ch - 'A' + 10;
	else if (ch >= 'a' && ch <= 'f')
		return ch - 'a' + 10;
	else
		return 0;
}

static Colour ColourFromString(const char *val) {
	int r = IntFromHexDigit(val[1]) * 16 + IntFromHexDigit(val[2]);
	int g = IntFromHexDigit(val[3]) * 16 + IntFromHexDigit(val[4]);
	int b = IntFromHexDigit(val[5]) * 16 + IntFromHexDigit(val[6]);
	return Colour(r, g, b);
}

void SciTEWindow::SetOneStyle(Window &win, int style, const char *s) {
	char *val = strdup(s);
	//Platform::DebugPrintf("Style %d is [%s]\n", style, val);
	char *opt = val;
	while (opt) {
		char *cpComma = strchr(opt, ',');
		if (cpComma)
			*cpComma = '\0';
		char *colon = strchr(opt, ':');
		if (colon)
			*colon++ = '\0';
		if (0 == strcmp(opt, "italics"))
			SendMessage(win.GetID(), SCI_STYLESETITALIC, style, 1);
		if (0 == strcmp(opt, "notitalics"))
			SendMessage(win.GetID(), SCI_STYLESETITALIC, style, 0);
		if (0 == strcmp(opt, "bold"))
			SendMessage(win.GetID(), SCI_STYLESETBOLD, style, 1);
		if (0 == strcmp(opt, "notbold"))
			SendMessage(win.GetID(), SCI_STYLESETBOLD, style, 0);
		if (0 == strcmp(opt, "font"))
			SendMessage(win.GetID(), SCI_STYLESETFONT, style, reinterpret_cast<LPARAM>(colon));
		if (0 == strcmp(opt, "fore"))
			SendMessage(win.GetID(), SCI_STYLESETFORE, style, ColourFromString(colon).AsLong());
		if (0 == strcmp(opt, "back"))
			SendMessage(win.GetID(), SCI_STYLESETBACK, style, ColourFromString(colon).AsLong());
		if (0 == strcmp(opt, "size"))
			SendMessage(win.GetID(), SCI_STYLESETSIZE, style, atoi(colon));
		if (0 == strcmp(opt, "eolfilled"))
			SendMessage(win.GetID(), SCI_STYLESETEOLFILLED, style, 1);
		if (0 == strcmp(opt, "noteolfilled"))
			SendMessage(win.GetID(), SCI_STYLESETEOLFILLED, style, 0);
		if (cpComma)
			opt = cpComma + 1;
		else
			opt = 0;
	}
	if (val)
		free(val);
}

void SciTEWindow::SetStyleFor(Window &win, const char *lang) {
	for (int style = 0; style <= STYLE_MAX; style++) {
		if (style != STYLE_DEFAULT) {
			char key[200];
			sprintf(key, "style.%s.%0d", lang, style);
			SString sval = props.GetNewExpand(key, "");
			SetOneStyle(win, style, sval.c_str());
		}
	}
}

static bool iswordsep(char ch, bool onlyLineEnds) {
	if (!isspace(ch))
		return false;
	if (!onlyLineEnds)
		return true;
	return ch == '\r' || ch == '\n';
}

// Creates an array that points into each word in the string and puts \0 terminators
// after each word.
char **ArrayFromWordList(char *wordlist, bool onlyLineEnds = false) {
	char prev = '\n';
	int words = 0;
	for (int j = 0; wordlist[j]; j++) {
		if (!iswordsep(wordlist[j], onlyLineEnds) && iswordsep(prev, onlyLineEnds))
			words++;
		prev = wordlist[j];
	}
	char **keywords = new char * [words + 1];
	if (keywords) {
		words = 0;
		prev = '\0';
		int len = strlen(wordlist);
		for (int k = 0; k < len; k++) {
			if (!iswordsep(wordlist[k], onlyLineEnds)) {
				if (!prev) {
					keywords[words] = &wordlist[k];
					words++;
				}
			} else {
				wordlist[k] = '\0';
			}
			prev = wordlist[k];
		}
		keywords[words] = &wordlist[len];
	}
	return keywords;
}

void SciTEWindow::ReadProperties() {
	SString fileNameForExtension = fileName;
	if (!fileNameForExtension.length())
		fileNameForExtension = props.Get("default.file.ext");

	language = props.GetNewExpand("lexer.", fileNameForExtension.c_str());

	if (keywords) {
		delete []keywords;
		free(keywordlist);
	}
	if (keywords2) {
		delete []keywords2;
		free(keywordlist2);
	}
	if (apis) {
		delete []apis;
		apis = 0;
		free(apilist);
		apilist = 0;
	}

	keywordlist = strdup(props.GetNewExpand("keywords.", fileNameForExtension.c_str()).c_str());
	keywords = ArrayFromWordList(keywordlist);

	keywordlist2 = strdup(props.GetNewExpand("keywords2.", fileNameForExtension.c_str()).c_str());
	keywords2 = ArrayFromWordList(keywordlist2);

	SString apifilename = props.GetNewExpand("api.", fileNameForExtension.c_str());
	if (apifilename.length()) {
		FILE *fp = fopen(apifilename.c_str(), "rb");
		if (fp) {
			fseek(fp, 0, SEEK_END);
			int len = ftell(fp);
			apilist = reinterpret_cast<char *>(malloc(len + 1));
			if (apilist) {
				fseek(fp, 0, SEEK_SET);
				fread(apilist, 1, len, fp);
				apilist[len] = '\0';
				apis = ArrayFromWordList(apilist, true);
			}
			fclose(fp);
			//Platform::DebugPrintf("Finished api file %d\n", len);
		}
	}

	codePage = props.GetInt("code.page");
	SendEditor(SCI_SETCODEPAGE, codePage);

	SString colour;
	colour = props.Get("caret.fore");
	if (colour.length()) {
		SendEditor(SCI_SETCARETFORE, ColourFromString(colour.c_str()).AsLong());
	}

	colour = props.Get("calltip.back");
	if (colour.length()) {
		SendEditor(SCI_CALLTIPSETBACK, ColourFromString(colour.c_str()).AsLong());
	}

	SString caretPeriod = props.Get("caret.period");
	if (caretPeriod.length()) {
		SendEditor(SCI_SETCARETPERIOD, caretPeriod.value());
		SendOutput(SCI_SETCARETPERIOD, caretPeriod.value());
	}

	SendEditor(SCI_SETEDGECOLUMN, props.GetInt("edge.column", 0));
	SendEditor(SCI_SETEDGEMODE, props.GetInt("edge.mode", EDGE_NONE));
	colour = props.Get("edge.colour");
	if (colour.length()) {
		SendEditor(SCI_SETEDGECOLOUR, ColourFromString(colour.c_str()).AsLong());
	}

	SString selfore = props.Get("selection.fore");
	if (selfore.length()) {
		SendChildren(SCI_SETSELFORE, 1, ColourFromString(selfore.c_str()).AsLong());
	} else {
		SendChildren(SCI_SETSELFORE, 0, 0);
	}
	colour = props.Get("selection.back");
	if (colour.length()) {
		SendChildren(SCI_SETSELBACK, 1, ColourFromString(colour.c_str()).AsLong());
	} else {
		if (selfore.length())
			SendChildren(SCI_SETSELBACK, 0, 0);
		else	// Have to show selection somehow
			SendChildren(SCI_SETSELBACK, 1, Colour(0xC0, 0xC0, 0xC0).AsLong());
	}

	char bracesStyleKey[200];
	sprintf(bracesStyleKey, "braces.%s.style", language.c_str());
	bracesStyle = props.GetInt(bracesStyleKey, 0);

	char key[200];
	SString sval;

	// Set styles
	// For each window set the global default style, then the language default style, then the other global styles, then the other language styles

	SendMessage(wEditor.GetID(), SCI_STYLERESETDEFAULT, 0, 0);
	SendMessage(wOutput.GetID(), SCI_STYLERESETDEFAULT, 0, 0);

	sprintf(key, "style.%s.%0d", "*", STYLE_DEFAULT);
	sval = props.GetNewExpand(key, "");
	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());
	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());

	sprintf(key, "style.%s.%0d", language.c_str(), STYLE_DEFAULT);
	sval = props.GetNewExpand(key, "");
	SetOneStyle(wEditor, STYLE_DEFAULT, sval.c_str());

	SendMessage(wEditor.GetID(), SCI_STYLECLEARALL, 0, 0);

	SetStyleFor(wEditor, "*");
	SetStyleFor(wEditor, language.c_str());

	SendMessage(wOutput.GetID(), SCI_STYLECLEARALL, 0, 0);

	sprintf(key, "style.%s.%0d", "errorlist", STYLE_DEFAULT);
	sval = props.GetNewExpand(key, "");
	SetOneStyle(wOutput, STYLE_DEFAULT, sval.c_str());

	SendMessage(wOutput.GetID(), SCI_STYLECLEARALL, 0, 0);

	SetStyleFor(wOutput, "*");
	SetStyleFor(wOutput, "errorlist");

	// As user can change vertical versus horizontal split, only read at app startup
	if (firstPropertiesRead) {
		splitVertical = props.GetInt("split.vertical");
		SendEditor(SCI_SETVIEWWS, props.GetInt("view.whitespace"));
		SendEditor(SCI_SETVIEWEOL, props.GetInt("view.eol"));

		lineNumbersWidth = 0;
		SString linenums = props.Get("line.numbers");
		if (linenums.length())
			lineNumbersWidth = linenums.value();
		lineNumbers = lineNumbersWidth;
		if (lineNumbersWidth == 0)
			lineNumbersWidth = lineNumbersWidthDefault;

		marginWidth = 0;
		SString margwidth = props.Get("margin.width");
		if (margwidth.length())
			marginWidth = margwidth.value();
		margin = marginWidth;
		if (marginWidth == 0)
			marginWidth = marginWidthDefault;
	}

	usePalette = props.GetInt("use.palette");
	SendEditor(SCI_SETUSEPALETTE, usePalette);

	clearBeforeExecute = props.GetInt("clear.before.execute");
	SendEditor(SCI_SETUSEPALETTE, usePalette);

	SendEditor(SCI_SETMARGINWIDTH, margin ? marginWidth : 0);

	SendEditor(SCI_SETLINENUMBERWIDTH, lineNumbers ? lineNumbersWidth : 0);

	bufferedDraw = props.GetInt("buffered.draw", 1);
	SendEditor(SCI_SETBUFFEREDDRAW, bufferedDraw);

	bracesCheck = props.GetInt("braces.check");
	bracesSloppy = props.GetInt("braces.sloppy");

	SString wordCharacters = props.GetNewExpand("word.characters.", fileNameForExtension.c_str());
	if (wordCharacters.length()) {
		SendEditor(SCI_SETWORDCHARS, 0, reinterpret_cast<LPARAM>(wordCharacters.c_str()));
	} else {
		SendEditor(SCI_SETWORDCHARS, 0, 0);
	}

	SendEditor(SCI_MARKERDELETEALL, -1);

	int tabSize = props.GetInt("tabsize");
	if (tabSize) {
		SendEditor(SCI_SETTABWIDTH, tabSize);
	}

	SetToolsMenu();

	firstPropertiesRead = false;
}

int SciTEWindow::LengthDocument() {
	return SendEditor(SCI_GETLENGTH);
}

void SciTEWindow::GetRange(Window &win, int start, int end, char *text) {
	TEXTRANGE tr;
	tr.chrg.cpMin = start;
	tr.chrg.cpMax = end;
	tr.lpstrText = text;
	SendMessage(win.GetID(), EM_GETTEXTRANGE, 0, reinterpret_cast<LPARAM>(&tr));
}

void SciTEWindow::Colourise(int start, int end, bool editor) {
	//DWORD dwStart = timeGetTime();
	Window &win = editor ? wEditor : wOutput;
	int lengthDoc = SendMessage(win.GetID(), SCI_GETLENGTH, 0, 0);
	if (end == -1)
		end = lengthDoc;
	int len = end - start;
	char *cdoc = new char[len + 1];  	// +1 for NUL
	//memset(cdoc, '\003', len + 1);
	//Platform::DebugPrintf("Colourise size=%d ptr=%x lan=%s\n", lengthDoc, cdoc, language);

	GetRange(win, start, end, cdoc);

	int styleStart = 0;
	if (start > 0)
		styleStart = SendMessage(win.GetID(), SCI_GETSTYLEAT, start - 1, 0);
	//Platform::DebugPrintf("Style start = %d %x .. %x %x %x (%0d)\n", styleStart,
	//	SendMessage(win.GetID(), SCI_GETCHARAT, start - 1, 0),
	//	cdoc[end-start-1], cdoc[end-start], cdoc[end-start+1], end-start+1
	//	);
	if (editor || isSlaved)
		ColouriseDoc(codePage, cdoc, start, len, styleStart, language.c_str(), keywords, keywords2, win.GetID());
	else
		ColouriseDoc(codePage, cdoc, start, len, 0, "errorlist", 0, 0, win.GetID());
	delete []cdoc;
	//DWORD dwEnd = timeGetTime();
	//Platform::DebugPrintf("end colourise %d\n", dwEnd - dwStart);
}

void SciTEWindow::FindMatchingBracePosition(bool editor, int &braceAtCaret, int &braceOpposite) {
	Window &win = editor ? wEditor : wOutput;
	int bracesStyleCheck = editor ? bracesStyle : 0;
	int caretPos = SendMessage(win.GetID(), SCI_GETCURRENTPOS, 0, 0);
	braceAtCaret = -1;
	braceOpposite = -1;
	char charBefore = '\0';
	char styleBefore = '\0';
	if (caretPos > 0) {
		charBefore = SendMessage(win.GetID(), SCI_GETCHARAT, caretPos - 1, 0);
		styleBefore = SendMessage(win.GetID(), SCI_GETSTYLEAT, caretPos - 1, 0) & 31;
	}
	// Priority goes to character before caret
	if (charBefore && strchr("[](){}", charBefore) && (styleBefore == bracesStyleCheck)) {
		braceAtCaret = caretPos - 1;
	}
	if (bracesSloppy && (braceAtCaret < 0)) {
		// No brace found so check other side
		char charAfter = SendMessage(win.GetID(), SCI_GETCHARAT, caretPos, 0);
		char styleAfter = SendMessage(win.GetID(), SCI_GETSTYLEAT, caretPos, 0) & 31;
		if (charAfter && strchr("[](){}", charAfter) && (styleAfter == bracesStyleCheck)) {
			braceAtCaret = caretPos;
		}
	}
	if (braceAtCaret >= 0) {
		braceOpposite =
		    SendMessage(win.GetID(), SCI_BRACEMATCH, braceAtCaret, 0);
	}
}

void SciTEWindow::BraceMatch(bool editor) {
	if (!bracesCheck)
		return;
	int braceAtCaret = -1;
	int braceOpposite = -1;
	FindMatchingBracePosition(editor, braceAtCaret, braceOpposite);
	Window &win = editor ? wEditor : wOutput;
	if ((braceAtCaret != -1) && (braceOpposite == -1))
		SendMessage(win.GetID(), SCI_BRACEBADLIGHT, braceAtCaret, 0);
	else
		SendMessage(win.GetID(), SCI_BRACEHIGHLIGHT, braceAtCaret, braceOpposite);
}

SciTEWindow::~SciTEWindow() {
	delete []keywords;
	keywords = 0;
	free(keywordlist);
	keywordlist = 0;
	delete []keywords2;
	keywords2 = 0;
	free(keywordlist2);
	keywordlist2 = 0;
	delete []apis;
	apis = 0;
	free(apilist);
	apilist = 0;
};

void SciTEWindow::SetWindowName() {
	if (fileName[0] == '\0')
		strcpy(windowName, "(Untitled)");
	else
		strcpy(windowName, fileName);
	if (isDirty)
		strcat(windowName, " * ");
	else
		strcat(windowName, " - ");
	strcat(windowName, appName);
	wSciTE.SetTitle(windowName);
}

void SciTEWindow::SetFileName(const char *openName) {
	char pathCopy[MAX_PATH + 1];
	pathCopy[0] = '\0';

	if (openName[0] == '\"') {
		strncpy(pathCopy, openName + 1, MAX_PATH);
		pathCopy[MAX_PATH] = '\0';
		if (pathCopy[strlen(pathCopy) - 1] == '\"')
			pathCopy[strlen(pathCopy) - 1] = '\0';
		_fullpath(fullPath, pathCopy, MAX_PATH);
	} else if (openName[0]) {
		_fullpath(fullPath, openName, MAX_PATH);
	} else {
		fullPath[0] = '\0';
	}

	char *cpDirEnd = strrchr(fullPath, pathSepChar);
	if (cpDirEnd) {
		strcpy(fileName, cpDirEnd + 1);
		strcpy(dirName, fullPath);
		dirName[cpDirEnd - fullPath] = '\0';
	} else {
		strcpy(fileName, fullPath);
		getcwd(dirName, sizeof(dirName));
		//Platform::DebugPrintf("Working directory: <%s>\n", dirName);
		strcpy(fullPath, dirName);
		strcat(fullPath, pathSepString);
		strcat(fullPath, fileName);
	}
#ifdef PLAT_WIN
	// On windows file comparison is done case insensitively so the user can
	// enter scite.cxx and still open this file, SciTE.cxx. To ensure that the file
	// is saved with correct capitalisation FindFirstFile is used to find out the
	// real name of the file.
	WIN32_FIND_DATA FindFileData;
	HANDLE hFind = FindFirstFile(fullPath, &FindFileData);
	if (hFind != INVALID_HANDLE_VALUE) {	// FindFirstFile found the file
		cpDirEnd = strrchr(fullPath, pathSepChar);
		if (cpDirEnd) {
			strcpy(fileName, FindFileData.cFileName);
			strcpy(dirName, fullPath);
			dirName[cpDirEnd - fullPath] = '\0';
			strcpy(fullPath, dirName);
			strcat(fullPath, pathSepString);
			strcat(fullPath, fileName);
		}
		FindClose(hFind);
	}
#endif
	char fileBase[MAX_PATH];
	strcpy(fileBase, fileName);
	char *cpExt = strrchr(fileBase, '.');
	if (cpExt) {
		*cpExt = '\0';
		strcpy(fileExt, cpExt + 1);
	} else { // No extension
		fileExt[0] = '\0';
	}

	chdir(dirName);

	ReadLocalPropFile();

	props.Set("FilePath", fullPath);
	props.Set("FileDir", dirName);
	props.Set("FileName", fileBase);
	props.Set("FileExt", fileExt);
	props.Set("FileNameExt", fileName);

	SetWindowName();
}

#ifdef PLAT_GTK
void SciTEWindow::OpenCancelSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->fileSelector.Destroy();
}

void SciTEWindow::OpenKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew) {
	if (event->keyval == GDK_Escape) {
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		scitew->fileSelector.Destroy();
	}
}

void SciTEWindow::OpenOKSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->Open(gtk_file_selection_get_filename(
	                 GTK_FILE_SELECTION(scitew->fileSelector.GetID())));
	scitew->fileSelector.Destroy();
}

void SciTEWindow::SaveAsSignal(GtkWidget *, SciTEWindow *scitew) {
	//Platform::DebugPrintf("Do Save As\n");
	if (scitew->savingHTML)
		scitew->SaveToHTML(gtk_file_selection_get_filename(
		                       GTK_FILE_SELECTION(scitew->fileSelector.GetID())));
	else
		scitew->SaveAs(gtk_file_selection_get_filename(
		                   GTK_FILE_SELECTION(scitew->fileSelector.GetID())));
	scitew->fileSelector.Destroy();
}
#endif

void SciTEWindow::New() {
	SendEditor(SCI_CLEARALL);
	fullPath[0] = '\0';
	fileName[0] = '\0';
	fileExt[0] = '\0';
	dirName[0] = '\0';
	SetWindowName();
	isDirty = false;
	isBuilt = false;
	SendEditor(SCI_SETSAVEPOINT);
}

bool SciTEWindow::Exists(const char *dir, const char *path, char *testPath) {
	if (dir) {
		strcpy(testPath, path);
	} else {
		if ((strlen(dir) + strlen(pathSepString) + strlen(path) + 1) > MAX_PATH)
			return false;
		strcpy(testPath, dir);
		strcat(testPath, pathSepString);
		strcat(testPath, path);
	}
	FILE *fp = fopen(testPath, "rb");
	if (!fp)
		return false;
	fclose(fp);
	return true;
}

static time_t GetModTime(const char *fullPath) {
	struct stat statusFile;
	if (stat(fullPath, &statusFile) != -1)
		return statusFile.st_mtime;
	else
		return 0;
}

const int blockSize = 131072;

void SciTEWindow::Open(const char *file, bool initialCmdLine) {
	//Platform::DebugPrintf("Opening\n");
	RememberLineNumberStack(fullPath, GetCurrentLineNumber());

	if (file) {
		New();
		//Platform::DebugPrintf("Opening %s\n", file);
		SetFileName(file);
		ReadProperties();
		if (fileName[0]) {
			SendEditor(SCI_CANCEL);
			SendEditor(SCI_SETUNDOCOLLECTION, 0);

			fileModTime = GetModTime(fullPath);

			FILE *fp = fopen(fullPath, "rb");
			if (fp || initialCmdLine) {
				if (fp) {
					AddFileToStack(fullPath);
					char data[blockSize];
					int lenFile = fread(data, 1, sizeof(data), fp);
					while (lenFile > 0) {
						SendEditor(SCI_ADDTEXT, lenFile,
						           reinterpret_cast<LPARAM>(static_cast<char *>(data)));
						lenFile = fread(data, 1, sizeof(data), fp);
					}
					fclose(fp);
				}

				SendEditor(SCI_SETUNDOCOLLECTION, 1);

				// Flick focus to the output window and back to
				// ensure palette realised correctly.
				SetFocus(wOutput.GetID());
				SetFocus(wEditor.GetID());
				SendEditor(EM_EMPTYUNDOBUFFER);
				SendEditor(SCI_SETSAVEPOINT);
			} else {
				char msg[MAX_PATH + 100];
				strcpy(msg, "Could not open file \"");
				strcat(msg, fullPath);
				strcat(msg, "\".");
				MessageBox(wSciTE.GetID(), msg, appName, MB_OK);
			}
			SendEditor(SCI_GOTOPOS, 0);
		}
		Redraw();
	} else {
#ifdef PLAT_GTK
		if (!fileSelector.Created()) {
			fileSelector = gtk_file_selection_new("Open File");
			gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->ok_button),
			                   "clicked", GtkSignalFunc(OpenOKSignal), this);
			gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->cancel_button),
			                   "clicked", GtkSignalFunc(OpenCancelSignal), this);
			gtk_signal_connect(GTK_OBJECT(fileSelector.GetID()),
			                   "key_press_event", GtkSignalFunc(OpenKeySignal),
			                   this);
			// Other ways to destroy
			// Mark it as a modal transient dialog
			gtk_window_set_modal(GTK_WINDOW(fileSelector.GetID()), TRUE);
			gtk_window_set_transient_for (GTK_WINDOW(fileSelector.GetID()),
			                              GTK_WINDOW(wSciTE.GetID()));
			fileSelector.Show();
		}
#else
		char openName[MAX_PATH] = "\0";
		OPENFILENAME ofn = {sizeof(OPENFILENAME)};
		ofn.hwndOwner = wSciTE.GetID();
		ofn.hInstance = hInstance;
		ofn.lpstrFile = openName;
		ofn.nMaxFile = sizeof(openName);
		char *filter = 0;
		SString openFilter = props.Get("open.filter");
		if (openFilter.length()) {
			filter = strdup(openFilter.c_str());
			for (int fc = 0; filter[fc]; fc++)
				if (filter[fc] == '|')
					filter[fc] = '\0';
		}
		ofn.lpstrFilter = filter;
		ofn.lpstrCustomFilter = openWhat;
		ofn.nMaxCustFilter = sizeof(openWhat);
		ofn.nFilterIndex = filterDefault;
		ofn.lpstrTitle = "Open File";
		ofn.Flags = OFN_HIDEREADONLY;

		if (::GetOpenFileName(&ofn)) {
			filterDefault = ofn.nFilterIndex;
			free(filter);
			//Platform::DebugPrintf("Open: <%s>\n", openName);
			Open(openName);
		} else {
			free(filter);
			return;
		}
#endif
	}
	AddFileToStack(fullPath);
	SetWindowName();
}

int SciTEWindow::SaveIfUnsure(bool forceQuestion) {
	if (isDirty) {
		if (props.GetInt("are.you.sure", 1) || forceQuestion) {
			char msg[MAX_PATH + 100];
			strcpy(msg, "Save changes to \"");
			strcat(msg, fullPath);
			strcat(msg, "\"?");
			dialogsOnScreen++;
			int decision = MessageBox(wSciTE.GetID(), msg, appName, MB_YESNOCANCEL);
			if (decision == IDYES) {
				Save();
			}
			dialogsOnScreen--;
			return decision;
		} else {
			Save();
		}
	}
	return IDYES;
}

void SciTEWindow::Save() {
	if (fileName[0]) {
		//Platform::DebugPrintf("Saving <%s>\n", fileName);
		//DWORD dwStart = timeGetTime();
		FILE *fp = fopen(fullPath, "wb");
		if (fp) {
			char data[blockSize + 1];
			AddFileToStack(fullPath, GetCurrentLineNumber());
			int lengthDoc = LengthDocument();
			for (int i = 0; i < lengthDoc; i += blockSize) {
				int grabSize = lengthDoc - i;
				if (grabSize > blockSize)
					grabSize = blockSize;
				GetRange(wEditor, i, i + grabSize, data);
				fwrite(data, grabSize, 1, fp);
			}
			fclose(fp);
			//DWORD dwEnd = timeGetTime();
			//Platform::DebugPrintf("Saved file=%d\n", dwEnd - dwStart);
			fileModTime = GetModTime(fullPath);
			SendEditor(SCI_SETSAVEPOINT);
			if ((EqualCaseInsensitive(fileName, propFileName)) ||
			        (EqualCaseInsensitive(fileName, propGlobalFileName))) {
				ReadGlobalPropFile();
				ReadLocalPropFile();
				ReadProperties();
				Redraw();
			}
		} else {
			char msg[200];
			strcpy(msg, "Could not save file \"");
			strcat(msg, fullPath);
			strcat(msg, "\".");
			dialogsOnScreen++;
			MessageBox(wSciTE.GetID(), msg, appName, MB_OK);
			dialogsOnScreen--;
		}
	} else {
		SaveAs();
	}
}

void SciTEWindow::SaveAs(char *file) {
	if (file && *file) {
		SetFileName(file);
		Save();
		ReadProperties();
		Colourise();  	// In case extension was changed
		Redraw();
		SetWindowName();
	} else {
#ifdef PLAT_GTK
		savingHTML = false;
		if (!fileSelector.Created()) {
			fileSelector = gtk_file_selection_new("Save File As");
			gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->ok_button),
			                   "clicked", GtkSignalFunc(SaveAsSignal), this);
			gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->cancel_button),
			                   "clicked", GtkSignalFunc(OpenCancelSignal), this);
			gtk_signal_connect(GTK_OBJECT(fileSelector.GetID()),
			                   "key_press_event", GtkSignalFunc(OpenKeySignal),
			                   this);
			// Other ways to destroy
			// Mark it as a modal transient dialog
			gtk_window_set_modal(GTK_WINDOW(fileSelector.GetID()), TRUE);
			gtk_window_set_transient_for (GTK_WINDOW(fileSelector.GetID()),
			                              GTK_WINDOW(wSciTE.GetID()));
			fileSelector.Show();
		}
#else
		if (0 == dialogsOnScreen) {
			char openName[MAX_PATH] = "\0";
			strcpy(openName, fileName);
			OPENFILENAME ofn = {sizeof(ofn)};
			ofn.hwndOwner = wSciTE.GetID();
			ofn.hInstance = hInstance;
			ofn.lpstrFile = openName;
			ofn.nMaxFile = sizeof(openName);
			ofn.lpstrTitle = "Save File";
			ofn.Flags = OFN_HIDEREADONLY;

			dialogsOnScreen++;
			if (::GetSaveFileName(&ofn)) {
				//Platform::DebugPrintf("Save: <%s>\n", openName);
				SetFileName(openName);
				Save();
				ReadProperties();
				Colourise();  	// In case extension was changed
				wEditor.InvalidateAll();
			}
			dialogsOnScreen--;
		}
#endif
	}
}

void SciTEWindow::SaveToHTML(const char *saveName) {
	Colourise();  	// Ensure whole file styled
	int tabSize = props.GetInt("tabsize");
	if (tabSize == 0)
		tabSize = 4;
	FILE *fp = fopen(saveName, "wt");
	if (fp) {
		int styleCurrent = -1;
		fputs("<HTML>\n", fp);
		fputs("<HEAD>\n", fp);
		fputs("<STYLE>\n", fp);
		SString colour;
		for (int istyle = 0; istyle <= STYLE_DEFAULT; istyle++) {
			char key[200];
			sprintf(key, "style.*.%0d", istyle);
			char *valdef = strdup(props.Get(key).c_str());
			sprintf(key, "style.%s.%0d", language.c_str(), istyle);
			char *val = strdup(props.Get(key).c_str());
			SString family;
			SString fore;
			SString back;
			bool italics = false;
			bool bold = false;
			int size = 0;
			if ((valdef && *valdef) || (val && *val)) {
				if (istyle == STYLE_DEFAULT) 
					fprintf(fp, "SPAN {\n");
				else
					fprintf(fp, ".S%0d {\n", istyle);
				if (valdef && *valdef) {
					char *opt = valdef;
					while (opt) {
						char *cpComma = strchr(opt, ',');
						if (cpComma)
							*cpComma = '\0';
						char *colon = strchr(opt, ':');
						if (colon)
							*colon++ = '\0';
						if (0 == strcmp(opt, "italics"))
							italics = true;
						if (0 == strcmp(opt, "notitalics"))
							italics = false;
						if (0 == strcmp(opt, "bold"))
							bold = true;
						if (0 == strcmp(opt, "notbold"))
							bold = false;
						if (0 == strcmp(opt, "font"))
							family = colon;
						if (0 == strcmp(opt, "fore"))
							fore = colon;
						if (0 == strcmp(opt, "back"))
							back = colon;
						if (0 == strcmp(opt, "size"))
							size = atoi(colon);
						if (cpComma)
							opt = cpComma + 1;
						else
							opt = 0;
					}
				}
				if (val && *val) {
					char *opt = val;
					while (opt) {
						char *cpComma = strchr(opt, ',');
						if (cpComma)
							*cpComma = '\0';
						char *colon = strchr(opt, ':');
						if (colon)
							*colon++ = '\0';
						if (0 == strcmp(opt, "italics"))
							italics = true;
						if (0 == strcmp(opt, "notitalics"))
							italics = false;
						if (0 == strcmp(opt, "bold"))
							bold = true;
						if (0 == strcmp(opt, "notbold"))
							bold = false;
						if (0 == strcmp(opt, "font"))
							family = colon;
						if (0 == strcmp(opt, "fore"))
							fore = colon;
						if (0 == strcmp(opt, "back"))
							back = colon;
						if (0 == strcmp(opt, "size"))
							size = atoi(colon);
						if (cpComma)
							opt = cpComma + 1;
						else
							opt = 0;
					}
				}
				if (italics)
					fprintf(fp, "\tfont-style: italic;\n");
				if (bold)
					fprintf(fp, "\tfont-weight: bold;\n");
				if (family.length())
					fprintf(fp, "\tfont-family: %s;\n", family.c_str());
				if (fore.length())
					fprintf(fp, "\tcolor: %s;\n", fore.c_str());
				if (back.length())
					fprintf(fp, "\tbackground: %s;\n", back.c_str());
				if (size)
					fprintf(fp, "\tfont-size: %0dpt;\n", size);
				fprintf(fp, "}\n");
			}
			if (val)
				free(val);
			if (valdef)
				free(valdef);
		}
		fputs("</STYLE>\n", fp);
		fputs("</HEAD>\n", fp);
		fputs("<BODY>\n", fp);
		fputs("<SPAN class=\'S0\'>", fp);
		int lengthDoc = LengthDocument();
		bool prevCR = false;
		for (int i = 0; i < lengthDoc; i++) {
			char ch = SendEditor(SCI_GETCHARAT, i);
			int style = SendEditor(SCI_GETSTYLEAT, i);
			if (style != styleCurrent) {
				fputs("</SPAN>", fp);
				fprintf(fp, "<SPAN class=\'S%0d\'>", style);
				styleCurrent = style;
			}
			if (ch == ' ') {
				fputs("&nbsp;", fp);
			} else if (ch == '\t') {
				for (int itab = 0; itab < tabSize; itab++)
					fputs("&nbsp;", fp);
			} else if (ch == '\r') {
				fputs("<BR>\n", fp);
			} else if (ch == '\n') {
				if (!prevCR)
					fputs("<BR>\n", fp);
			} else if (ch == '<') {
				fputs("&lt;", fp);
			} else if (ch == '>') {
				fputs("&gt;", fp);
			} else if (ch == '&') {
				fputs("&amp;", fp);
			} else {
				fputc(ch, fp);
			}
			prevCR = ch == '\r';
		}
		fputs("</SPAN>", fp);
		fputs("</BODY>\n</HTML>\n", fp);
		fclose(fp);
	} else {
		char msg[200];
		strcpy(msg, "Could not save file \"");
		strcat(msg, fullPath);
		strcat(msg, "\".");
		dialogsOnScreen++;
		MessageBox(wSciTE.GetID(), msg, appName, MB_OK);
		dialogsOnScreen--;
	}
}
void SciTEWindow::SaveAsHTML() {
#ifdef PLAT_GTK
	if (!fileSelector.Created()) {
		savingHTML = true;
		fileSelector = gtk_file_selection_new("Save File As HTML");
		gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->ok_button),
		                   "clicked", GtkSignalFunc(SaveAsSignal), this);
		gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(fileSelector.GetID())->cancel_button),
		                   "clicked", GtkSignalFunc(OpenCancelSignal), this);
		gtk_signal_connect(GTK_OBJECT(fileSelector.GetID()),
		                   "key_press_event", GtkSignalFunc(OpenKeySignal),
		                   this);
		// Other ways to destroy
		// Mark it as a modal transient dialog
		gtk_window_set_modal(GTK_WINDOW(fileSelector.GetID()), TRUE);
		gtk_window_set_transient_for (GTK_WINDOW(fileSelector.GetID()),
		                              GTK_WINDOW(wSciTE.GetID()));
		fileSelector.Show();
	}
#else
	if (0 == dialogsOnScreen) {
		char saveName[MAX_PATH] = "\0";
		strcpy(saveName, fileName);
		char *cpDot = strchr(saveName, '.');
		strcpy(cpDot, ".html");
		OPENFILENAME ofn = {sizeof(ofn)};
		ofn.hwndOwner = wSciTE.GetID();
		ofn.hInstance = hInstance;
		ofn.lpstrFile = saveName;
		ofn.nMaxFile = sizeof(saveName);
		ofn.lpstrTitle = "Save File As HTML";
		ofn.Flags = OFN_HIDEREADONLY;

		ofn.lpstrFilter = "Web (.html;.htm)\0*.html;*.htm\0";

		dialogsOnScreen++;
		if (::GetSaveFileName(&ofn)) {
			//Platform::DebugPrintf("Save As HTML: <%s>\n", saveName);
			SaveToHTML(saveName);
		}
		dialogsOnScreen--;
	}
#endif
}

void SciTEWindow::OpenProperties(bool local) {
	if (SaveIfUnsure() != IDCANCEL) {
		char propfile[MAX_PATH + 20];
		if (local) {
			getcwd(propfile, sizeof(propfile));
			strcat(propfile, pathSepString);
			strcat(propfile, propFileName);
			Open(propfile);
		} else {
			if (GetDefaultPropertiesFileName(propfile, sizeof(propfile))) {
				Open(propfile);
			}
		}
	}
}

// TODO: Use margins from Page Setup dialog.
void SciTEWindow::Print() {
#ifdef PLAT_WIN
#ifdef OLD_INI_WAY
	// This code is old style (before common dialogs) code and could be used for default printing without showing dialog
	char szDevice[1000];
	int length = GetProfileString("windows", "device", "", szDevice, sizeof(szDevice));
	if (length == 0) {
		MessageBox(wSciTE.GetID(), "No default printer found.", 0, MB_OK);
		return;
	}
	Platform::DebugPrintf("Print to %s\n", szDevice);

	PSTR spDeviceName = strtok(szDevice, ",");
	PSTR spFileName = strtok(NULL, ", ");
	PSTR spPort = strtok(NULL, ", ");

	//HDC hdc = ::CreateDC(spFileName, spDeviceName, spPort, 0);
	Platform::DebugPrintf("Print to %s|%s|%s\n", spDeviceName, spFileName, spPort);
	if (hdc == 0) {
		::MessageBox(wSciTE.GetID(), "Can not write to printer.", 0, MB_OK);
		return;
	}
#endif

	PRINTDLG pdlg = {sizeof(PRINTDLG)};
	pdlg.hwndOwner = wSciTE.GetID();
	pdlg.hInstance = hInstance;
	pdlg.Flags = PD_USEDEVMODECOPIES | PD_ALLPAGES |
	             PD_NOPAGENUMS | PD_NOSELECTION | PD_RETURNDC;
	pdlg.nFromPage = 1;
	pdlg.nToPage = 1;
	pdlg.nMinPage = 1;
	pdlg.nMaxPage = 1000;
	pdlg.nCopies = 1;
	pdlg.hDC = 0;
	if (!::PrintDlg(&pdlg)) {
		return;
	}
	if (pdlg.hDevMode)
		::GlobalFree(pdlg.hDevMode);
	if (pdlg.hDevNames != NULL)
		::GlobalFree(pdlg.hDevNames);


	HDC hdc = pdlg.hDC;

	PRectangle rc;
	rc.right = ::GetDeviceCaps(hdc, HORZRES);
	rc.bottom = ::GetDeviceCaps(hdc, VERTRES);
	// Set top to be 2% down page, otherwise frst line is half printed, should use Page Setup margins but
	// should not have to do a Page Setup. May need to set other margins.
	rc.top = rc.bottom / 50;
	DOCINFO di = {sizeof(DOCINFO)};
	di.lpszDocName = windowName;
	di.lpszOutput = 0;
	di.lpszDatatype = 0;
	di.fwType = 0;
	if (::StartDoc(hdc, &di) < 0) {
		MessageBox(wSciTE.GetID(), "Can not start printer document.", 0, MB_OK);
		return;
	}

	LONG lengthDoc = SendEditor(SCI_GETLENGTH);

	// Print each page
	LONG lengthPrinted = 0;
	while (lengthPrinted < lengthDoc) {
		//Platform::DebugPrintf("Printing a page %d\n", lengthPrinted);
		::StartPage(hdc);

		FORMATRANGE frPrint = {0};
		frPrint.hdc = hdc;
		frPrint.hdcTarget = hdc;
		frPrint.rc.left = rc.left;
		frPrint.rc.top = rc.top;
		frPrint.rc.right = rc.right;
		frPrint.rc.bottom = rc.bottom;
		frPrint.rcPage = frPrint.rc;
		frPrint.chrg.cpMin = lengthPrinted;
		frPrint.chrg.cpMax = lengthDoc;

		lengthPrinted = SendEditor(EM_FORMATRANGE, TRUE,
		                           reinterpret_cast<LPARAM>(&frPrint));
		::EndPage(hdc);
	};
	SendEditor(EM_FORMATRANGE, FALSE, 0);

	::EndDoc(hdc);
	::DeleteDC(hdc);

#endif
}

void SciTEWindow::PrintSetup() {
#ifdef PLAT_WIN
#ifdef PAGE_SETUP_IMPLEMENTED
	PAGESETUPDLG pdlg = {sizeof(PAGESETUPDLG)};
	pdlg.hwndOwner = wSciTE.GetID();
	pdlg.hInstance = hInstance;
	pdlg.Flags = 0;
	if (!PageSetupDlg(&pdlg))
		return;
	Platform::DebugPrintf("PrintSize %d %d\n",
	                      pdlg.ptPaperSize.x, pdlg.ptPaperSize.y);
	Platform::DebugPrintf("MinMargin %d %d %d %d\n",
	                      pdlg.rtMinMargin.left, pdlg.rtMinMargin.top,
	                      pdlg.rtMinMargin.right, pdlg.rtMinMargin.bottom);
	Platform::DebugPrintf("Margin %d %d %d %d\n",
	                      pdlg.rtMargin.left, pdlg.rtMargin.top,
	                      pdlg.rtMargin.right, pdlg.rtMargin.bottom);
#endif
#endif
}

void SciTEWindow::SetSelection(int anchor, int currentPos) {
	CHARRANGE crange;
	crange.cpMin = anchor;
	crange.cpMax = currentPos;
	SendEditor(EM_EXSETSEL, 0, reinterpret_cast<LPARAM>(&crange));
}

void SciTEWindow::HandleFindReplace() {
#ifdef PLAT_WIN
	matchCase = fr.Flags & FR_MATCHCASE;
	wholeWord = fr.Flags & FR_WHOLEWORD;
	reverseFind = !(fr.Flags & FR_DOWN);
	if (fr.Flags & FR_DIALOGTERM) {
		wFindReplace = 0;
		currentDlg = 0;
		//Platform::DebugPrintf("Cancel FindReplace\n");
	} else if (fr.Flags & FR_FINDNEXT) {
		//Platform::DebugPrintf("FindNext <%s> %x\n", fr.lpstrFindWhat, fr.Flags);
		FindNext();
	} else if (fr.Flags & FR_REPLACE) {
		if (havefound) {
			SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(fr.lpstrReplaceWith));
			havefound = false;
			//Platform::DebugPrintf("Replace <%s> -> <%s>\n", fr.lpstrFindWhat, fr.lpstrReplaceWith);
		}
		FindNext();
	} else if (fr.Flags & FR_REPLACEALL) {
		FINDTEXTEX ft = {0};
		ft.chrg.cpMin = 0;
		ft.chrg.cpMax = LengthDocument();
		ft.lpstrText = fr.lpstrFindWhat;
		ft.chrgText.cpMin = 0;
		ft.chrgText.cpMax = 0;
		int posFind = SendEditor(EM_FINDTEXTEX, fr.Flags & (FR_WHOLEWORD | FR_MATCHCASE),
		                         reinterpret_cast<LPARAM>(&ft));
		if (posFind != -1) {
			SendEditor(SCI_BEGINUNDOACTION);
			while (posFind != -1) {
				SetSelection(ft.chrgText.cpMin, ft.chrgText.cpMax);
				SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(fr.lpstrReplaceWith));
				if (fr.Flags & FR_DOWN) {
					ft.chrg.cpMin = posFind + strlen(fr.lpstrReplaceWith) + 1;
					ft.chrg.cpMax = LengthDocument();
				} else {
					ft.chrg.cpMin = 0;
					ft.chrg.cpMax = posFind;
				}
				posFind = SendEditor(EM_FINDTEXTEX, fr.Flags & (FR_WHOLEWORD | FR_MATCHCASE),
				                     reinterpret_cast<LPARAM>(&ft));
			}
			SendEditor(SCI_ENDUNDOACTION);
		} else {
			char msg[200];
			strcpy(msg, "No replacements because string \"");
			strncat(msg, fr.lpstrFindWhat, sizeof(msg) / 2);
			strcat(msg, "\" was not present.");
			MessageBox(wFindReplace.GetID(), msg, appName, MB_OK | MB_ICONWARNING);
		}
		//Platform::DebugPrintf("ReplaceAll <%s> -> <%s>\n", fr.lpstrFindWhat, fr.lpstrReplaceWith);
	} else {
		//Platform::DebugPrintf("Find/replacemessage %x\n", fr.Flags);
	}
#endif
}

static bool iswordchar(char ch) {
	return isalnum(ch) || ch == '_';
}

void SciTEWindow::SelectionIntoProperties() {
	int selStart = 0;
	int selEnd = 0;
	SendEditor(EM_GETSEL, reinterpret_cast<WPARAM>(&selStart),
	           reinterpret_cast<LPARAM>(&selEnd));
	char currentSelection[1000];
	if ((selStart < selEnd) && ((selEnd - selStart + 1) < static_cast<int>(sizeof(currentSelection)))) {
		GetRange(wEditor, selStart, selEnd, currentSelection);
		props.Set("CurrentSelection", currentSelection);
	}
}

void SciTEWindow::SelectionIntoFind() {
	int selStart = 0;
	int selEnd = 0;
	int lengthDoc = LengthDocument();
	SendEditor(EM_GETSEL, reinterpret_cast<WPARAM>(&selStart),
	           reinterpret_cast<LPARAM>(&selEnd));
	if (selStart == selEnd) {
		// Try and find a word at the caret
		if (iswordchar(SendEditor(SCI_GETCHARAT, selStart))) {
			while ((selStart > 0) && (iswordchar(SendEditor(SCI_GETCHARAT, selStart - 1))))
				selStart--;
			while ((selEnd < lengthDoc - 1) && (iswordchar(SendEditor(SCI_GETCHARAT, selEnd + 1))))
				selEnd++;
			if (selStart < selEnd)
				selEnd++;  	// Because normal selections end one past
		}
	}
	if ((selStart < selEnd) && ((selEnd - selStart + 1) < static_cast<int>(sizeof(findWhat)))) {
		GetRange(wEditor, selStart, selEnd, findWhat);
	}
}

void SciTEWindow::Find() {
#ifdef PLAT_GTK
	SelectionIntoFind();
	FindReplace(false);
#else
	if (wFindReplace.Created())
		return;
	SelectionIntoFind();

	memset(&fr, 0, sizeof(fr));
	fr.lStructSize = sizeof(fr);
	fr.hwndOwner = wSciTE.GetID();
	fr.hInstance = hInstance;
	fr.Flags = 0;
	if (!reverseFind)
		fr.Flags |= FR_DOWN;
	fr.lpstrFindWhat = findWhat;
	fr.wFindWhatLen = sizeof(findWhat);
	wFindReplace = ::FindText(&fr);
	replacing = false;
	currentDlg = wFindReplace.GetID();
#endif
}

void SciTEWindow::FindNext() {
	if (!findWhat[0]) {
		Find();
		return;
	}
	FINDTEXTEX ft = {0};
	if (reverseFind) {
		ft.chrg.cpMin = SendEditor(SCI_GETANCHOR) - 1;
		ft.chrg.cpMax = 0;
	} else {
		ft.chrg.cpMin = SendEditor(SCI_GETCURRENTPOS) + 1;
		ft.chrg.cpMax = LengthDocument();
	}
	ft.lpstrText = findWhat;
	ft.chrgText.cpMin = 0;
	ft.chrgText.cpMax = 0;
	int flags = (wholeWord ? FR_WHOLEWORD : 0) | (matchCase ? FR_MATCHCASE : 0);
	//DWORD dwStart = timeGetTime();
	int posFind = SendEditor(EM_FINDTEXTEX, flags, reinterpret_cast<LPARAM>(&ft));
	//DWORD dwEnd = timeGetTime();
	//Platform::DebugPrintf("<%s> found at %d took %d\n", findWhat, posFind, dwEnd - dwStart);
	if (posFind == -1) {
		// Failed to find in indicated direction so search on other side
		if (reverseFind) {
			ft.chrg.cpMin = LengthDocument();
			ft.chrg.cpMax = 0;
		} else {
			ft.chrg.cpMin = 0;
			ft.chrg.cpMax = LengthDocument();
		}
		posFind = SendEditor(EM_FINDTEXTEX, flags, reinterpret_cast<LPARAM>(&ft));
	}
	if (posFind == -1) {
		havefound = false;
		char msg[300];
		strcpy(msg, "Cannot find the string \"");
		strcat(msg, findWhat);
		strcat(msg, "\".");
		dialogsOnScreen++;
		if (wFindReplace.Created())
			MessageBox(wFindReplace.GetID(), msg, appName, MB_OK | MB_ICONWARNING);
		else
			MessageBox(wSciTE.GetID(), msg, appName, MB_OK);
		dialogsOnScreen--;
	} else {
		havefound = true;
		SetSelection(ft.chrgText.cpMin, ft.chrgText.cpMax);
		if (!replacing) {
#ifdef PLAT_GTK
			wFindReplace.Destroy();
#else
			if (wFindReplace.Created()) {
				EndDialog(wFindReplace.GetID(), IDCANCEL);
				wFindReplace = 0;
				currentDlg = 0;
			}
#endif
		}
	}
}

#ifdef PLAT_GTK
static void FillComboFromMemory(GtkWidget *combo, const ComboMemory &mem) {
	GtkWidget *list = GTK_COMBO(combo)->list;
	for (int i = 0; i < mem.Length(); i++) {
		GtkWidget *item = gtk_list_item_new_with_label(mem.At(i).c_str());
		gtk_container_add(GTK_CONTAINER(list), item);
		gtk_widget_show(item);
	}
}
#else
static void FillComboFromMemory(HWND combo, const ComboMemory &mem) {
	for (int i = 0; i < mem.Length(); i++) {
		//Platform::DebugPrintf("Combo[%0d] = %s\n", i, mem.At(i).c_str());
		SendMessage(combo, CB_ADDSTRING, 0,
		            reinterpret_cast<LPARAM>(mem.At(i).c_str()));
	}
}
#endif

#ifdef PLAT_GTK
void SciTEWindow::FindInFilesSignal(GtkWidget *, SciTEWindow *scitew) {
	char *findEntry = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(scitew->comboFind)->entry));
	scitew->props.Set("find.what", findEntry);
	scitew->memFinds.Insert(findEntry);
	char *filesEntry = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(scitew->comboFiles)->entry));
	scitew->props.Set("find.files", filesEntry);
	scitew->memFiles.Insert(filesEntry);

	scitew->findInFilesDialog.Destroy();

	//printf("Grepping for <%s> in <%s>\n",
	//	scitew->props.Get("find.what"),
	//	scitew->props.Get("find.files"));
	scitew->SelectionIntoProperties();
	scitew->AddCommand(scitew->props.GetNewExpand("find.command", ""), "", jobCLI);
	if (scitew->commandCurrent > 0)
		scitew->Execute();
}

void SciTEWindow::FindInFilesCancelSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->findInFilesDialog.Destroy();
}

void SciTEWindow::FindInFilesKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew) {
	if (event->keyval == GDK_Escape) {
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		scitew->findInFilesDialog.Destroy();
	}
}

#else
BOOL CALLBACK SciTEWindow::GrepDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	static SciTEWindow *sci;
	HWND hFindWhat;
	HWND hFiles;

	switch (message) {

	case WM_INITDIALOG:
		sci = reinterpret_cast<SciTEWindow *>(lParam);
		SetDlgItemText(hDlg, IDFINDWHAT, sci->props.Get("find.what").c_str());
		SetDlgItemText(hDlg, IDFILES, sci->props.Get("find.files").c_str());
		hFindWhat = GetDlgItem(hDlg, IDFINDWHAT);
		FillComboFromMemory(hFindWhat, sci->memFinds);
		hFiles = GetDlgItem(hDlg, IDFILES);
		FillComboFromMemory(hFiles, sci->memFiles);
		//SetDlgItemText(hDlg, IDDIRECTORY, props->Get("find.directory"));
		return TRUE;

	case WM_CLOSE:
		SendMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
		break;

	case WM_COMMAND:
		if (ControlIDOfCommand(wParam) == IDCANCEL) {
			EndDialog(hDlg, IDCANCEL);
			return FALSE;
		} else if (ControlIDOfCommand(wParam) == IDOK) {
			//Platform::DebugPrintf("Finding\n");
			char s[200];
			GetDlgItemText(hDlg, IDFINDWHAT, s, sizeof(s));
			sci->props.Set("find.what", s);
			strcpy(sci->findWhat, s);
			sci->memFinds.Insert(s);
			GetDlgItemText(hDlg, IDFILES, s, sizeof(s));
			sci->props.Set("find.files", s);
			sci->memFiles.Insert(s);
			//GetDlgItemText(hDlg, IDDIRECTORY, s, sizeof(s));
			//props->Set("find.directory", s);
			EndDialog(hDlg, IDOK);
			return TRUE;
		}
	}

	return FALSE;
}
#endif

void SciTEWindow::FindInFiles() {
	SelectionIntoFind();
	props.Set("find.what", findWhat);
	props.Set("find.directory", ".");
#ifdef PLAT_GTK
	findInFilesDialog = gtk_dialog_new();
	gtk_window_set_policy(GTK_WINDOW(findInFilesDialog.GetID()), TRUE, TRUE, TRUE);
	findInFilesDialog.SetTitle("Find In Files");

	gtk_signal_connect(GTK_OBJECT(findInFilesDialog.GetID()),
	                   "destroy", GtkSignalFunc(destroyDialog), 0);

	GtkWidget *table = gtk_table_new(2, 2, FALSE);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog.GetID())->vbox),
	                   table, TRUE, TRUE, 0);

	GtkAttachOptions opts = static_cast<GtkAttachOptions>(
	                            GTK_SHRINK | GTK_FILL);

	GtkAttachOptions optse = static_cast<GtkAttachOptions>(
	                             GTK_SHRINK | GTK_FILL | GTK_EXPAND);

	int row = 0;

	GtkWidget *labelFind = gtk_label_new("Find what:");
	gtk_table_attach(GTK_TABLE(table), labelFind, 0, 1,
	                 row, row + 1, opts, opts, 5, 5);
	gtk_widget_show(labelFind);

	comboFind = gtk_combo_new();
	FillComboFromMemory(comboFind, memFinds);

	gtk_table_attach(GTK_TABLE(table), comboFind, 1, 2,
	                 row, row + 1, optse, opts, 5, 5);
	gtk_widget_show(comboFind);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(comboFind)->entry), findWhat);
	gtk_entry_select_region(GTK_ENTRY(GTK_COMBO(comboFind)->entry), 0, strlen(findWhat));
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(comboFind)->entry),
	                   "activate", GtkSignalFunc(FindInFilesSignal), this);

	row++;

	GtkWidget *labelFiles = gtk_label_new("Files:");
	gtk_table_attach(GTK_TABLE(table), labelFiles, 0, 1,
	                 row, row + 1, opts, opts, 5, 5);
	gtk_widget_show(labelFiles);

	comboFiles = gtk_combo_new();
	FillComboFromMemory(comboFiles, memFiles);

	gtk_table_attach(GTK_TABLE(table), comboFiles, 1, 2,
	                 row, row + 1, optse, opts, 5, 5);
	gtk_widget_show(comboFiles);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(comboFiles)->entry), props.Get("find.files").c_str());
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(comboFiles)->entry),
	                   "activate", GtkSignalFunc(FindInFilesSignal), this);

	gtk_widget_show(table);

	GtkWidget *buttonFind = gtk_button_new_with_label("Find");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog.GetID())->action_area),
	                   buttonFind, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(buttonFind),
	                   "clicked", GtkSignalFunc(FindInFilesSignal), this);
	gtk_widget_show(buttonFind);

	GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(findInFilesDialog.GetID())->action_area),
	                   buttonCancel, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(buttonCancel),
	                   "clicked", GtkSignalFunc(FindInFilesCancelSignal), this);
	gtk_signal_connect(GTK_OBJECT(findInFilesDialog.GetID()),
	                   "key_press_event", GtkSignalFunc(FindInFilesKeySignal),
	                   this);

	gtk_widget_show(buttonCancel);

	GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonFind), GTK_CAN_DEFAULT);
	gtk_widget_grab_default(GTK_WIDGET(buttonFind));
	gtk_widget_grab_focus(GTK_WIDGET(GTK_COMBO(comboFind)->entry));

	// Mark it as a modal transient dialog
	gtk_window_set_modal(GTK_WINDOW(findInFilesDialog.GetID()), TRUE);
	gtk_window_set_transient_for (GTK_WINDOW(findInFilesDialog.GetID()),
	                              GTK_WINDOW(wSciTE.GetID()));

	findInFilesDialog.Show();
#else
	if (DoDialog(hInstance, "Grep", wSciTE.GetID(),
	             reinterpret_cast<DLGPROC>(GrepDlg),
	             reinterpret_cast<DWORD>(this)) == IDOK) {
		//Platform::DebugPrintf("asked to find %s %s %s\n", props.Get("find.what"), props.Get("find.files"), props.Get("find.directory"));
		SelectionIntoProperties();
		AddCommand(props.GetNewExpand("find.command", ""), "", jobCLI);
		if (commandCurrent > 0)
			Execute();
	}
#endif
}

void SciTEWindow::Replace() {
#ifdef PLAT_GTK
	SelectionIntoFind();
	FindReplace(true);
#else
	if (wFindReplace.Created())
		return;
	SelectionIntoFind();

	memset(&fr, 0, sizeof(fr));
	fr.lStructSize = sizeof(fr);
	fr.hwndOwner = wSciTE.GetID();
	fr.hInstance = hInstance;
	fr.Flags = FR_REPLACE;
	fr.lpstrFindWhat = findWhat;
	fr.lpstrReplaceWith = replaceWhat;
	fr.wFindWhatLen = sizeof(findWhat);
	fr.wReplaceWithLen = sizeof(replaceWhat);
	wFindReplace = ReplaceText(&fr);
	replacing = true;
	havefound = false;
	currentDlg = wFindReplace.GetID();
#endif
}

void SciTEWindow::ReplaceOnce() {
	if (havefound) {
		SendEditor(EM_REPLACESEL, 0,
		           reinterpret_cast<LPARAM>(static_cast<char *>(replaceWhat)));
		havefound = false;
		//Platform::DebugPrintf("Replace <%s> -> <%s>\n", findWhat, replaceWhat);
	}
	FindNext();
}

void SciTEWindow::ReplaceAll() {
	FINDTEXTEX ft;
	ft.chrg.cpMin = 0;
	ft.chrg.cpMax = LengthDocument();
	ft.lpstrText = findWhat;
	ft.chrgText.cpMin = 0;
	ft.chrgText.cpMax = 0;
	int flags = (wholeWord ? FR_WHOLEWORD : 0) | (matchCase ? FR_MATCHCASE : 0);
	int posFind = SendEditor(EM_FINDTEXTEX, flags,
	                         reinterpret_cast<LPARAM>(&ft));
	if (posFind != -1) {
		while (posFind != -1) {
			SetSelection(ft.chrgText.cpMin, ft.chrgText.cpMax);
			SendEditor(EM_REPLACESEL, 0,
			           reinterpret_cast<LPARAM>(static_cast<char *>(replaceWhat)));
			ft.chrg.cpMin = posFind + strlen(replaceWhat) + 1;
			ft.chrg.cpMax = LengthDocument();
			posFind = SendEditor(EM_FINDTEXTEX, flags,
			                     reinterpret_cast<LPARAM>(&ft));
		}
	} else {
		char msg[200];
		strcpy(msg, "No replacements because string \"");
		strcat(msg, findWhat);
		strcat(msg, "\" was not present.");
		dialogsOnScreen++;
#ifdef PLAT_GTK
		MessageBox(wSciTE.GetID(), msg, appName, MB_OK);
#else
		MessageBox(wFindReplace.GetID(), msg, appName, MB_OK);
#endif
		dialogsOnScreen--;
	}
	//Platform::DebugPrintf("ReplaceAll <%s> -> <%s>\n", findWhat, replaceWhat);
}

void SciTEWindow::OutputAppendString(const char *s, int len) {
	if (len == -1)
		len = strlen(s);
	SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
	SendOutput(SCI_ADDTEXT, len, reinterpret_cast<LPARAM>(s));
	SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
}

#ifdef PLAT_GTK
void SciTEWindow::ContinueExecute() {
	char buf[256];
	int count = 0;
	count = read(fdFIFO, buf, sizeof(buf) - 1 );
	if (count < 0) {
		OutputAppendString(">End Bad\n");
		return;
	}
	if (count == 0) {
		int status;
		int pidWaited = wait(&status);
		char exitmessage[80];
		sprintf(exitmessage, ">Exit code: %d\n", status);
		OutputAppendString(exitmessage);
		// Move selection back to beginning of this run so that F4 will go
		// to first error of this run.
		SendOutput(SCI_GOTOPOS, originalEnd);
		gdk_input_remove(inputHandle);
		close(fdFIFO);
		unlink(resultsFile);
		icmd++;
		if (icmd < commandCurrent && icmd < commandMax) {
			Execute();
		} else {
			icmd = 0;
			executing = false;
			CheckMenus();
			for (int ic = 0; ic < commandMax; ic++) {
				jobQueue[ic].Clear();
			}
			commandCurrent = 0;
		}
		return;
	}
	if (count > 0) {
		buf[count] = '\0';
		OutputAppendString(buf);
		if (count < 0) {
			OutputAppendString(">Continue no data\n");
			return;
		}
	}
}

void SciTEWindow::IOSignal(SciTEWindow *scitew) {
	scitew->ContinueExecute();
}
#else
// ProcessExecute runs a command with redirected input and output streams
// so the output can be put in a window.
// It is based upon several usenet posts and a knowledge base article.
void SciTEWindow::ProcessExecute() {

	DWORD exitcode = 0;

	SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
	int originalEnd = SendOutput(SCI_GETCURRENTPOS);

	for (int icmd = 0; icmd < commandCurrent && icmd < commandMax && exitcode == 0; icmd++) {

		OSVERSIONINFO osv = {sizeof(OSVERSIONINFO)};
		GetVersionEx(&osv);
		bool windows95 = osv.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS;

		SECURITY_ATTRIBUTES sa = {sizeof(SECURITY_ATTRIBUTES)};
		PROCESS_INFORMATION pi = {0};
		HANDLE hPipeWrite = NULL;
		HANDLE hPipeRead = NULL;
		HANDLE hWrite2 = NULL;
		HANDLE hRead2 = NULL;
		char buffer[256];
		//Platform::DebugPrintf("Execute <%s>\n", command);
		OutputAppendString(">");
		OutputAppendString(jobQueue[icmd].command.c_str());
		OutputAppendString("\n");

		sa.bInheritHandle = TRUE;
		sa.lpSecurityDescriptor = NULL;

		SECURITY_DESCRIPTOR sd;
		// If NT make a real security thing to allow inheriting handles
		if (!windows95) {
			InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
			SetSecurityDescriptorDacl(&sd, TRUE, NULL, FALSE);
			sa.nLength = sizeof(SECURITY_ATTRIBUTES);
			sa.lpSecurityDescriptor = &sd;
		}

		// Create pipe for output redirection
		CreatePipe(&hPipeRead,    // read handle
		           &hPipeWrite,   // write handle
		           &sa,        // security attributes
		           0      // number of bytes reserved for pipe - 0 default
		          );

		//Platform::DebugPrintf("2Execute <%s>\n");
		// Create pipe for input redirection. In this code, you do not
		// redirect the output of the child process, but you need a handle
		// to set the hStdInput field in the STARTUP_INFO struct. For safety,
		// you should not set the handles to an invalid handle.

		CreatePipe(&hRead2,     // read handle
		           &hWrite2,      // write handle
		           &sa,        // security attributes
		           0      // number of bytes reserved for pipe - 0 default
		          );

		//Platform::DebugPrintf("3Execute <%s>\n");
		// Make child process use hPipeWrite as standard out, and make
		// sure it does not show on screen.
		STARTUPINFO si = {sizeof(STARTUPINFO)};
		si.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
		if (jobQueue[icmd].jobType == jobCLI)
			si.wShowWindow = SW_HIDE;
		else
			si.wShowWindow = SW_SHOW;
		si.hStdInput = hWrite2;
		si.hStdOutput = hPipeWrite;
		si.hStdError = hPipeWrite;

		bool worked = CreateProcess(
		                  NULL,
		                  const_cast<char *>(jobQueue[icmd].command.c_str()),
		                  NULL, NULL,
		                  TRUE, 0,
		                  NULL, NULL,
		                  &si, &pi);

		if (!worked) {
			OutputAppendString(">Failed to CreateProcess\n");
		}

		/* now that this has been inherited, close it to be safe.
		You don't want to write to it accidentally */
		CloseHandle(hPipeWrite);

		bool completed = !worked;
		DWORD timeDetectedDeath = 0;
		while (!completed) {

			Sleep(100L);

			if (cancelFlag) {
				TerminateProcess(pi.hProcess, 1);
				completed = true;
				break;
			}

			bool NTOrData = true;

			DWORD bytesRead = 0;

			if (windows95) {
				DWORD bytesAvail = 0;
				if (PeekNamedPipe(hPipeRead, buffer, sizeof(buffer), &bytesRead, &bytesAvail, NULL)) {
					if (0 == bytesAvail) {
						NTOrData = false;
						DWORD dwExitCode = STILL_ACTIVE;
						if (GetExitCodeProcess(pi.hProcess, &dwExitCode)) {
							if (STILL_ACTIVE != dwExitCode) {
								// Process is dead, but wait a second in case there is some output in transit
								if (timeDetectedDeath == 0) {
									timeDetectedDeath = timeGetTime();
								} else {
									if ((timeGetTime() - timeDetectedDeath) > 1000) {
										completed = true;   // It's a dead process
									}
								}
							}
						}
					}
				}
			}

			if (!completed && NTOrData) {
				int bTest = ReadFile(
				                hPipeRead,
				                buffer,
				                sizeof(buffer),           // number of bytes to read
				                &bytesRead,
				                NULL         // non-overlapped.
				            );

				if (bTest && bytesRead) {
					// Display the data
					OutputAppendString(buffer, bytesRead);
					UpdateWindow(wSciTE.GetID());
				} else {
					completed = true;
				}
			}
		}

		if (worked) {
			WaitForSingleObject(pi.hProcess, INFINITE);
			GetExitCodeProcess(pi.hProcess, &exitcode);
			char exitmessage[80];
			if (exitcode == 0)
				isBuilt = true;
			sprintf(exitmessage, ">Exit code: %ld\n", exitcode);
			OutputAppendString(exitmessage);
			CloseHandle(pi.hProcess);
		}
		CloseHandle(hPipeRead);
	}

	// Move selection back to beginning of this run so that F4 will go
	// to first error of this run.
	SendOutput(SCI_GOTOPOS, originalEnd);
	SendMessage(wSciTE.GetID(), WM_COMMAND, IDM_FINISHEDEXECUTE, 0);
}

void ExecThread(void *ptw) {
	SciTEWindow *tw = reinterpret_cast<SciTEWindow *>(ptw);
	tw->ProcessExecute();
}
#endif

#ifdef PLAT_GTK
int xsystem(const char *s, const char *resultsFile) {
	int pid = 0;
	//printf("xsystem %s %s\n", s, resultsFile);
	if ((pid = fork()) == 0) {
		close(0);
		int fh = open(resultsFile, O_WRONLY);
		close(1);
		dup(fh);
		close(2);
		dup(fh);
		execlp("/bin/sh", "sh", "-c", s, 0);
		exit(127);
	}
	return pid;
}
#endif

void SciTEWindow::Execute() {
	if (clearBeforeExecute) {
		SendOutput(SCI_CLEARALL);
	}

	SendOutput(SCI_MARKERDELETEALL, -1);
	SendEditor(SCI_MARKERDELETEALL, -1);
	// Ensure the output pane is visible
	if (heightOutput < 20) {
		if (splitVertical)
			heightOutput = NormaliseSplit(300);
		else
			heightOutput = NormaliseSplit(100);
		SizeSubWindows();
		Redraw();
	}

	InterlockedExchange(&cancelFlag, 0L);
	executing = true;
	CheckMenus();
	chdir(dirName);
	strcpy(dirNameAtExecute, dirName);

#ifdef PLAT_GTK
	SendOutput(SCI_GOTOPOS, SendOutput(WM_GETTEXTLENGTH));
	originalEnd = SendOutput(SCI_GETCURRENTPOS);

	OutputAppendString(">");
	OutputAppendString(jobQueue[icmd].command.c_str());
	OutputAppendString("\n");

	unlink(resultsFile);

	int fdp = mkfifo(resultsFile, S_IRUSR | S_IWUSR);
	if (fdp < 0) {
		OutputAppendString(">Failed to create FIFO\n");
		return;
	}
	close(fdp);

	SString commandPlus = jobQueue[icmd].command;
	pidShell = xsystem(commandPlus.c_str(), resultsFile);
	fdFIFO = open(resultsFile, O_RDONLY);
	if (fdFIFO < 0) {
		OutputAppendString(">Failed to open\n");
		return;
	}

	inputHandle = gdk_input_add(fdFIFO, GDK_INPUT_READ,
	                            (GdkInputFunction) IOSignal, this);
#else
	_beginthread(ExecThread, 1024 * 1024, reinterpret_cast<void *>(this));
#endif
}

void SciTEWindow::CheckReload() {
	if (props.GetInt("load.on.activate")) {
		// Make a copy of fullPath as otherwise it gets aliased in Open
		char fullPathToCheck[MAX_PATH];
		strcpy(fullPathToCheck, fullPath);
		time_t newModTime = GetModTime(fullPathToCheck);
		//Platform::DebugPrintf("Times are %d %d\n", fileModTime, newModTime);
		if (newModTime > fileModTime) {
			if (isDirty) {
				static bool entered = false;  	// Stop reentrancy
				if (!entered && (0 == dialogsOnScreen)) {
					entered = true;
					char msg[MAX_PATH + 100];
					strcpy(msg, "The file \"");
					strcat(msg, fullPathToCheck);
					strcat(msg, "\" has been modified. Should it be reloaded?");
					dialogsOnScreen++;
					int decision = MessageBox(wSciTE.GetID(), msg, appName, MB_YESNO);
					dialogsOnScreen--;
					if (decision == IDYES) {
						Open(fullPathToCheck);
					}
					entered = false;
				}
			} else {
				Open(fullPathToCheck);
			}
		}
	}
}

void SciTEWindow::Activate(bool activeApp) {
	if (activeApp) {
		CheckReload();
	} else {
		if (isDirty) {
			if (props.GetInt("save.on.deactivate") && fileName[0]) {
				// Trying to save at deactivation when no file name -> dialogs
				Save();
			}
		}
	}
}

void SciTEWindow::Redraw() {
	wSciTE.InvalidateAll();
	wEditor.InvalidateAll();
	wOutput.InvalidateAll();
}

int DecodeMessage(char *cdoc, char *sourcePath, int format) {
	sourcePath[0] = '\0';
	if (format == 1) {
		// Python
		char *startPath = strchr(cdoc, '\"') + 1;
		char *endPath = strchr(startPath, '\"');
		strncpy(sourcePath, startPath, endPath - startPath);
		sourcePath[endPath - startPath] = 0;
		endPath++;
		while (*endPath && !isdigit(*endPath))
			endPath++;
		int sourceNumber = atoi(endPath) - 1;
		return sourceNumber;
	} else if (format == 2) {
		// GCC - look for number followed by colon to be line number
		// This will be preceded by file name
		for (int i = 0; cdoc[i]; i++) {
			if (isdigit(cdoc[i]) && cdoc[i + 1] == ':') {
				int j = i;
				while (j > 0 && isdigit(cdoc[j - 1]))
					j--;
				int sourceNumber = atoi(cdoc + j) - 1;
				strncpy(sourcePath, cdoc, j - 1);
				sourcePath[j - 1] = 0;
				return sourceNumber;
			}
		}
	} else if (format == 3) {
		// Visual *
		char *endPath = strchr(cdoc, '(');
		strncpy(sourcePath, cdoc, endPath - cdoc);
		sourcePath[endPath - cdoc] = 0;
		endPath++;
		int sourceNumber = atoi(endPath) - 1;
		return sourceNumber;
	}
	return - 1;
}

void SciTEWindow::GoMessage(int dir) {
	CHARRANGE crange;
	SendOutput(EM_EXGETSEL, 0, reinterpret_cast<LPARAM>(&crange));
	int selStart = crange.cpMin;
	int curLine = SendOutput(EM_LINEFROMCHAR, selStart);
	int maxLine = SendOutput(EM_GETLINECOUNT, selStart, 0);
	int lookLine = curLine + dir;
	if (lookLine < 0)
		lookLine = maxLine - 1;
	else if (lookLine >= maxLine)
		lookLine = 0;
	while ((dir == 0) || (lookLine != curLine)) {
		int startPosLine = SendOutput(EM_LINEINDEX, lookLine, 0);
		int lineLength = SendOutput(SCI_LINELENGTH, lookLine, 0);
		//Platform::DebugPrintf("GOMessage %d %d %d of %d linestart = %d\n", selStart, curLine, lookLine, maxLine, startPosLine);
		char style = SendOutput(SCI_GETSTYLEAT, startPosLine, 0);
		if (style != 0 && style != 4) {
			//Platform::DebugPrintf("Marker to %d\n", lookLine);
			SendOutput(SCI_MARKERDELETEALL, -1);
			SendOutput(SCI_MARKERDEFINE, 0, SC_MARK_SMALLRECT);
			SendOutput(SCI_MARKERSETFORE, 0, Colour(0x7f, 0, 0).AsLong());
			SendOutput(SCI_MARKERSETBACK, 0, Colour(0xff, 0xff, 0).AsLong());
			SendOutput(SCI_MARKERADD, lookLine, 0);
			CHARRANGE crangeset = {startPosLine, startPosLine};
			SendOutput(EM_EXSETSEL, 0, reinterpret_cast<LPARAM>(&crangeset));
			char *cdoc = new char[lineLength + 1];
			if (!cdoc)
				return;
			GetRange(wOutput, startPosLine, startPosLine + lineLength, cdoc);
			char sourcePath[MAX_PATH];
			int sourceLine = DecodeMessage(cdoc, sourcePath, style);
			//Platform::DebugPrintf("<%s> %d %d\n",sourcePath, sourceLine, lookLine);
			if (sourceLine >= 0) {
				if (0 != strcmp(sourcePath, fileName)) {
					char messagePath[MAX_PATH];
					if (Exists(dirNameAtExecute, sourcePath, messagePath)) {
						if (SaveIfUnsure() == IDCANCEL) {
							delete []cdoc;
							return;
						} else {
							Open(messagePath);
						}
					} else if (Exists(dirName, sourcePath, messagePath)) {
						if (SaveIfUnsure() == IDCANCEL) {
							delete []cdoc;
							return;
						} else {
							Open(messagePath);
						}
					}
				}
				SendEditor(SCI_MARKERDELETEALL, -1);
				SendEditor(SCI_MARKERDEFINE, 0, SC_MARK_CIRCLE);
				SendEditor(SCI_MARKERSETFORE, 0, Colour(0x7f, 0, 0).AsLong());
				SendEditor(SCI_MARKERSETBACK, 0, Colour(0xff, 0xff, 0).AsLong());
				SendEditor(SCI_MARKERADD, sourceLine, 0);
				int startSourceLine = SendEditor(EM_LINEINDEX, sourceLine, 0);
				SetSelection(startSourceLine, startSourceLine);
			}
			delete []cdoc;
			return;
		}
		if (dir == 0)
			dir = 1;
		lookLine += dir;
		if (lookLine < 0)
			lookLine = maxLine - 1;
		else if (lookLine >= maxLine)
			lookLine = 0;
	}
}

void SciTEWindow::StartCallTip() {
	//Platform::DebugPrintf("StartCallTip\n");
	char linebuf[1000];
	int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf),
	                         reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));
	int pos = SendEditor(SCI_GETCURRENTPOS);

	int startword = current - 1;
	while (startword > 0 && !isspace(linebuf[startword - 1]))
		startword--;
	linebuf[current] = '\0';
	int rootlen = current - startword;
	functionDefinition = "";
	//Platform::DebugPrintf("word  is [%s] %d %d %d\n", linebuf + startword, rootlen, pos, pos - rootlen);
	if (apis) {
		int i;
		for (i = 0; apis[i][0]; i++) {
			if (0 == strncmp(linebuf + startword, apis[i], rootlen)) {
				functionDefinition = apis[i];
				SendEditor(SCI_CALLTIPSHOW, pos - rootlen, reinterpret_cast<LPARAM>(apis[i]));
				ContinueCallTip();
			}
		}
	}
}

void SciTEWindow::ContinueCallTip() {
	char linebuf[1000];
	int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf),
	                         reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));

	int commas = 0;
	for (int i = 0; i < current; i++) {
		if (linebuf[i] == ',')
			commas++;
	}

	int startHighlight = 0;
	while (functionDefinition[startHighlight] && functionDefinition[startHighlight] != '(')
		startHighlight++;
	if (functionDefinition[startHighlight] == '(')
		startHighlight++;
	while (functionDefinition[startHighlight] && commas > 0) {
		if (functionDefinition[startHighlight] == ',' || functionDefinition[startHighlight] == ')')
			commas--;
		startHighlight++;
	}
	if (functionDefinition[startHighlight] == ',' || functionDefinition[startHighlight] == ')')
		startHighlight++;
	int endHighlight = startHighlight;
	if (functionDefinition[endHighlight])
		endHighlight++;
	while (functionDefinition[endHighlight] && functionDefinition[endHighlight] != ',' && functionDefinition[endHighlight] != ')')
		endHighlight++;

	SendEditor(SCI_CALLTIPSETHLT, startHighlight, endHighlight);
}

void SciTEWindow::StartAutoComplete() {
	char linebuf[1000];
	int current = SendEditor(SCI_GETCURLINE, sizeof(linebuf),
	                         reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));

	int startword = current;
	while (startword > 0 && !isspace(linebuf[startword - 1]))
		startword--;
	linebuf[current] = '\0';
	const char *root = linebuf + startword;
	int rootlen = current - startword;
	if (apis) {
		int lenchoices = 0;
		int i;
		for (i = 0; apis[i][0]; i++) {
			if (0 == strncmp(root, apis[i], rootlen)) {
				char *brace = strchr(apis[i], '(');
				if (brace)
					lenchoices += brace - apis[i];
				else
					lenchoices += strlen(apis[i]);
				lenchoices++;
			}
		}
		//Platform::DebugPrintf("lenchoices = %d\n", lenchoices);
		if (lenchoices) {
			char *choices = new char[lenchoices + 1];
			if (choices) {
				char *endchoice = choices;
				for (i = 0; apis[i][0]; i++) {
					if (0 == strncmp(root, apis[i], rootlen)) {
						if (endchoice != choices)
							*endchoice++ = ' ';
						char *brace = strchr(apis[i], '(');
						int len = strlen(apis[i]);
						if (brace)
							len = brace - apis[i];
						strncpy(endchoice, apis[i], len);
						endchoice += len;
					}
					*endchoice = '\0';
				}
				//Platform::DebugPrintf("rootlen = %d\n", rootlen);
				//Platform::DebugPrintf("choices = [%s]\n", choices);
				SendEditor(SCI_AUTOCSHOW, rootlen, reinterpret_cast<LPARAM>(choices));
				delete []choices;
			}
		}
	}
	//char word[220];
	//strcpy(word, "autocompletion.");
	//strcat(word, linebuf + startword);

	//SString possibles = props.GetNewExpand(word, fileName);
	//if (possibles.length())
	//	SendEditor(SCI_AUTOCSHOW, 0, reinterpret_cast<LPARAM>(possibles.c_str()));
}

int SciTEWindow::GetCurrentLineNumber() {
	CHARRANGE crange;
	SendEditor(EM_EXGETSEL, 0, reinterpret_cast<LPARAM>(&crange));
	int selStart = crange.cpMin;
	return SendEditor(EM_LINEFROMCHAR, selStart);
}

// Upon a character being added, SciTE may decide to perform some action
// such as displaying a completion list.
void SciTEWindow::CharAdded(char ch) {
	CHARRANGE crange;
	SendEditor(EM_EXGETSEL, 0, reinterpret_cast<LPARAM>(&crange));
	int selStart = crange.cpMin;
	int selEnd = crange.cpMax;
	if ((selEnd == selStart) && (selStart > 0)) {
		char style = SendEditor(SCI_GETSTYLEAT, selStart - 1, 0);
		//Platform::DebugPrintf("Char added %d style = %d\n", ch, style);
		if (style != 1) {
			if (SendEditor(SCI_CALLTIPACTIVE)) {
				if (ch == ')') {
					SendEditor(SCI_CALLTIPCANCEL);
				} else {
					ContinueCallTip();
				}
			} else if (SendEditor(SCI_AUTOCACTIVE)) {
				if (ch == '(') {
					StartCallTip();
				} else if (!isalpha(ch) && (ch != '_')) {
					SendEditor(SCI_AUTOCCANCEL);
				}
			} else {
				if (ch == '(') {
					StartCallTip();
				} else if (ch == '\r' || ch == '\n') {
					char linebuf[1000];
					int curLine = GetCurrentLineNumber();
					int lineLength = SendEditor(SCI_LINELENGTH, curLine);
					//Platform::DebugPrintf("[CR] %d len = %d\n", curLine, lineLength);
					if (curLine > 0 && lineLength <= 2) {
						int prevLineLength = SendEditor(SCI_LINELENGTH, curLine - 1);
						if (prevLineLength < sizeof(linebuf)) {
							WORD buflen = sizeof(linebuf);
							memcpy(linebuf, &buflen, sizeof(buflen));
							SendEditor(EM_GETLINE, curLine - 1,
							           reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));
							linebuf[prevLineLength] = '\0';
							for (int pos = 0; linebuf[pos]; pos++) {
								if (linebuf[pos] != ' ' && linebuf[pos] != '\t')
									linebuf[pos] = '\0';
							}
							SendEditor(EM_REPLACESEL, 0, reinterpret_cast<LPARAM>(static_cast<char *>(linebuf)));
						}
					}
				}
			}
		}
	}
}

#ifdef PLAT_GTK

void SciTEWindow::GotoCancelSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->gotoDialog.Destroy();
}

void SciTEWindow::GotoKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew) {
	if (event->keyval == GDK_Escape) {
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		gtk_widget_destroy(GTK_WIDGET(w));
		scitew->gotoDialog = 0;
	}
}

void SciTEWindow::GotoSignal(GtkWidget *, SciTEWindow *scitew) {
	char *lineEntry = gtk_entry_get_text(GTK_ENTRY(scitew->gotoEntry));
	int lineNo = atoi(lineEntry);

	scitew->SendEditor(SCI_GOTOLINE, lineNo - 1);

	scitew->gotoDialog.Destroy();
}
#else
BOOL CALLBACK SciTEWindow::GoLineDlg(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {
	static int *pLineNo;

	switch (message) {

	case WM_INITDIALOG:
		pLineNo = reinterpret_cast<int *>(lParam);
		SendDlgItemMessage(hDlg, IDGOLINE, EM_LIMITTEXT, 10, 1);
		SetDlgItemInt(hDlg, IDCURRLINE, pLineNo[0], FALSE);
		SetDlgItemInt(hDlg, IDLASTLINE, pLineNo[1], FALSE);
		return TRUE;

	case WM_CLOSE:
		SendMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
		break;

	case WM_COMMAND:
		if (ControlIDOfCommand(wParam) == IDCANCEL) {
			EndDialog(hDlg, IDCANCEL);
			return FALSE;
		} else if (ControlIDOfCommand(wParam) == IDOK) {
			BOOL bOK;
			pLineNo[0] = static_cast<int>(GetDlgItemInt(hDlg, IDGOLINE, &bOK, FALSE));
			if (!bOK)
				pLineNo[0] = -1;
			//pLineNo[1] = (SendDlgItemMessage(hDlg, IDEXTEND, BM_GETCHECK, 0, 0L) ==
			//	BST_CHECKED ? TRUE : FALSE);
			EndDialog(hDlg, IDOK);
			return TRUE;
		}
	}

	return FALSE;
}
#endif

void SciTEWindow::GoLineDialog() {
#ifdef PLAT_GTK
	gotoDialog = gtk_dialog_new();
	gtk_window_set_title(GTK_WINDOW(gotoDialog.GetID()), "Go To");
	gtk_container_border_width(GTK_CONTAINER(gotoDialog.GetID()), 0);

	gtk_signal_connect(GTK_OBJECT(gotoDialog.GetID()),
	                   "destroy", GtkSignalFunc(destroyDialog), 0);

	GtkWidget *table = gtk_table_new(2, 1, FALSE);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog.GetID())->vbox),
	                   table, TRUE, TRUE, 0);

	GtkAttachOptions opts = static_cast<GtkAttachOptions>(
	                            GTK_EXPAND | GTK_SHRINK | GTK_FILL);
	GtkWidget *label = gtk_label_new("Go to line:");
	gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, opts, opts, 5, 5);
	gtk_widget_show(label);

	gotoEntry = gtk_entry_new();
	gtk_table_attach(GTK_TABLE(table), gotoEntry, 1, 2, 0, 1, opts, opts, 5, 5);
	gtk_signal_connect(GTK_OBJECT(gotoEntry),
	                   "activate", GtkSignalFunc(GotoSignal), this);
	gtk_widget_grab_focus(GTK_WIDGET(gotoEntry));
	gtk_widget_show(gotoEntry);

	gtk_widget_show(table);

	GtkWidget *buttonGoTo = gtk_button_new_with_label("Go To");
	gtk_signal_connect(GTK_OBJECT(buttonGoTo),
	                   "clicked", GtkSignalFunc(GotoSignal), this);
	GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonGoTo), GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog.GetID())->action_area),
	                   buttonGoTo, TRUE, TRUE, 0);
	gtk_widget_grab_default(GTK_WIDGET(buttonGoTo));
	gtk_widget_show(buttonGoTo);

	GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
	gtk_signal_connect(GTK_OBJECT(buttonCancel),
	                   "clicked", GtkSignalFunc(GotoCancelSignal), this);
	gtk_signal_connect(GTK_OBJECT(gotoDialog.GetID()),
	                   "key_press_event", GtkSignalFunc(GotoKeySignal),
	                   this);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(gotoDialog.GetID())->action_area),
	                   buttonCancel, TRUE, TRUE, 0);
	gtk_widget_show(buttonCancel);

	// Mark it as a modal transient dialog
	gtk_window_set_modal(GTK_WINDOW(gotoDialog.GetID()), TRUE);
	gtk_window_set_transient_for (GTK_WINDOW(gotoDialog.GetID()),
	                              GTK_WINDOW(wSciTE.GetID()));

	gotoDialog.Show();
#else
	int lineNo[2] = { 0, 0 };
	lineNo[0] = SendMessage(wEditor.GetID(), EM_LINEFROMCHAR, static_cast<WPARAM>( -1), 0L) + 1;
	lineNo[1] = SendMessage(wEditor.GetID(), EM_GETLINECOUNT, 0, 0L);
	if (DoDialog(hInstance, "GoLine", wSciTE.GetID(),
	             reinterpret_cast<DLGPROC>(GoLineDlg),
	             reinterpret_cast<DWORD>(lineNo)) == IDOK) {
		//Platform::DebugPrintf("asked to go to %d\n", lineNo);
		if (lineNo[0] != -1) {
			//if (lineNo[1] == TRUE) {
			//	int selStart, selStop;
			//	selStart = SendMessage(wEditor.GetID(), SCI_GETANCHOR, 0, 0L);
			//	selStop = SendMessage(wEditor.GetID(), EM_LINEINDEX, lineNo[0] - 1, 0L);
			//	SendEditor(EM_SETSEL, selStart, selStop);
			//} else {
			SendEditor(SCI_GOTOLINE, lineNo[0]-1);
			//}
		}
	}
	SetFocus(wEditor.GetID());
#endif
}

void SciTEWindow::GoMatchingBrace() {
	int braceAtCaret = -1;
	int braceOpposite = -1;
	FindMatchingBracePosition(true, braceAtCaret, braceOpposite);
	if (braceOpposite >= 0) {
		SendEditor(SCI_GETCHARAT, braceOpposite);
		braceOpposite++;
		SetSelection(braceOpposite, braceOpposite);
	}
}

#ifdef PLAT_GTK

void SciTEWindow::FindReplaceGrabFields() {
	char *findEntry = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(comboFind)->entry));
	strncpy(findWhat, findEntry, sizeof(findWhat));
	memFinds.Insert(findWhat);
	if (comboReplace) {
		char *replaceEntry = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(comboReplace)->entry));
		strncpy(replaceWhat, replaceEntry, sizeof(replaceWhat));
		memReplaces.Insert(replaceWhat);
	}
	wholeWord = GTK_TOGGLE_BUTTON(toggleWord)->active;
	matchCase = GTK_TOGGLE_BUTTON(toggleCase)->active;
	reverseFind = GTK_TOGGLE_BUTTON(toggleReverse)->active;
}

void SciTEWindow::FRCancelSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->wFindReplace.Destroy();
}

void SciTEWindow::FRKeySignal(GtkWidget *w, GdkEventKey *event, SciTEWindow *scitew) {
	if (event->keyval == GDK_Escape) {
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		scitew->wFindReplace.Destroy();
	}
}

void SciTEWindow::FRFindSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->FindReplaceGrabFields();
	if (scitew->findWhat[0]) {
		scitew->FindNext();
	}
}

void SciTEWindow::FRReplaceSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->FindReplaceGrabFields();
	scitew->ReplaceOnce();
}

void SciTEWindow::FRReplaceAllSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->FindReplaceGrabFields();
	scitew->ReplaceAll();
	scitew->wFindReplace.Destroy();
}
#endif

void SciTEWindow::FindReplace(bool replace) {
	replacing = replace;
#ifdef PLAT_GTK
	wFindReplace = gtk_dialog_new();
	gtk_window_set_policy(GTK_WINDOW(wFindReplace.GetID()), TRUE, TRUE, TRUE);
	wFindReplace.SetTitle(replace ? "Replace" : "Find");

	gtk_signal_connect(GTK_OBJECT(wFindReplace.GetID()),
	                   "destroy", GtkSignalFunc(destroyDialog), 0);

	GtkWidget *table = gtk_table_new(2, replace ? 4 : 3, FALSE);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace.GetID())->vbox),
	                   table, TRUE, TRUE, 0);

	GtkAttachOptions opts = static_cast<GtkAttachOptions>(
	                            GTK_SHRINK | GTK_FILL);

	GtkAttachOptions optse = static_cast<GtkAttachOptions>(
	                             GTK_SHRINK | GTK_FILL | GTK_EXPAND);

	int row = 0;

	GtkWidget *labelFind = gtk_label_new("Find:");
	gtk_table_attach(GTK_TABLE(table), labelFind, 0, 1,
	                 row, row + 1, opts, opts, 5, 5);
	gtk_widget_show(labelFind);

	comboFind = gtk_combo_new();
	FillComboFromMemory(comboFind, memFinds);

	gtk_table_attach(GTK_TABLE(table), comboFind, 1, 2,
	                 row, row + 1, optse, opts, 5, 5);
	gtk_widget_show(comboFind);
	gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(comboFind)->entry), findWhat);
	gtk_entry_select_region(GTK_ENTRY(GTK_COMBO(comboFind)->entry), 0, strlen(findWhat));
	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(comboFind)->entry),
	                   "activate", GtkSignalFunc(FRFindSignal), this);

	row++;

	if (replace) {
		GtkWidget *labelReplace = gtk_label_new("Replace:");
		gtk_table_attach(GTK_TABLE(table), labelReplace, 0, 1,
		                 row, row + 1, opts, opts, 5, 5);
		gtk_widget_show(labelReplace);

		comboReplace = gtk_combo_new();
		FillComboFromMemory(comboReplace, memReplaces);

		gtk_table_attach(GTK_TABLE(table), comboReplace, 1, 2,
		                 row, row + 1, optse, opts, 5, 5);
		gtk_widget_show(comboReplace);
		gtk_signal_connect(GTK_OBJECT(GTK_COMBO(comboReplace)->entry),
		                   "activate", GtkSignalFunc(FRFindSignal), this);

		row++;
	} else {
		comboReplace = 0;
	}

	// Case Sensitive
	toggleCase = gtk_check_button_new_with_label("Case Sensitive");
	gtk_table_attach(GTK_TABLE(table), toggleCase, 0, 2, row, row + 1, opts, opts, 3, 0);
	gtk_widget_show(toggleCase);
	row++;

	// Whole Word
	toggleWord = gtk_check_button_new_with_label("Whole Word");
	gtk_table_attach(GTK_TABLE(table), toggleWord, 0, 2, row, row + 1, opts, opts, 3, 0);
	gtk_widget_show(toggleWord);
	row++;

	// Whole Word
	toggleReverse = gtk_check_button_new_with_label("Reverse Direction");
	gtk_table_attach(GTK_TABLE(table), toggleReverse, 0, 2, row, row + 1, opts, opts, 3, 0);
	gtk_widget_show(toggleReverse);

	gtk_widget_show(table);

	GtkWidget *buttonFind = gtk_button_new_with_label("Find");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace.GetID())->action_area),
	                   buttonFind, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(buttonFind),
	                   "clicked", GtkSignalFunc(FRFindSignal), this);
	gtk_widget_show(buttonFind);

	if (replace) {
		GtkWidget *buttonReplace = gtk_button_new_with_label("Replace");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace.GetID())->action_area),
		                   buttonReplace, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonReplace),
		                   "clicked", GtkSignalFunc(FRReplaceSignal), this);
		gtk_widget_show(buttonReplace);

		GtkWidget *buttonReplaceAll = gtk_button_new_with_label("Replace All");
		gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace.GetID())->action_area),
		                   buttonReplaceAll, TRUE, TRUE, 0);
		gtk_signal_connect(GTK_OBJECT(buttonReplaceAll),
		                   "clicked", GtkSignalFunc(FRReplaceAllSignal), this);
		gtk_widget_show(buttonReplaceAll);
	}

	GtkWidget *buttonCancel = gtk_button_new_with_label("Cancel");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(wFindReplace.GetID())->action_area),
	                   buttonCancel, TRUE, TRUE, 0);
	gtk_signal_connect(GTK_OBJECT(buttonCancel),
	                   "clicked", GtkSignalFunc(FRCancelSignal), this);
	gtk_signal_connect(GTK_OBJECT(wFindReplace.GetID()),
	                   "key_press_event", GtkSignalFunc(FRKeySignal),
	                   this);
	gtk_widget_show(buttonCancel);

	GTK_WIDGET_SET_FLAGS(GTK_WIDGET(buttonFind), GTK_CAN_DEFAULT);
	gtk_widget_grab_default(GTK_WIDGET(buttonFind));
	gtk_widget_grab_focus(GTK_WIDGET(GTK_COMBO(comboFind)->entry));

	// Mark it as a transient dialog
	gtk_window_set_transient_for (GTK_WINDOW(wFindReplace.GetID()),
	                              GTK_WINDOW(wSciTE.GetID()));

	wFindReplace.Show();
#else
#endif
}

void SciTEWindow::AddCommand(const SString &cmd, const SString &dir, JobSubsystem jobType) {
	if (cmd.length()) {
		//Platform::DebugPrintf("AddCommand %s\n",cmd.c_str());
		jobQueue[commandCurrent].command = cmd;
		jobQueue[commandCurrent].directory = dir;
		jobQueue[commandCurrent].jobType = jobType;
		commandCurrent++;
	}
}

void SciTEWindow::AboutDialog() {
#ifdef PLAT_GTK
	//AboutDialog();
	MessageBox(wSciTE.GetID(), "SciTE\nby Neil Hodgson neilh@scintilla.org .",
	           appName, MB_OK | MB_ABOUTBOX);
#else
	DoDialog(hInstance, "About", wSciTE.GetID(), NULL, 0);
	SetFocus(wEditor.GetID());
#endif
}

void SciTEWindow::QuitProgram() {
	if (SaveIfUnsure() != IDCANCEL) {
#ifdef PLAT_GTK
		gtk_exit(0);
#else
		SendMessage(wSciTE.GetID(), WM_CLOSE, 0, 0L);
#endif
	}
}

void SciTEWindow::Command(WPARAM wParam, LPARAM lParam) {
	int cmdID = ControlIDOfCommand(wParam);
	switch (cmdID) {
	case IDM_NEW:
		// For the New command, the are you sure question is always asked as this gives
		// an opportunity to abandon the edits made to a file when are.you.sure is turned off.
		if (SaveIfUnsure(true) != IDCANCEL) {
			New();
			ReadProperties();
		}
		break;
	case IDM_OPEN:
		if (SaveIfUnsure() != IDCANCEL) {
			Open();
			SetFocus(wEditor.GetID());
		}
		break;
	case IDM_CLOSE:
		if (SaveIfUnsure() != IDCANCEL) {
			DropFileStackTop();
			StackMenu(0);
			// If none left, New()
			SetFocus(wEditor.GetID());
		}
		break;
	case IDM_SAVE:
		Save();
		SetFocus(wEditor.GetID());
		break;
	case IDM_SAVEAS:
		SaveAs();
		SetFocus(wEditor.GetID());
		break;
	case IDM_SAVEASHTML:
		SaveAsHTML();
		SetFocus(wEditor.GetID());
		break;
	case IDM_REVERT:
		break;
	case IDM_PRINT:
		Print();
		break;
	case IDM_PRINTSETUP:
		PrintSetup();
		break;
	case IDM_ABOUT:
		AboutDialog();
		break;
	case IDM_QUIT:
		QuitProgram();
		break;
	case IDM_NEXTFILE:
		if (SaveIfUnsure() != IDCANCEL) {
			StackMenu(1);
		}
		break;

	case IDM_UNDO:
		SendEditor(WM_UNDO);
		break;
	case IDM_REDO:
		SendEditor(SCI_REDO);
		break;

	case IDM_CUT:
		SendEditor(WM_CUT);
		break;
	case IDM_COPY:
		SendEditor(WM_COPY);
		break;
	case IDM_PASTE:
		SendEditor(WM_PASTE);
		break;
	case IDM_CLEAR:
		SendEditor(WM_CLEAR);
		break;
	case IDM_SELECTALL:
		SendEditor(SCI_SELECTALL);
		break;

	case IDM_FIND:
		Find();
		break;

	case IDM_FINDNEXT:
		FindNext();
		break;

	case IDM_FINDINFILES:
		FindInFiles();
		break;

	case IDM_REPLACE:
		Replace();
		break;

	case IDM_GOTO:
		GoLineDialog();
		break;

	case IDM_MATCHBRACE:
		GoMatchingBrace();
		break;

	case IDM_COMPLETE:
		StartAutoComplete();
		break;

	case IDM_SPLITVERTICAL:
		splitVertical = !splitVertical;
		heightOutput = NormaliseSplit(heightOutput);
		SizeSubWindows();
		CheckMenus();
		Redraw();
		break;

	case IDM_LINENUMBERMARGIN:
		lineNumbers = !lineNumbers;
		SendEditor(SCI_SETLINENUMBERWIDTH, lineNumbers ? lineNumbersWidth : 0);
		CheckMenus();
		break;

	case IDM_SELMARGIN:
		margin = !margin;
		SendEditor(SCI_SETMARGINWIDTH, margin ? marginWidth : 0);
		CheckMenus();
		break;

	case IDM_VIEWEOL:
		SendEditor(SCI_SETVIEWEOL, !SendEditor(SCI_GETVIEWEOL));
		CheckMenus();
		break;

#ifdef STATUSBAR
	case IDM_VIEWSTATUSBAR:
		sbVisible = ! sbVisible;
		if (sbVisible) {
			gtk_widget_show(GTK_WIDGET(wStatusBar.GetID()));
		} else {
			gtk_widget_hide(GTK_WIDGET(wStatusBar.GetID()));
		}
		CheckMenus();
		break;
#endif

	case IDM_SLAVE:
		if (!isSlaved) {
			// New scheme can only transition to slave mode
			isSlaved = true;
			SendOutput(SCI_SETDOCPOINTER, 0, SendEditor(SCI_GETDOCPOINTER));
		}
		CheckMenus();
		break;

	case IDM_EOL_CRLF:
		SendEditor(SCI_SETEOLMODE, SC_EOL_CRLF);
		CheckMenus();
		break;
	case IDM_EOL_CR:
		SendEditor(SCI_SETEOLMODE, SC_EOL_CR);
		CheckMenus();
		break;
	case IDM_EOL_LF:
		SendEditor(SCI_SETEOLMODE, SC_EOL_LF);
		CheckMenus();
		break;
	case IDM_EOL_CONVERT:
		SendEditor(SCI_CONVERTEOLS, SendEditor(SCI_GETEOLMODE));
		break;

	case IDM_UPRCASE:
		//Platform::DebugPrintf("IDM_UPRCASE\n");
		break;
	case IDM_LWRCASE:
		//Platform::DebugPrintf("IDM_LWRCASE\n");
		break;
	case IDM_VIEWSPACE: {
			int viewWS = SendEditor(SCI_GETVIEWWS, 0, 0);
			SendEditor(SCI_SETVIEWWS, !viewWS);
			CheckMenus();
			Redraw();
		}
		break;

	case IDM_COMPILE: {
			if (SaveIfUnsure() != IDCANCEL) {
				SelectionIntoProperties();
				AddCommand(props.GetNewExpand("command.compile.", fileName), "",
				           SubsystemType("command.compile.subsystem."));
				if (commandCurrent > 0)
					Execute();
			}
		}
		break;

	case IDM_BUILD: {
			if (SaveIfUnsure() != IDCANCEL) {
				SelectionIntoProperties();
				AddCommand(props.GetNewExpand("command.build.", fileName), "",
				           SubsystemType("command.build.subsystem."));
				if (commandCurrent > 0)
					Execute();
			}
		}
		break;

	case IDM_GO: {
			if (SaveIfUnsure() != IDCANCEL) {
				SelectionIntoProperties();
				if (!isBuilt)
					AddCommand(props.GetNewExpand("command.go.needs.", fileName), "",
					           SubsystemType("command.go.needs.subsystem."));
				AddCommand(props.GetNewExpand("command.go.", fileName), "",
				           SubsystemType("command.go.subsystem."));
				if (commandCurrent > 0)
					Execute();
			}
		}
		break;

	case IDM_STOPEXECUTE:
#ifdef PLAT_GTK
		kill(pidShell, SIGKILL);
#else
		InterlockedExchange(&cancelFlag, 1L);
#endif
		break;

#ifdef PLAT_WIN
	case IDM_FINISHEDEXECUTE: {
			executing = false;
			CheckMenus();
			for (int icmd = 0; icmd < commandMax; icmd++) {
				jobQueue[icmd].Clear();
			}
			commandCurrent = 0;
			CheckReload();
		}
		break;
#endif

	case IDM_NEXTMSG:
		GoMessage(1);
		break;

	case IDM_PREVMSG:
		GoMessage( -1);
		break;

	case IDM_OPENLOCALPROPERTIES:
		OpenProperties(true);
		SetFocus(wEditor.GetID());
		break;

	case IDM_OPENGLOBALPROPERTIES:
		OpenProperties(false);
		SetFocus(wEditor.GetID());
		break;

	case IDM_SRCWIN:
		break;

	case IDM_ACTIVATE:
		Activate(lParam);
		break;

	default:
		if ((cmdID >= fileStackCmdID) &&
		        (cmdID < fileStackCmdID + fileStackMax)) {
			if (SaveIfUnsure() != IDCANCEL) {
				StackMenu(cmdID - fileStackCmdID);
			}
		} else if (cmdID >= IDM_TOOLS && cmdID < IDM_TOOLS + 10) {
			ToolsMenu(cmdID - IDM_TOOLS);
		}
		break;
	}
}

void SciTEWindow::Notify(SCNotification *notification) {
	switch (notification->nmhdr.code) {
	case SCN_STYLENEEDED: {
			if (notification->nmhdr.idFrom == IDM_SRCWIN) {
				int endStyled = SendEditor(SCI_GETENDSTYLED);
				int lineEndStyled = SendEditor(EM_LINEFROMCHAR, endStyled);
				endStyled = SendEditor(EM_LINEINDEX, lineEndStyled);
				Colourise(endStyled, notification->position);
			} else {
				int endStyled = SendOutput(SCI_GETENDSTYLED);
				int lineEndStyled = SendOutput(EM_LINEFROMCHAR, endStyled);
				endStyled = SendOutput(EM_LINEINDEX, lineEndStyled);
				Colourise(endStyled, notification->position, false);
			}
		}
		break;

	case SCN_CHARADDED:
		CharAdded(notification->ch);
		break;

	case SCN_SAVEPOINTREACHED:
		if (notification->nmhdr.idFrom == IDM_SRCWIN) {
			isDirty = false;
			CheckMenus();
			SetWindowName();
		}
		break;

	case SCN_SAVEPOINTLEFT:
		if (notification->nmhdr.idFrom == IDM_SRCWIN) {
			isDirty = true;
			isBuilt = false;
			CheckMenus();
			SetWindowName();
		}
		break;

#ifdef PLAT_GTK
	case SCN_KEY: {
			int mods = 0;
			if (notification->modifiers & SHIFT_PRESSED)
				mods |= GDK_SHIFT_MASK;
			if (notification->modifiers & LEFT_CTRL_PRESSED)
				mods |= GDK_CONTROL_MASK;
			if (notification->modifiers & LEFT_ALT_PRESSED)
				mods |= GDK_MOD1_MASK;
			//Platform::DebugPrintf("SCN_KEY: %d %d\n", notification->ch, mods);
			if ((mods == GDK_CONTROL_MASK) && (notification->ch == VK_TAB)) {
				Command(IDM_NEXTFILE, 0);
			} else {
				gtk_accel_group_activate(accelGroup, notification->ch,
				                         static_cast<GdkModifierType>(mods));
			}
		}
		break;
#endif

	case SCN_DOUBLECLICK:
		if (notification->nmhdr.idFrom == IDM_RUNWIN) {
			//Platform::DebugPrintf("Double click 0\n");
			GoMessage(0);
		}
		break;

	case SCN_CHECKBRACE:
		BraceMatch(notification->nmhdr.idFrom == IDM_SRCWIN);
#ifdef STATUSBAR
		if (notification->nmhdr.idFrom == IDM_SRCWIN) {
			if (sbVisible) {
				SString msg;
				int caretPos = SendEditor(SCI_GETCURRENTPOS);
				int caretLine = SendEditor(EM_LINEFROMCHAR, caretPos);
				int caretLineStart = SendEditor(EM_LINEINDEX, caretLine);
				msg = "Column=";
				msg += SString(caretPos - caretLineStart + 1).c_str();
				msg += "    Line=";
				msg += SString(caretLine + 1).c_str();
				if (!(sbValue == msg)) {
					gtk_statusbar_pop(GTK_STATUSBAR(wStatusBar.GetID()), sbContextID);
					gtk_statusbar_push(GTK_STATUSBAR(wStatusBar.GetID()), sbContextID, msg.c_str());
					sbValue = msg;
				}
			} else {
				sbValue = "";
			}
		}
#endif
		break;

	case SCN_MODIFIED:
		/*
				if ((!isSlaved) && (notification->nmhdr.idFrom == IDM_SRCWIN)) {
					char msg[400];
					char chType = 'I';
					if (notification->modificationType & SC_MOD_DELETETEXT)
						chType = 'D';
					if (notification->modificationType & SC_MOD_CHANGESTYLE)
						chType = 'C';
					char chSource = 'N';
					if (notification->modificationType & SC_PERFORMED_UNDO)
						chSource = 'U';
					if (notification->modificationType & SC_PERFORMED_REDO)
						chSource = 'R';
					char chLast = (notification->modificationType & SC_LASTSTEPINUNDOREDO) ? 'E' : ' ';
					char chLines = (notification->linesAdded != 0) ? '+' : ' ';
					sprintf(msg, "%c%c%c%c %d ", chLast, chLines, chType, chSource, notification->position);
					int len = Platform::Minimum(notification->length, 100);
					if (notification->text)
						strncat(msg, notification->text, len);
					strcat(msg, "\r\n");
					OutputAppendString(msg);
				}
		*/
		break;

	}
}

void SciTEWindow::CheckAMenuItem(WORD wIDCheckItem, BOOL val) {
#ifdef PLAT_GTK
	GtkWidget *item = gtk_item_factory_get_widget_by_action(itemFactory, wIDCheckItem);
	allowMenuActions = false;
	if (item)
		gtk_check_menu_item_set_state(GTK_CHECK_MENU_ITEM(item), val ? TRUE : FALSE);
	else
		Platform::DebugPrintf("Could not find %x\n", wIDCheckItem);
	allowMenuActions = true;
#else
	if (val)
		CheckMenuItem(GetMenu(wSciTE.GetID()), wIDCheckItem, MF_CHECKED | MF_BYCOMMAND);
	else
		CheckMenuItem(GetMenu(wSciTE.GetID()), wIDCheckItem, MF_UNCHECKED | MF_BYCOMMAND);
#endif
}

void SciTEWindow::EnableAMenuItem(WORD wIDCheckItem, BOOL val) {
#ifdef PLAT_GTK
	GtkWidget *item = gtk_item_factory_get_widget_by_action(itemFactory, wIDCheckItem);
	if (item) {
		//Platform::DebugPrintf("Set %d to %d\n", wIDCheckItem, val);
		if (GTK_IS_WIDGET(item))
			gtk_widget_set_sensitive(item, val);
	}
#else
	if (val)
		EnableMenuItem(GetMenu(wSciTE.GetID()), wIDCheckItem, MF_ENABLED | MF_BYCOMMAND);
	else
		EnableMenuItem(GetMenu(wSciTE.GetID()), wIDCheckItem, MF_DISABLED | MF_GRAYED | MF_BYCOMMAND);
#endif
}

void SciTEWindow::CheckMenus() {
	EnableAMenuItem(IDM_SAVE, isDirty);
	EnableAMenuItem(IDM_UNDO, SendEditor(EM_CANUNDO));
	EnableAMenuItem(IDM_REDO, SendEditor(SCI_CANREDO));
	EnableAMenuItem(IDM_PASTE, SendEditor(EM_CANPASTE));
	EnableAMenuItem(IDM_FINDINFILES, !executing);
	EnableAMenuItem(IDM_COMPLETE, apis != 0);
	CheckAMenuItem(IDM_SPLITVERTICAL, splitVertical);
	CheckAMenuItem(IDM_VIEWSPACE, SendEditor(SCI_GETVIEWWS));
	CheckAMenuItem(IDM_LINENUMBERMARGIN, lineNumbers);
	CheckAMenuItem(IDM_SELMARGIN, margin);
	CheckAMenuItem(IDM_VIEWEOL, SendEditor(SCI_GETVIEWEOL));
	//CheckAMenuItem(IDM_SLAVE, isSlaved);
	EnableAMenuItem(IDM_COMPILE, !executing);
	EnableAMenuItem(IDM_BUILD, !executing);
	EnableAMenuItem(IDM_GO, !executing);
	for (int toolItem = 0; toolItem < toolMax; toolItem++)
		EnableAMenuItem(IDM_TOOLS + toolItem, !executing);
	EnableAMenuItem(IDM_STOPEXECUTE, executing);
#ifdef PLAT_GTK
	//Platform::DebugPrintf("EOL Mode is %d\n", SendEditor(SCI_GETEOLMODE));
	CheckAMenuItem(IDM_EOL_CRLF, SendEditor(SCI_GETEOLMODE) == SC_EOL_CRLF);
	CheckAMenuItem(IDM_EOL_CR, SendEditor(SCI_GETEOLMODE) == SC_EOL_CR);
	CheckAMenuItem(IDM_EOL_LF, SendEditor(SCI_GETEOLMODE) == SC_EOL_LF);
#ifdef STATUSBAR
	CheckAMenuItem(IDM_VIEWSTATUSBAR, sbVisible);
#endif
#else
	CheckMenuRadioItem(GetMenu(wSciTE.GetID()), IDM_EOL_CRLF, IDM_EOL_LF,
	                   SendEditor(SCI_GETEOLMODE) - SC_EOL_CRLF + IDM_EOL_CRLF, 0);
#endif
}

void SciTEWindow::SizeSubWindows() {
	PRectangle rcClient = GetClientRectangle();
	int w = rcClient.right - rcClient.left;
	int h = rcClient.bottom - rcClient.top;
	heightOutput = NormaliseSplit(heightOutput);
	if (splitVertical) {
		wEditor.SetPosition(PRectangle(0, 0, w - heightOutput - heightBar, h));
#ifdef PLAT_GTK
		wDivider.SetPosition(PRectangle(w - heightOutput - heightBar, 0, w - heightOutput, h));
#endif
		wOutput.SetPosition(PRectangle(w - heightOutput, 0, w, h));
	} else {
		wEditor.SetPosition(PRectangle(0, 0, w, h - heightOutput - heightBar));
#ifdef PLAT_GTK
		wDivider.SetPosition(PRectangle(0, h - heightOutput - heightBar, w, h - heightOutput));
#endif
		wOutput.SetPosition(PRectangle(0, h - heightOutput, w, h));
	}
}

PRectangle SciTEWindow::GetClientRectangle() {
#ifdef PLAT_GTK
	return content.GetClientPosition();
#else
	return wSciTE.GetClientPosition();
#endif
}

// Ensure that a splitter bar position is inside the main window
int SciTEWindow::NormaliseSplit(int splitPos) {
	PRectangle rcClient = GetClientRectangle();
	int w = rcClient.right - rcClient.left;
	int h = rcClient.bottom - rcClient.top;
	if (splitPos < 20)
		splitPos = 0;
	if (splitVertical) {
		if (splitPos > w - heightBar - 20)
			splitPos = w - heightBar;
	} else {
		if (splitPos > h - heightBar - 20)
			splitPos = h - heightBar;
	}
	return splitPos;
}

void SciTEWindow::MoveSplit(Point ptNewDrag) {
	int newHeightOutput = heightOutputStartDrag + (ptStartDrag.y - ptNewDrag.y);
	if (splitVertical)
		newHeightOutput = heightOutputStartDrag + (ptStartDrag.x - ptNewDrag.x);
	newHeightOutput = NormaliseSplit(newHeightOutput);
	if (heightOutput != newHeightOutput) {
		heightOutput = newHeightOutput;
		SizeSubWindows();
		Redraw();
	}

}

#ifdef PLAT_WIN
void SciTEWindow::Run(const char *cmdLine) {
	Open(cmdLine, true);
	wSciTE.Show();
}
#endif

#ifdef PLAT_GTK
gint SciTEWindow::MoveResize(GtkWidget *, GtkAllocation *allocation, SciTEWindow *scitew) {
	//Platform::DebugPrintf("tw move resize %d %d\n", allocation->width, allocation->height);
	scitew->SizeSubWindows();
	return TRUE;
}

void SciTEWindow::QuitSignal(GtkWidget *, SciTEWindow *scitew) {
	scitew->Command(IDM_QUIT, 0);
}

void SciTEWindow::MenuSignal(SciTEWindow *scitew, guint action, GtkWidget *) {
	//Platform::DebugPrintf("action %d %x \n", action, w);
	if (scitew->allowMenuActions)
		scitew->Command(action, 0);
}

void SciTEWindow::CommandSignal(GtkWidget *, gint wParam, gpointer lParam, SciTEWindow *scitew) {
	//Platform::DebugPrintf("Command: %x %x %x\n", w, wParam, lParam);
	scitew->Command(wParam, reinterpret_cast<LPARAM>(lParam));
}

void SciTEWindow::NotifySignal(GtkWidget *, gint wParam, gpointer lParam, SciTEWindow *scitew) {
	//Platform::DebugPrintf("Notify: %x %x %x\n", w, wParam, lParam);
	scitew->Notify(reinterpret_cast<SCNotification *>(lParam));
}

void SciTEWindow::DividerXOR(Point pt) {
	if (!xor_gc) {
		GdkGCValues values;
		values.foreground = wSciTE.GetID()->style->white;
		values.function = GDK_XOR;
		values.subwindow_mode = GDK_INCLUDE_INFERIORS;
		xor_gc = gdk_gc_new_with_values(wSciTE.GetID()->window,
		                                &values,
		                                static_cast<enum GdkGCValuesMask>(
		                                    GDK_GC_FOREGROUND | GDK_GC_FUNCTION | GDK_GC_SUBWINDOW));
	}
	if (splitVertical) {
		gdk_draw_line(wSciTE.GetID()->window, xor_gc,
		              pt.x,
		              0,
		              pt.x,
		              wSciTE.GetID()->allocation.height - 1);
	} else {
		gdk_draw_line(wSciTE.GetID()->window, xor_gc,
		              0,
		              pt.y,
		              wSciTE.GetID()->allocation.width - 1,
		              pt.y);
	}
	ptOld = pt;
}

gint SciTEWindow::DividerExpose(GtkWidget *widget, GdkEventExpose *, SciTEWindow *sciThis) {
	//GtkStyle style = gtk_widget_get_default_style();
	GdkRectangle area;
	area.x = 0;
	area.y = 0;
	area.width = widget->allocation.width;
	area.height = widget->allocation.height;
	gdk_window_clear_area(widget->window,
	                      area.x, area.y, area.width, area.height);
	if (widget->allocation.width > widget->allocation.height) {
		// Horizontal divider
		gtk_paint_hline(widget->style, widget->window, GTK_STATE_NORMAL,
		                &area, widget, "vpaned",
		                0, widget->allocation.width - 1,
		                area.height / 2 - 1);
		gtk_paint_box (widget->style, widget->window,
		               GTK_STATE_NORMAL,
		               GTK_SHADOW_OUT,
		               &area, widget, "paned",
		               area.width - sciThis->heightBar * 2, 1,
		               sciThis->heightBar - 2, sciThis->heightBar - 2);
	} else {
		// Vertical divider
		gtk_paint_vline(widget->style, widget->window, GTK_STATE_NORMAL,
		                &area, widget, "hpaned",
		                0, widget->allocation.height - 1,
		                area.width / 2 - 1);
		gtk_paint_box (widget->style, widget->window,
		               GTK_STATE_NORMAL,
		               GTK_SHADOW_OUT,
		               &area, widget, "paned",
		               1, area.height - sciThis->heightBar * 2,
		               sciThis->heightBar - 2, sciThis->heightBar - 2);
	}
	return TRUE;
}

gint SciTEWindow::DividerMotion(GtkWidget *, GdkEventMotion *event, SciTEWindow *scitew) {
	if (scitew->capturedMouse) {
		int x = 0;
		int y = 0;
		GdkModifierType state;
		if (event->is_hint) {
			gdk_window_get_pointer(scitew->wSciTE.GetID()->window, &x, &y, &state);
			if (state & GDK_BUTTON1_MASK) {
				scitew->DividerXOR(scitew->ptOld);
				scitew->DividerXOR(Point(x, y));
			}
		}
	}
	return TRUE;
}

gint SciTEWindow::DividerPress(GtkWidget *, GdkEventButton *, SciTEWindow *scitew) {
	int x = 0;
	int y = 0;
	GdkModifierType state;
	gdk_window_get_pointer(scitew->wSciTE.GetID()->window, &x, &y, &state);
	scitew->ptStartDrag = Point(x, y);
	scitew->capturedMouse = true;
	scitew->heightOutputStartDrag = scitew->heightOutput;
	gtk_widget_grab_focus(GTK_WIDGET(scitew->wDivider.GetID()));
	gtk_grab_add(GTK_WIDGET(scitew->wDivider.GetID()));
	scitew->DividerXOR(scitew->ptStartDrag);
	return TRUE;
}

gint SciTEWindow::DividerRelease(GtkWidget *, GdkEventButton *, SciTEWindow *scitew) {
	if (scitew->capturedMouse) {
		scitew->capturedMouse = false;
		gtk_grab_remove(GTK_WIDGET(scitew->wDivider.GetID()));
		scitew->DividerXOR(scitew->ptOld);
		int x = 0;
		int y = 0;
		GdkModifierType state;
		gdk_window_get_pointer(scitew->wSciTE.GetID()->window, &x, &y, &state);
		scitew->MoveSplit(Point(x, y));
	}
	return TRUE;
}

void SciTEWindow::Run(const char *cmdLine) {

	GtkItemFactoryCallback menuSig = GtkItemFactoryCallback(MenuSignal);
	GtkItemFactoryEntry menuItems[] = {
	    {"/_File", NULL, NULL, 0, "<Branch>"},
	    {"/_File/tear", NULL, NULL, 0, "<Tearoff>"},
	    {"/File/_New", "<control>N", menuSig, IDM_NEW, 0},
	    {"/File/_Open", "<control>O", menuSig, IDM_OPEN, 0},
	    {"/File/_Close", "<control>W", menuSig, IDM_CLOSE, 0},
	    {"/File/_Save", "<control>S", menuSig, IDM_SAVE, 0},
	    {"/File/Save _As", NULL, menuSig, IDM_SAVEAS, 0},
	    {"/File/Save As _HTML", NULL, menuSig, IDM_SAVEASHTML, 0},
	    {"/File/sep1", NULL, NULL, 0, "<Separator>"},
	    {"/File/File1", "", menuSig, fileStackCmdID + 1, 0},
	    {"/File/File2", "", menuSig, fileStackCmdID + 2, 0},
	    {"/File/File3", "", menuSig, fileStackCmdID + 3, 0},
	    {"/File/File4", "", menuSig, fileStackCmdID + 4, 0},
	    {"/File/File5", "", menuSig, fileStackCmdID + 5, 0},
	    {"/File/File6", "", menuSig, fileStackCmdID + 6, 0},
	    {"/File/File7", "", menuSig, fileStackCmdID + 7, 0},
	    {"/File/File8", "", menuSig, fileStackCmdID + 8, 0},
	    {"/File/File9", "", menuSig, fileStackCmdID + 9, 0},
	    {"/File/sep2", NULL, menuSig, IDM_MRU_SEP, "<Separator>"},
	    {"/File/_Quit", "<control>Q", menuSig, IDM_QUIT, 0},

	    {"/_Edit", NULL, NULL, 0, "<Branch>"},
	    {"/_Edit/tear", NULL, NULL, 0, "<Tearoff>"},
	    {"/Edit/_Undo", "<control>Z", menuSig, IDM_UNDO, 0},
	    {"/Edit/_Redo", "<control>Y", menuSig, IDM_REDO, 0},
	    {"/Edit/sep1", NULL, NULL, 0, "<Separator>"},
	    {"/Edit/Cu_t", "<control>X", menuSig, IDM_CUT, 0},
	    {"/Edit/_Copy", "<control>C", menuSig, IDM_COPY, 0},
	    {"/Edit/_Paste", "<control>V", menuSig, IDM_PASTE, 0},
	    {"/Edit/_Delete", "Del", menuSig, IDM_CLEAR, 0},
	    {"/Edit/Select _All", "", menuSig, IDM_SELECTALL, 0},
	    {"/Edit/sep2", NULL, NULL, 0, "<Separator>"},
	    {"/Edit/_Find...", "<control>F", menuSig, IDM_FIND, 0},
	    {"/Edit/Find _Next", "F3", menuSig, IDM_FINDNEXT, 0},
	    {"/Edit/F_ind in Files...", "", menuSig, IDM_FINDINFILES, 0},
	    {"/Edit/R_eplace", "<control>H", menuSig, IDM_REPLACE, 0},
	    {"/Edit/sep3", NULL, NULL, 0, "<Separator>"},
	    {"/Edit/_Go To", "<control>G", menuSig, IDM_GOTO, 0},
	    {"/Edit/Match _Brace", "<control>B", menuSig, IDM_MATCHBRACE, 0},
	    {"/Edit/C_omplete Identifier", "<control>I", menuSig, IDM_COMPLETE, 0},

	    {"/_Tools", NULL, NULL, 0, "<Branch>"},
	    {"/_Tools/tear", NULL, NULL, 0, "<Tearoff>"},
	    {"/Tools/_Compile", "<control>F7", menuSig, IDM_COMPILE, 0},
	    {"/Tools/_Build", "F7", menuSig, IDM_BUILD, 0},
	    {"/Tools/_Go", "F5", menuSig, IDM_GO, 0},
	    {"/Tools/Tool0", "<control>0", menuSig, IDM_TOOLS + 0, 0},
	    {"/Tools/Tool1", "<control>1", menuSig, IDM_TOOLS + 1, 0},
	    {"/Tools/Tool2", "<control>2", menuSig, IDM_TOOLS + 2, 0},
	    {"/Tools/Tool3", "<control>3", menuSig, IDM_TOOLS + 3, 0},
	    {"/Tools/Tool4", "<control>4", menuSig, IDM_TOOLS + 4, 0},
	    {"/Tools/Tool5", "<control>5", menuSig, IDM_TOOLS + 5, 0},
	    {"/Tools/Tool6", "<control>6", menuSig, IDM_TOOLS + 6, 0},
	    {"/Tools/Tool7", "<control>7", menuSig, IDM_TOOLS + 7, 0},
	    {"/Tools/Tool8", "<control>8", menuSig, IDM_TOOLS + 8, 0},
	    {"/Tools/Tool9", "<control>9", menuSig, IDM_TOOLS + 9, 0},
	    {"/Tools/_Stop Executing", "<control>.", menuSig, IDM_STOPEXECUTE, NULL},
	    {"/Tools/sep1", NULL, NULL, 0, "<Separator>"},
	    {"/Tools/_Next Message", "F4", menuSig, IDM_NEXTMSG, 0},
	    {"/Tools/_Previous Message", "<shift>F4", menuSig, IDM_PREVMSG, 0},

	    {"/_Options", NULL, NULL, 0, "<Branch>"},
	    {"/_Options/tear", NULL, NULL, 0, "<Tearoff>"},
	    {"/Options/Vertical _Split", "", menuSig, IDM_SPLITVERTICAL, "<CheckItem>"},
	    {"/Options/sep1", NULL, NULL, 0, "<Separator>"},
	    {"/Options/View _Whitespace", "<control>T", menuSig, IDM_VIEWSPACE, "<CheckItem>"},
	    {"/Options/View _Margin", NULL, menuSig, IDM_SELMARGIN, "<CheckItem>"},
	    {"/Options/View Line _Numbers", "", menuSig, IDM_LINENUMBERMARGIN, "<CheckItem>"},
	    {"/Options/View _End of Line", "", menuSig, IDM_VIEWEOL, "<CheckItem>"},
#ifdef STATUSBAR
	    {"/Options/View Status Bar", "", menuSig, IDM_VIEWSTATUSBAR, "<CheckItem>"},
#endif
	    //{"/Options/Slave Output to Edit", "", menuSig, IDM_SLAVE, "<CheckItem>"},
	    {"/Options/sep2", NULL, NULL, 0, "<Separator>"},
	    {"/Options/Line End C_haracters", "", 0, 0, "<Branch>"},
	    {"/Options/Line End Characters/CR _+ LF", "", menuSig, IDM_EOL_CRLF, "<RadioItem>"},
	    {"/Options/_Line End Characters/_CR", "", menuSig, IDM_EOL_CR, "/Options/Line End Characters/CR + LF"},
	    {"/Options/_Line End Characters/_LF", "", menuSig, IDM_EOL_LF, "/Options/Line End Characters/CR + LF"},
	    {"/Options/_Convert Line End Characters", "", menuSig, IDM_EOL_CONVERT, 0},
	    {"/Options/sep3", NULL, NULL, 0, "<Separator>"},
	    {"/Options/Open _Local Options File", "", menuSig, IDM_OPENLOCALPROPERTIES, 0},
	    {"/Options/Open _Global Options File", "", menuSig, IDM_OPENGLOBALPROPERTIES, 0},

	    {"/_Help", NULL, NULL, 0, "<Branch>"},
	    {"/_Help/tear", NULL, NULL, 0, "<Tearoff>"},
	    {"/Help/About SciTE", "", menuSig, IDM_ABOUT, 0},
	};

	char *gthis = reinterpret_cast<char *>(this);

	wSciTE = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	GTK_WIDGET_UNSET_FLAGS(wSciTE.GetID(), GTK_CAN_FOCUS);
	gtk_window_set_policy(GTK_WINDOW(wSciTE.GetID()), TRUE, TRUE, FALSE);

	gtk_widget_set_events(wSciTE.GetID(),
	                      GDK_EXPOSURE_MASK
	                      | GDK_LEAVE_NOTIFY_MASK
	                      | GDK_BUTTON_PRESS_MASK
	                      | GDK_BUTTON_RELEASE_MASK
	                     );
	gtk_signal_connect(GTK_OBJECT(wSciTE.GetID()), "destroy",
	                   GTK_SIGNAL_FUNC(QuitSignal), gthis);

	gtk_window_set_title(GTK_WINDOW(wSciTE.GetID()), appName);
	int left = props.GetInt("position.left", 10);
	int top = props.GetInt("position.top", 30);
	int width = props.GetInt("position.width", 300);
	int height = props.GetInt("position.height", 400);
	gtk_widget_set_usize(GTK_WIDGET(wSciTE.GetID()), width, height);

	GtkWidget *boxMain = gtk_vbox_new(FALSE, 1);
	gtk_container_add(GTK_CONTAINER(wSciTE.GetID()), boxMain);
	GTK_WIDGET_UNSET_FLAGS(boxMain, GTK_CAN_FOCUS);

	int nItems = sizeof(menuItems) / sizeof(menuItems[0]);
	accelGroup = gtk_accel_group_new();
	itemFactory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accelGroup);
	gtk_item_factory_create_items(itemFactory, nItems, menuItems, gthis);

	gtk_accel_group_attach(accelGroup, GTK_OBJECT(wSciTE.GetID()));

	GtkWidget *handle_box = gtk_handle_box_new();

	gtk_container_add(GTK_CONTAINER(handle_box),
	                  gtk_item_factory_get_widget(itemFactory, "<main>"));

	gtk_box_pack_start(GTK_BOX(boxMain),
	                   handle_box,
	                   FALSE, FALSE, 0);

	content = gtk_fixed_new();
	GTK_WIDGET_UNSET_FLAGS(content.GetID(), GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(boxMain), content.GetID(), TRUE, TRUE, 0);

	gtk_signal_connect(GTK_OBJECT(content.GetID()), "size_allocate",
	                   GTK_SIGNAL_FUNC(MoveResize), gthis);

	wEditor = scintilla_new();
	scintilla_set_id(SCINTILLA(wEditor.GetID()), IDM_SRCWIN);
	gtk_fixed_put(GTK_FIXED(content.GetID()), wEditor.GetID(), 0, 0);
	gtk_widget_set_usize(wEditor.GetID(), 600, 600);
	gtk_signal_connect(GTK_OBJECT(wEditor.GetID()), "command",
	                   GtkSignalFunc(CommandSignal), this);
	gtk_signal_connect(GTK_OBJECT(wEditor.GetID()), "notify",
	                   GtkSignalFunc(NotifySignal), this);

	wDivider = gtk_drawing_area_new();
	gtk_signal_connect(GTK_OBJECT(wDivider.GetID()), "expose_event",
	                   GtkSignalFunc(DividerExpose), this);
	gtk_signal_connect(GTK_OBJECT(wDivider.GetID()), "motion_notify_event",
	                   GtkSignalFunc(DividerMotion), this);
	gtk_signal_connect(GTK_OBJECT(wDivider.GetID()), "button_press_event",
	                   GtkSignalFunc(DividerPress), this);
	gtk_signal_connect(GTK_OBJECT(wDivider.GetID()), "button_release_event",
	                   GtkSignalFunc(DividerRelease), this);
	gtk_widget_set_events(wDivider.GetID(),
	                      GDK_EXPOSURE_MASK
	                      | GDK_LEAVE_NOTIFY_MASK
	                      | GDK_BUTTON_PRESS_MASK
	                      | GDK_BUTTON_RELEASE_MASK
	                      | GDK_POINTER_MOTION_MASK
	                      | GDK_POINTER_MOTION_HINT_MASK
	                     );
	gtk_drawing_area_size(GTK_DRAWING_AREA(wDivider.GetID()), width, 10);
	gtk_fixed_put(GTK_FIXED(content.GetID()), wDivider.GetID(), 0, 600);

	wOutput = scintilla_new();
	scintilla_set_id(SCINTILLA(wOutput.GetID()), IDM_RUNWIN);
	gtk_fixed_put(GTK_FIXED(content.GetID()), wOutput.GetID(), 0, width);
	gtk_widget_set_usize(wOutput.GetID(), width, 100);
	gtk_signal_connect(GTK_OBJECT(wOutput.GetID()), "command",
	                   GtkSignalFunc(CommandSignal), this);
	gtk_signal_connect(GTK_OBJECT(wOutput.GetID()), "notify",
	                   GtkSignalFunc(NotifySignal), this);

	SendOutput(SCI_SETMARGINWIDTH, 0, 0);
	//SendOutput(SCI_SETCARETPERIOD, 0);

#ifdef STATUSBAR
	wStatusBar = gtk_statusbar_new();
	sbContextID = gtk_statusbar_get_context_id(
	                  GTK_STATUSBAR(wStatusBar.GetID()), "global");
	gtk_box_pack_start(GTK_BOX(boxMain), wStatusBar.GetID(), FALSE, FALSE, 0);
	gtk_statusbar_push(GTK_STATUSBAR(wStatusBar.GetID()), sbContextID, "Initial");
	sbVisible = false;
#endif

	gtk_widget_show_all(wSciTE.GetID());
#ifdef STATUSBAR
	gtk_widget_hide(GTK_WIDGET(wStatusBar.GetID()));
#endif
	gtk_widget_set_uposition(GTK_WIDGET(wSciTE.GetID()), left, top);
	SetFocus(wOutput.GetID());

	Open(cmdLine, true);
	CheckMenus();

	SizeSubWindows();

	gtk_main();
}

int main(int argc, char *argv[]) {
	gtk_init(&argc, &argv);
	SciTEWindow scite;
	if (argc > 1) {
		//Platform::DebugPrintf("args: %d %s\n", argc, argv[1]);
		scite.Run(argv[1]);
	} else {
		scite.Run("");
	}

	return 0;
}
#else
void SciTEWindow::Paint(Surface *surfaceWindow, PRectangle) {
	PRectangle rcClient = GetClientRectangle();
	int heightClient = rcClient.bottom - rcClient.top;
	int widthClient = rcClient.right - rcClient.left;

	surfaceWindow->FillRectangle(rcClient, GetSysColor(COLOR_3DFACE));

	if (splitVertical) {
		surfaceWindow->PenColour(GetSysColor(COLOR_3DHILIGHT));
		surfaceWindow->MoveTo(widthClient - (heightOutput + heightBar - 1), 0);
		surfaceWindow->LineTo(widthClient - (heightOutput + heightBar - 1), heightClient);

		surfaceWindow->PenColour(GetSysColor(COLOR_3DSHADOW));
		surfaceWindow->MoveTo(widthClient - (heightOutput + 2), 0);
		surfaceWindow->LineTo(widthClient - (heightOutput + 2), heightClient);

		surfaceWindow->PenColour(GetSysColor(COLOR_3DDKSHADOW));
		surfaceWindow->MoveTo(widthClient - (heightOutput + 1), 0);
		surfaceWindow->LineTo(widthClient - (heightOutput + 1), heightClient);
	} else {
		surfaceWindow->PenColour(GetSysColor(COLOR_3DHILIGHT));
		surfaceWindow->MoveTo(0, heightClient - (heightOutput + heightBar - 1));
		surfaceWindow->LineTo(widthClient, heightClient - (heightOutput + heightBar - 1));

		surfaceWindow->PenColour(GetSysColor(COLOR_3DSHADOW));
		surfaceWindow->MoveTo(0, heightClient - (heightOutput + 2));
		surfaceWindow->LineTo(widthClient, heightClient - (heightOutput + 2));

		surfaceWindow->PenColour(GetSysColor(COLOR_3DDKSHADOW));
		surfaceWindow->MoveTo(0, heightClient - (heightOutput + 1));
		surfaceWindow->LineTo(widthClient, heightClient - (heightOutput + 1));
	}
}

LRESULT SciTEWindow::WndProc(UINT iMessage, WPARAM wParam, LPARAM lParam) {
	//Platform::DebugPrintf("start wnd proc %x %x\n",iMessage, wSciTE.GetID());
	switch (iMessage) {

	case WM_CREATE:
		wEditor = ::CreateWindow(
		              "Scintilla",
		              "Source",
		              WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_CLIPCHILDREN,
		              0, 0,
		              100, 100,
		              wSciTE.GetID(),
		              reinterpret_cast<HMENU>(IDM_SRCWIN),
		              hInstance,
		              0);
		if (!wEditor.Created())
			exit(FALSE);
		wEditor.Show();
		SendEditor(SCI_ASSIGNCMDKEY, VK_RETURN, SCI_NEWLINE);
		SendEditor(SCI_ASSIGNCMDKEY, VK_TAB, SCI_TAB);
		SendEditor(SCI_ASSIGNCMDKEY, VK_TAB | (SHIFT_PRESSED << 16), SCI_BACKTAB);
		SetFocus(wEditor.GetID());

		wOutput = ::CreateWindow(
		              "Scintilla",
		              "Run",
		              WS_CHILD | WS_VSCROLL | WS_HSCROLL | WS_CLIPCHILDREN,
		              0, 0,
		              100, 100,
		              wSciTE.GetID(),
		              reinterpret_cast<HMENU>(IDM_RUNWIN),
		              hInstance,
		              0);
		if (!wOutput.Created())
			exit(FALSE);
		wOutput.Show();
		// No selection margin on output window
		SendOutput(SCI_SETMARGINWIDTH, 0, 0);
		//SendOutput(SCI_SETCARETPERIOD, 0);
		SendOutput(SCI_ASSIGNCMDKEY, VK_RETURN, SCI_NEWLINE);
		SendOutput(SCI_ASSIGNCMDKEY, VK_TAB, SCI_TAB);
		SendOutput(SCI_ASSIGNCMDKEY, VK_TAB | (SHIFT_PRESSED << 16), SCI_BACKTAB);
		::DragAcceptFiles(wSciTE.GetID(), true);
		break;

	case WM_PAINT: {
			PAINTSTRUCT ps;
			::BeginPaint(wSciTE.GetID(), &ps);
			Surface surfaceWindow;
			surfaceWindow.Init(ps.hdc);
			PRectangle rcPaint(ps.rcPaint.left, ps.rcPaint.top, ps.rcPaint.right, ps.rcPaint.bottom);
			Paint(&surfaceWindow, rcPaint);
			surfaceWindow.Release();
			::EndPaint(wSciTE.GetID(), &ps);
			return 0;
		}

	case WM_COMMAND:
		Command(wParam, lParam);
		break;

	case WM_SYSCOMMAND:
		if (wParam == SC_CLOSE) {
			QuitProgram();
		} else {
			return ::DefWindowProc(wSciTE.GetID(), iMessage, wParam, lParam);
		}
		break;

	case WM_NOTIFY:
		Notify(reinterpret_cast<SCNotification *>(lParam));
		break;

	case WM_KEYDOWN:
		//Platform::DebugPrintf("keydown %d %x %x\n",iMessage, wParam, lParam);
		break;

	case WM_KEYUP:
		//Platform::DebugPrintf("keyup %d %x %x\n",iMessage, wParam, lParam);
		break;

	case WM_SIZE:
		//Platform::DebugPrintf("size %d %x %x\n",iMessage, wParam, lParam);
		if (wParam != 1)
			SizeSubWindows();
		break;

	case WM_GETMINMAXINFO:
		return ::DefWindowProc(wSciTE.GetID(), iMessage, wParam, lParam);

	case WM_LBUTTONDOWN:
		ptStartDrag = Point::FromLong(lParam);
		capturedMouse = true;
		heightOutputStartDrag = heightOutput;
		::SetCapture(wSciTE.GetID());
		//Platform::DebugPrintf("Click %x %x\n", wParam, lParam);
		break;

	case WM_MOUSEMOVE:
		if (capturedMouse) {
			MoveSplit(Point::FromLong(lParam));
		}
		break;

	case WM_LBUTTONUP:
		if (capturedMouse) {
			MoveSplit(Point::FromLong(lParam));
			capturedMouse = false;
			::ReleaseCapture();
		}
		break;

	case WM_SETCURSOR:
		if (ControlIDOfCommand(lParam) == HTCLIENT) {
			Point ptCursor;
			GetCursorPos(reinterpret_cast<POINT *>(&ptCursor));
			PRectangle rcScintilla = wEditor.GetPosition();
			PRectangle rcOutput = wOutput.GetPosition();
			if (!rcScintilla.Contains(ptCursor) && !rcOutput.Contains(ptCursor)) {
				wSciTE.SetCursor(splitVertical ? Window::cursorHoriz : Window::cursorVert);
				return TRUE;
			}
		}
		return ::DefWindowProc(wSciTE.GetID(), iMessage, wParam, lParam);

	case WM_INITMENU:
		CheckMenus();
		break;

	case WM_DESTROY:
		// Unhook before destroying wEditor and thus its document.
		wEditor.Destroy();
		wOutput.Destroy();
		::PostQuitMessage(0);
		break;

	case WM_SETTINGCHANGE:
		//Platform::DebugPrintf("** Setting Changed\n");
		SendEditor(WM_SETTINGCHANGE, wParam, lParam);
		SendOutput(WM_SETTINGCHANGE, wParam, lParam);
		break;

	case WM_SYSCOLORCHANGE:
		//Platform::DebugPrintf("** Color Changed\n");
		SendEditor(WM_SYSCOLORCHANGE, wParam, lParam);
		SendOutput(WM_SYSCOLORCHANGE, wParam, lParam);
		break;

	case WM_PALETTECHANGED:
		//Platform::DebugPrintf("** Palette Changed\n");
		if (wParam != reinterpret_cast<WPARAM>(wSciTE.GetID())) {
			SendEditor(WM_PALETTECHANGED, wParam, lParam);
			//SendOutput(WM_PALETTECHANGED, wParam, lParam);
		}

		break;

	case WM_QUERYNEWPALETTE:
		//Platform::DebugPrintf("** Query palette\n");
		SendEditor(WM_QUERYNEWPALETTE, wParam, lParam);
		//SendOutput(WM_QUERYNEWPALETTE, wParam, lParam);
		return TRUE;

	case WM_ACTIVATEAPP:
		SendEditor(EM_HIDESELECTION, !wParam);
		// Do not want to display dialog yet as may be in middle of system mouse capture
		::PostMessage(wSciTE.GetID(), WM_COMMAND, IDM_ACTIVATE, wParam);
		break;

	case WM_ACTIVATE:
		SetFocus(wEditor.GetID());
		break;

	case WM_DROPFILES:
		if (SaveIfUnsure() != IDCANCEL) {
			char pathDropped[MAX_PATH];
			HDROP hdrop = reinterpret_cast<HDROP>(wParam);
			int filesDropped = DragQueryFile(hdrop, 0xffffffff, pathDropped, sizeof(pathDropped));
			if (filesDropped == 1) {
				DragQueryFile(hdrop, 0, pathDropped, sizeof(pathDropped));
				Open(pathDropped);
			}
			DragFinish(hdrop);
		}
		break;

	default:
		if (iMessage == RegisterWindowMessage(FINDMSGSTRING)) {
			HandleFindReplace();
			//Platform::DebugPrintf("Find/replacemessage %x %x\n", wParam, lParam);
		} else {
			//Platform::DebugPrintf("default wnd proc %x %d %d\n",iMessage, wParam, lParam);
			return ::DefWindowProc(wSciTE.GetID(), iMessage, wParam, lParam);
		}
	}
	//Platform::DebugPrintf("end wnd proc\n");
	return 0l;
}

LRESULT PASCAL SciTEWindow::TWndProc(
    HWND hWnd, UINT iMessage, WPARAM wParam, LPARAM lParam) {
	//Platform::DebugPrintf("W:%x M:%d WP:%x L:%x\n", hWnd, iMessage, wParam, lParam);

	// Find C++ object associated with window.
	SciTEWindow *scite = reinterpret_cast<SciTEWindow *>(GetWindowLong(hWnd, 0));
	// scite will be zero if WM_CREATE not seen yet
	if (scite == 0) {
		if (iMessage == WM_CREATE) {
			LPCREATESTRUCT cs = reinterpret_cast<LPCREATESTRUCT>(lParam);
			scite = reinterpret_cast<SciTEWindow *>(cs->lpCreateParams);
			scite->wSciTE = hWnd;
			SetWindowLong(hWnd, 0, reinterpret_cast<LONG>(scite));
			return scite->WndProc(iMessage, wParam, lParam);
		} else
			return DefWindowProc(hWnd, iMessage, wParam, lParam);
	} else
		return scite->WndProc(iMessage, wParam, lParam);
}

int PASCAL WinMain(HINSTANCE hInstance, HINSTANCE, LPSTR lpszCmdLine, int) {

	//Platform::DebugPrintf("Command line is \n%s\n<<", lpszCmdLine);

	//	HRESULT hr = OleInitialize(NULL);
	//	if (FAILED(hr)) {
	//		return 0;
	//	}

	HACCEL hAccTable = LoadAccelerators(hInstance, "ACCELS");

	SciTEWindow::Register(hInstance);
#ifdef STATIC_BUILD
	Scintilla_RegisterClasses(hInstance);
#else
::LoadLibrary("Scintilla.DLL");
#endif

	SciTEWindow MainWind;

	MainWind.Run(lpszCmdLine);

	bool going = true;
	MSG msg;
	msg.wParam = 0;
	while (going) {
		going = GetMessage(&msg, NULL, 0, 0);
		if (currentDlg && going) {
			//Platform::DebugPrintf("DLG %x %x\n", currentDlg, msg.message);
			if (!IsDialogMessage(currentDlg, &msg)) {
				if (TranslateAccelerator(msg.hwnd, hAccTable, &msg) == 0) {
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
		} else if (going) {
			//if (TranslateAccelerator(msg.hwnd,hAccTable,&msg) == 0) {
			if (TranslateAccelerator(MainWind.GetID(), hAccTable, &msg) == 0) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}

	//	OleUninitialize();
	return msg.wParam;
}
#endif
